#include <Mono1MG.h>
#include <FiniteElement.h>
#include <ElmMatVec.h>
#include <readOrMakeGrid.h>
#include <ErrorNorms.h>
#include <PrecML.h>
#include <MLIter.h>

// functor for analytical solution, used below in resultReport and fillEssBC
class Mono1MGAnalSol : public FieldFunc
{
  Mono1MG* data;
public:
  Mono1MGAnalSol (Mono1MG* boss) : data(boss) {}
  virtual real valuePt (const Ptv(real)& x, real t);
};

real Mono1MGAnalSol:: valuePt (const Ptv(real)& x, real t)
{ 
  const int nsd = x.size();
  real prod= exp(- nsd*sqr(M_PI)*data->diffusion_coeff*t);
  for (int i = 1; i <= nsd; i++)
    prod *= sin (M_PI*x(i));
  return prod;
}


Mono1MG:: Mono1MG ()
{
  DPTRACE("Mono1MG constructor");
}

Mono1MG:: Mono1MG (String cmt)
{
  DPTRACE("Mono1MG constructor");

  cellmodeltype = cmt;
}



Mono1MG:: ~Mono1MG ()
{
  DPTRACE("Mono1MG destructor");
}

void Mono1MG:: adm (MenuSystem & menu)
{  
  DPTRACE("Mono1MG::adm");
  SimCase::attach (menu);
  define (menu);
  menu.prompt();
  scan();
}

void Mono1MG:: define (MenuSystem & menu, int level)
{ 
  DPTRACE("Heat::define"); 
  
  menu.addItem (level, "time parameters", "step and domain for time",
                "dt =0.1 t in [0,1]");
  
  menu.addItem (level,          // menu level (1 is main, 2 is first submenu)
                "gridfile",     // menu command/name
                "file or preprocessor command",  // help/description
                "mygrid.grid"); // default answer

  menu.addItem (level, "diffusion coefficient", "k", "1.0");
  
  menu.addItem (level, "error integration point type", "err_itg",
                "type of itg rule used in ErrorNorms::Lnorm",
                "GAUSS_POINTS", "S/GAUSS_POINTS/NODAL_POINTS");
  
  menu.addItem (level, "theta", "theta",
		"time discretization parameter",
		"1.0", "R1");

  menu.addItem (level, "compute error norms", "e_norms",
		"compute error norms",
		"false", "S");

  SaveSimRes::defineStatic (menu, level+1);
  LinEqAdmFE::defineStatic(menu,level+1);
  FEM::defineStatic(menu,level+1);
  MGtools:: defineStatic(menu,level+1);

  Cells::defineStatic(menu, level+1);

  cells = new Cells(cellmodeltype, menu, level);

}

void Mono1MG:: scan ()
{ 
  DPTRACE("Mono1MG::scan");
  MenuSystem& menu = SimCase::getMenuSystem();
  
  lineq.rebind(new LinEqAdmFE());
  lineq->scan(menu);

  mgtools.rebind(new MGtools(*this));
  mgtools->attach(*lineq);
  mgtools->scan(menu);

  if (lineq->getSolver().description().contains("multilevel"))
    {
      MLIter& ml = CAST_REF(lineq->getSolver(), MLIter);
      ml.attach(mgtools->getMLSolver());
    }

  if (lineq->getPrec().description().contains("multilevel"))
    {
      PrecML& ml = CAST_REF(lineq->getPrec(), PrecML);
      ml.attach(mgtools->getMLSolver());
    }

  mgtools->initStructure();

  no_of_grids = mgtools->getNoOfSpaces();
  grid.rebind (mgtools->getGrid(no_of_grids));
  dof.rebind (mgtools->getDof (no_of_grids));

  //String gridfile = menu.get ("gridfile");
  assignEnum(error_itg_pt_tp,menu.get("error integration point type"));
  // read and make grid 
  //grid.rebind(new GridFE());
  //readOrMakeGrid (*grid, gridfile);  // fill grid
  //s_o << "\nGrid: " << gridfile << "\n";

  diffusion_coeff = menu.get ("diffusion coefficient").getReal();
  compute_error_norms = menu.get("compute error norms").getBool();
 
  FEM::scan(menu);
  vanal.rebind (new Mono1MGAnalSol (this));
  database.rebind (new SaveSimRes());
  database->scan (menu, grid->getNoSpaceDim());

  // fill handles with objects:
  v.     rebind (new FieldFE  (*grid, "v"));
  v_prev.rebind (new FieldFE  (*grid, "v_prev"));
  error. rebind (new FieldFE  (*grid, "error"));
  flux.  rebind (new FieldsFE (*grid, "flux"));
  v_summary.attach (*v);  v_summary.init ();  // see man page

  //dof.rebind( new DegFreeFE(*grid,1));
  tip.rebind(new TimePrm());
  tip->scan(menu.get("time parameters"));
  linsol.redim(grid->getNoNodes()); 
  linsol.fill(0.0);
  // attach linsol to lineq as the unknown in the linear system 
  lineq->attach(linsol);

  theta = menu.get("theta").getReal();

  cells->scan(menu);
  cells->attachGridAndTimer(grid(),tip());
  cells->attachDiff(this);

  mgtools->print(s_o, 1); 

  createTrackPoints();
}


void Mono1MG:: createTrackPoints()
{
  int nsd = grid->getNoSpaceDim();

  real distance; bool exact;

  Ptv(real) track1(nsd), track2(nsd), track3(nsd);

  tracknodes.redim(3);

  //2D code
  track1(1) = 0; track1(2) = 2.2;
  track2(1) = 0; track2(2) = 3.1;
  track3(1) = 0; track3(2) = 5.0;

  tracknodes(1) = grid->nearestPoint(track1, distance, exact);
  tracknodes(2) = grid->nearestPoint(track2, distance, exact);
  tracknodes(3) = grid->nearestPoint(track3, distance, exact);
}

void Mono1MG:: dumpTrackValues()
{
  int no_track_points = tracknodes.size();

  Os trackfile("points.track", APPEND);

  trackfile << tip->time();
  for (int i = 1; i <= no_track_points; i++)
    trackfile << " " << v->valueNode(tracknodes(i));

  trackfile << endl;

}


void Mono1MG:: mgFillEssBC (int grid_level)
{
  /* no ess bc for this version
  GridFE& current_grid = mgtools->getGrid(grid_level);
  DegFreeFE& current_dof = mgtools->getDof(grid_level);

  current_dof.initEssBC ();
  Ptv(real) x;
  const int nno = current_grid.getNoNodes();
  for (int i = 1; i <= nno; i++)
    {  
      if (current_grid.boNode (i)) { // at any of the boundaries ?
	x = current_grid.getCoor(i);
        current_dof.fillEssBC (i, g(x));    
      }
    }
  */
  #ifdef DP_DEBUG
   current_dof.printEssBC (s_o, 2);  // for checking the ess BC
  #endif
  
}

void Mono1MG:: solveProblem () { timeLoop(); }

void Mono1MG:: setIC ()
{
  /*
  v_prev->fill (*vanal, 0.0); // call uanal for t=0 at all nodes
  *v = *v_prev;               // make u ready for dump also at t=0
  
  */
  
  int nsd = grid->getNoSpaceDim();

  //smooth initial conditions independent of ICfile:
  Ptv(real) x;
  real r, r1, val;
  real peak = 0;
  real rest = -95.87;
  for (int i = 1; i <= grid->getNoNodes(); i++){
    x = grid->getCoor(i);
    r = x.norm();
    if (r <= 1.0)
      val = peak;
    else if (r >=2.0)
      val = rest;
    else 
      val = (peak-rest)*(2.0*r*r*r-9.0*r*r+12.0*r-5.0)+peak;
 
    v_prev->valueNode(i) = val;
  }
  *v = *v_prev;
  //dump();
}

void Mono1MG:: timeLoop()
{
  tip->initTimeLoop(); 
  setIC();
  dump();

  real t0 = tip->time();
  real dt = tip->Delta();

  cells->forward(t0, dt*theta);
 
  while(!tip->finished()) {
    solveAtThisTimeStep();
    t0 = tip->time();
    tip->increaseTime();
    
    if (!tip->finished()){
      if (database->noPlot4thisTimeLevel(tip.getPtr()))
	cells->forward(t0+theta*dt, dt);
      else {
	cells->forward(t0+theta*dt, dt*(1.0-theta));
	dump();
	if (compute_error_norms)
	  computeErrorNorms();
	cells->forward(t0+dt, dt*theta);
      }
    }
    
    *v_prev = *v;
    dumpTrackValues(); 
  }
  cells->forward(t0+theta, dt*(1.0-theta));
  dump();
  if (compute_error_norms)
    computeErrorNorms();

  LinEqStatBlk& statistics= lineq->getPerformance( );
  statistics.print(s_o);
  lineq->getMonitorListPrm().print(s_o);
}

void Mono1MG:: solveAtThisTimeStep ()
{
  //fillEssBC ();                // incorporate time-dep. ess. b.c.

  mgtools->makeSystemMl();
  //makeSystem (*dof,*lineq);    // FEM's assembly algorithm

  dof->field2vec (*v, linsol); // use most recent u as start vector
  lineq->solve ();             // solve linear system
  s_o << "t=" << tip->time();
  int niterations; bool c;     // for iterative solver statistics
  if (lineq->getStatistics(niterations,c))  // iterative solver?
    s_o << oform("  solver%sconverged in %3d iterations\n",
                 c ? " " : " not ",niterations);
  s_o << '\n'; s_o.flush();    // flush() forces output _now_

  dof->vec2field (linsol, *v); // load linsol into the field v
  //database->dump (*v, tip.getPtr(), "some comment if desired..."); 
  v_summary.update (tip->time()); // keep track of max/min u etc

  // compute smooth flux -k*grad(v);
  //FEM::makeFlux (*flux, *v);
  //database->dump(*flux, tip.getPtr(), "smooth flux -k*grad(v)");
}

void Mono1MG:: fillEssBC ()
{
  DPTRACE("Mono1MG::fillEssBC");
  /* no essential boundary conditions for now...
  dof->initEssBC ();
  Ptv(real) x;
  const int nno = grid->getNoNodes();
  for (int i = 1; i <= nno; i++)
    {  
      if (grid->boNode (i)) { // at any of the boundaries ?
	x = grid->getCoor(i);
        dof->fillEssBC (i, g(x));    
      }
    }
  */
#ifdef DP_DEBUG
   dof->printEssBC (s_o, 2);  // for checking the essential boundary cond.
#endif
}

// inherited from class FEM, no need to implement this if we only have
// to perform volume integration, but it is often used for debugging
void Mono1MG:: calcElmMatVec (int e, ElmMatVec& elmat, FiniteElement& fe)
{
  FEM::calcElmMatVec (e, elmat, fe);
#if DP_DEBUG >= 2
  elmat.print (s_o);
#endif
}


void Mono1MG:: integrands (ElmMatVec& elmat, const FiniteElement& fe)
{     
  const real detJxW = fe.detJxW();
  const int  nsd = fe.getNoSpaceDim();
  const int  nbf = fe.getNoBasisFunc();
  const real dt  = tip->Delta();
  const real t   = tip->time();
  Ptv(real) x = fe.getGlobalEvalPt();
  real gradNi_gradNj, gradNi_gradv;
  // v at the previous time level at the current integration point:
  real vp_pt = v_prev->valueFEM(fe);  
  Ptv(real) v_deriv;
  v_prev->derivativeFEM(v_deriv,fe);  

  const real k_value = k (fe, t);

  real r;
  //correct scale factor for radial problem solved as 1D:
  if (nsd == 1)
    r = x(1);  //for 2D solved as 1D (3D: x(1)*x(1))
  else 
    r = 1.0;

  int i,j,s;
  for(i = 1; i <= nbf; i++) {   
    for(j = 1; j <= nbf; j++) {
      gradNi_gradNj =0;
      for(s = 1; s <= nsd; s++)
        gradNi_gradNj += fe.dN(i,s)*fe.dN(j,s); 

      elmat.A(i,j) +=(fe.N(i)*fe.N(j) + 
		      theta*dt*k_value*gradNi_gradNj)*r*detJxW;
    } 
    gradNi_gradv = 0;
    for (s = 1; s <= nsd; s++)
      gradNi_gradv += fe.dN(i,s)*v_deriv(s);

    elmat.b(i)+= (fe.N(i)*vp_pt-(1.0-theta)*dt*k_value*gradNi_gradv)*r*detJxW;
  }
}

void Mono1MG:: computeErrorNorms()
{ 
  DBP("entering computeErrorNorms");
  SimResFile reference_file("test");
  real t = tip->getTime();
  
  Handle(FieldFE) v_ref;
  v_ref.rebind(new FieldFE(grid(),"v_ref"));
  Handle(FieldFE) v_ref_1D;
  v_ref_1D.rebind(new FieldFE);

  DBP("ok");
  SimResFile::readField(*v_ref_1D, reference_file, "v", t,true);

  //must convert 1d field to 2d...
  int nno = grid->getNoNodes();
  DBP("ok");
  int nsd = grid->getNoSpaceDim();
  Ptv(real) x(nsd);
  Ptv(real) r(1); 
  real ref_value;
  for (int n = 1; n <= nno; n++){
    grid->getCoor(x,n);
    r(1) = x.norm();
    //hack to avoid out-of-range values:
    if (r(1) > 5)
      r(1) = 5.0; 
    ref_value = v_ref_1D->valuePt(r);
    v_ref->valueNode(n) = ref_value;
  }
  //database->dump (v_ref_1D(),tip.getPtr(),"v_ref");
  database->dump (v_ref(),tip.getPtr(),"v_ref");

  real v_1, v_2, v_inf;

  ErrorNorms::discreteLnorm(*v_ref,*v,v_1, v_2, v_inf);
  Os errorfile ("errors", NEWFILE); 
  errorfile << "dt = " <<tip->Delta() << " nodes = " << nno;
  errorfile << " L1 = "<< v_1 << " L2 = " << v_2 << " L_inf= " << v_inf << endl; 
 
}

void Mono1MG:: resultReport ()
{ 
  DPTRACE("Mono1MG::resultReport");

}

void Mono1MG:: dump()
{
  database->dump(v(),tip.getPtr(), "v");
  //if (compute_error_norms)
  //  computeErrorNorms();
  
  if (grid->getNoSpaceDim() == 2){
    //dumping 5 line plots to matlab files:
    real angle, r;
    int no_of_points = 500;
    Mat(real) values(no_of_points, 5);
    Vec(real) radius(no_of_points);
    real step = 4.0/(no_of_points-1);
    
    Ptv(real) pt(2);
    for (int i = 1; i <= no_of_points; i++){
      r = (i-1)*step;
      for (int j = 1; j <= 5; j++){
	angle = M_PI*(j-1)/4.0;
	pt(1) = cos(angle)*r;
	pt(2) = sin(angle)*r;
	
	values(i,j) = v->valuePt(pt);
      }
      radius(i) = r;
    }
    
    //write the results to a file:
    Os matlabfile ("lineplots", NEWFILE);
    for (int i = 1; i <= no_of_points; i++){
      matlabfile << radius(i); 
      for (int j = 1; j <= 5; j++)
	matlabfile << " " << values(i,j);

      matlabfile << endl;
    }
  }
}

real Mono1MG:: f (const FiniteElement& /*fe*/, real /*t*/) { return 0; }
real Mono1MG:: k (const FiniteElement& /*fe*/, real /*t*/)
               { return diffusion_coeff; }
real Mono1MG:: I (const Ptv(real)& x) { return vanal->valuePt(x,0); }
real Mono1MG:: g (const Ptv(real)& /*x*/, real /*t*/) { return 0; }
