#include <SDIRK.h>

void SDIRK::attachAndInit(Winslow* prob)
{
  DPTRACE("SDIRK::attachAndInit");

  problem = prob;
  no_of_states = problem->evalSize();

  work = 0;
  rejects = 0;

  //accuracy parameters
  min_dt = 1e-5;
  tol = 0.00001;
  s_fac = 0.85;

  first_step = true;

  //parameters for the Newton iterations
  etamin = 1.0e-10;
  eta    = etamin;
  kappa  = 1.0e-1;
  maxits = 7;

  //set method coefficients
  gamma = 0.43586652150845899942;
  a21   = 0.43586652150845899942;
  b1_   = 0.49056338842178057060;
  b2_   = 0.073570090069760429950;
  b1    = 0.30880996997674652335;
  b2    = 1.4905633884217805707;
  b3    = -1.2352398799069860932;
  e1    = b1 - b1_;
  e2    = b2 - b2_;
  e3    = b3 - gamma;

  c2 = 0.87173304301691799884;
  c3 = 1.0;

  J.redim(no_of_states);
  A.redim(no_of_states);

  current_state.redim(no_of_states);
  internal_state.redim(no_of_states);
  y2.redim(no_of_states);
  y3.redim(no_of_states);
  z2.redim(no_of_states);
  z3.redim(no_of_states);
  z4.redim(no_of_states);
  k1.redim(no_of_states);
  k2.redim(no_of_states);
  k3.redim(no_of_states);
  k4.redim(no_of_states);

  temp.redim(no_of_states);
  error.redim(no_of_states);
  prev_steps.redim(no_of_states);

  //vectors used in the Newton iterations
  dk.redim(no_of_states);
  prev_dk.redim(no_of_states);
  b_vec.redim(no_of_states);
}

void SDIRK::computeFirstJacobian()
{
  DPTRACE("SDIRK::computeFirstJacobian");
  J.fill(0.0);
  problem->setConstantJacobianElms(J);
  recomputeJacobian();
  new_jacobian = true;
}


void SDIRK::recomputeJacobian()
{
  DPTRACE("SDIRK::recomputeJacobian");
  /*
  Vec(real) s;
 
  problemEvalStateDot(ydot1, current_state);
  for (int j = 1; j <= no_of_states; j++){
    internal_state = current_state;
    internal_state.y(j) += h;
    problemEvalStateDot(ydot2,internal_state);  
    dF.add(ydot2,'-',ydot1);
    dF.mult(1.0/h);
    for (int i = 1; i <= no_of_states; i++)
      J(i,j) = dF(i);      
  } 
  */
  problem->recomputeJacobian(J,current_state); 
  new_jacobian = true;
}


void SDIRK::advance
(Vec(real)& y, real global_t, real global_dt/*, bool db*/)
{
  DPTRACE("SDIRK::advance"); 
  real dt;
  real gtl=0;
  //cout << "y = "; y.print(cout); cout << endl;

  for (int i = 1; i <= no_of_states; i++)
    current_state.y(i) = y(i);  
  current_state.time() = global_t;

  local_t = 0.0;

  if (first_step){
    dt = global_dt;
    computeFirstJacobian();
  }
  else {
    dt = max(s_fac*prev_dt*pow((tol/loc_error),0.25),min_dt); 
    dt = min(dt, global_dt);
    //recomputeJacobian();
  }

  //  cout << y(31) << " ";

  while (local_t < global_dt){
    gtl++;
    for (int i = 1; i <= no_of_states; i++)
      current_state.y(i) = y(i);  
    current_state.time() = global_t+local_t;

    if (dt*1.1 >= global_dt-local_t)
      dt = global_dt-local_t;

    if (its >= 2)
      recomputeJacobian();
    if (abs((dt-prev_dt)/prev_dt) > 0.2 || new_jacobian)
      computeNewA(dt);

    advanceOneStep(y, dt);
    //    cout << dt << " ";
    prev_dt = dt;
    local_t += dt;
    dt = max(s_fac*prev_dt*pow((tol/loc_error),0.25),min_dt);

  }

  first_step = false;
  //  cout << "@@@@@@@@@@@@@@@@@@@@@ " << dt << " " << gtl << "\n";
}

void SDIRK::advanceOneStep(Vec(real)& y, real& dt)
{
  bool accepted = doStep(y, dt);
  while (!accepted && dt > min_dt){
    if (!new_jacobian)
      recomputeJacobian();
    reduceTimeStep(dt);
    computeNewA(dt);
    accepted = doStep(y,dt);
  }
  new_jacobian = false;  
}

bool SDIRK::doStep(Vec(real)& y, const real& delta_t)
{
  DPTRACE("SDIRK::advanceOneStep");
  Vec(real) prev_steps, error(no_of_states);
  //DBP("y0");
  //current_state.yvec().print(cout); cout << endl; 

  bool step_ok;  
  //compute stage derivative k1 (explicitely)
  problemEvalStateDot(k1, current_state);
  /*
  if (debug){ 
    cout << "k1 = " << endl;
    k1.print(cout); cout << endl;
  }
  */

  //k1 predicts k2 (seems to be the best choice)  
  k2 = k1;
  //compute k2
  prev_steps = k1;
  prev_steps.mult(a21);  
  step_ok = computeStageVal(k2, prev_steps, delta_t);
  /*
  if (debug){ 
    cout << "k2 = " << endl;
    k2.print(cout); cout << endl;
  }  
  */
  if (step_ok){
    //compute a predictor for k3
    k3 = k2;
    prev_steps = k1;
    prev_steps.mult(b1_);
    prev_steps.add(prev_steps,b2_,k2);
    step_ok = computeStageVal(k3, prev_steps, delta_t);

  }
  /*
  if (debug){ 
    cout << "k3 = " << endl;
    k3.print(cout); cout << endl;
  }
  */
  if (step_ok){
    //s_o << "Step 4" << endl;
    //use k3 as predictor for k4
    k4 = k3;
    prev_steps = k1;
    prev_steps.mult(b1);
    prev_steps.add(prev_steps, b2, k2);
    prev_steps.add(prev_steps, b3, k3);
    step_ok = computeStageVal(k4, prev_steps, delta_t);
  }
  /*
  if (debug){ 
    cout << "k4 = " << endl;
    k4.print(cout); cout << endl;
  }
  */
  if (step_ok){
    //set new value of y = y0+delta_t*(b1*k1+...gamma*k4)
    temp.add(b1,k1,b2,k2);
    temp.add(temp,b3,k3);
    temp.add(temp,gamma,k4);
 
    y.add(current_state.yvec(),delta_t,temp);
    error.add(b1-b1_,k1,b2-b2_,k2);
    error.add(error,b3-gamma,k3);
    error.add(error, gamma, k4);
    error.mult(delta_t);
    problem->forwBackLU(A,error,error);
 
    loc_error = error.norm();
    if (loc_error > tol){
      
      step_ok = false;
    }          
  }

  return step_ok;
 
}

bool SDIRK::computeStageVal
    (Vec(real)& k, const Vec(real)& prev_steps, const real& dt)
{
  DPTRACE("SDIRK::computeStageVal");
  its = 0;
  real theta;
  bool conv_ok = true;

  eta = pow(max(eta,etamin),0.8);

  do {
    internal_state = current_state;
    for (int i = 1; i <= no_of_states; i++)
      internal_state.y(i) += dt*prev_steps(i)+gamma*dt*k(i);
   
    problemEvalStateDot(temp, internal_state);
    b_vec.add(-1.0,k,1.0, temp);

    //solve the linear system to find dk
    prev_dk = dk;
    problem->forwBackLU(A,dk,b_vec);
   
    if (its >= 1){
      theta = dk.norm()/prev_dk.norm();
      if (!(theta <= 1)){
        //s_o << "Solver diverges, reducing time step" << endl;
	rejects ++;
        conv_ok = false;
        break;
      }
      eta = theta/(1.0-theta);
    }

    if (its > maxits){
      //s_o << "Not converged in "<< maxits <<" iterations, reducing time step" << endl;
      rejects ++;
      conv_ok = false;
      break;
    } 

    //compute new k-value k=k+dk
    k.add(k,dk); 
    
    its ++;
  } while (eta * dk.norm() > kappa*tol);
  work += its;      
  //s_o << "Newton solver converged in " << its << " iterations." << endl; 
  //k.print(s_o);
  //s_o << endl; s_o << endl;
  return conv_ok;
}

void SDIRK::computeNewA(const real& dt)
{
  DPTRACE("SDIRK::computeNewA"); 
  A = J;
  A.mult(-dt*gamma);
 
  for(int i = 1; i <=31; i++){
    A(i,i) += 1.0;
  }

  A.resetFact();
  A.factLU();

}

void SDIRK::reduceTimeStep(real& delta_t)
{  
  real reduction_factor = 1.0;
  real e = loc_error;
  do {
    reduction_factor *= 0.6; //seems to be the best choice 
    e *= 0.1;
  } while  (e > tol); 
  

  //s_o << "Time step reduced from " << delta_t;
  delta_t = max(reduction_factor*delta_t,min_dt);
  //s_o << " to " << delta_t <<endl;delta_t = max(0.5*delta_t,min_dt);
}

void SDIRK::problemEvalStateDot(Vec(real)& ydot,const DynamicState& state)
{ 
  problem->evalStateDot(ydot,state); 
}








