#ifndef DynSys_h_IS_INCLUDED
#define DynSys_h_IS_INCLUDED

#include <HandleId.h>
#include <btTypeInfo.h> // btbtTypeInfo.h iflg namechanges2.7
#include <Dpmacros.h>
#include <DynamicState.h>
#include <default_op.h>
#include<Ptv_real.h> 

class ExplicitODEeqUDC;
class ODEsolver;

/*<DynSys:*/
class DynSys: public virtual HandleId
{
  friend class CombinedDynSys;

  protected:
  Handle(DynamicState)        sys_state;
  Handle(DynamicState)  comunicate_state;  // for comunication

  public:

  DynSys() {        sys_state.rebind(new DynamicState);
             comunicate_state.rebind(new DynamicState); }

  virtual ~DynSys() {}

  virtual void init()
  {
    sys_state->redim(evalSize() ,evalNoDiscrite());
    setIV(sys_state());
  }

  DynamicState&      state() { return sys_state();        }

  // communication between models (for intermediate state values)
  DynamicState&   comState() { return comunicate_state(); }

  virtual int evalSize() const = 0;
  virtual int evalNoDiscrite() const = 0;

  virtual void setIV(DynamicState& state) = 0;

  virtual real evalErrorEstimate(const Vec(real)& l) const
  { return l.norm(Linf); }

  virtual int evalNoConstraints() const { return 0; }

  // evaluate the constraint functions
  virtual void evalConstraints (Vec(real)& /*results*/,
				DynamicState& /*state*/);

  virtual void handleRoot(DynamicState& s, int i);

  virtual void setParameters(Ptv(real)& p ) = 0;

  CLASS_INFO
  DEF_VIRTUAL_CAST(ExplicitODEeqUDC)


};
/*>DynSys:*/

/*Class:DynSys
NAME: DynSys - specification of a dynamic system.

SYNTAX: @DynSys

KEYWORDS:
    base class, IVP , ODE

DESCRIPTION:

   The class is an abstract base class for all classes specifying an
   dynamic system like a system of ordinary differential equations.
    A user shall not derive directly from  this  class but from a class like
   "ExplicitODEeqUDC". The class will be an input class for the "ODEsolver"
    class hirarchy.

CONSTRUCTORS AND INITIALIZATION:

   The only constructor is the empty constructor, as the class
   must be given an implenentation in a derived class.


MEMBER FUNCTIONS:

  $\begin{description}

     `\item[`"evalSize"`]` - evaluate the size of the system, must be
      defined by a derived class.

     `\item[`"evalNoDiscrite"`]` - evaluate the discrite size of the
       system, must be defined by a derived class.

     `\item[`"setIV"`]` - set the initial values of the dynamic state
       given by the argument, must be defined by a derived class.

       `\item[`"state"`]` - returns with the current dynamic state of
       the system.

        `\item[`"comState"`]` - returns a dynamic state, undefined
	within this class ( cf. class CombinedExplixitODE )

     `\item[`"evalErrorEstimate"`]` -
       estimates a scalar error to be compared to a tolerance
       parameter in eg. adaptive time loops.
       The function can be redefined for eg.
       considering the error of only some of the components of the
       vector or to scale the vector by some quantity.
       NOTE: Default is the max norm, meaning that the worst case
       component is compared to the given tolerance.
       Although there is a default implementation of the function
       any derived problem class should carefully consider different
       implementations in accordance with the given problem.

  $\end{description}


EXAMPLES:

   See "class Oscillator"

SEE ALSO:

   class "ExplicitODEeqUDC", class "ODEsolver" , class "ExplicitODEsolver"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */


#define ClassType DynSys
#include <Handle.h>
#undef ClassType


#endif
