#ifndef DynamicState_h_IS_INCLUDED
#define DynamicState_h_IS_INCLUDED

#include <Vec_real.h>
#include <VecSimple_int.h>

class CombinedDynamicState;
class Handle(DynamicState);

/*<DynamicState:*/
class DynamicState: public HandleId
{
  friend class CombinedDynamicState;

  real defaultT;
  real*         t; // time

  protected:

  Vec(real)    y_; // state vector
  Vec(real) ydot_; // state derivatives
  Vec(real)    d_; // discrete states

  public:

  DynamicState()
  { t = &(defaultT); }

  DynamicState(int n,int d=0)
  {
    t = &(defaultT);
    redim(n,d);
  }

  DynamicState(const DynamicState& state);

  virtual bool redim(int states, int discrites = 0);

  int size() const        { return y_.size(); }
  int noDiscrite() const  { return d_.size(); }

  DynamicState& operator=(const DynamicState& s)
  {
    y_ = s.y_; ydot_ = s.ydot_; time() = s.time(); d_ = s.d_; return *this;
  }



   // assignement
  const real& y   (int i) const  { return y_(i);    }
        real& y   (int i)        { return y_(i);    }
  const real& ydot(int i) const  { return ydot_(i); }
        real& ydot(int i)        { return ydot_(i); }
  const real& d   (int i) const  { return d_(i);    }
        real& d   (int i)        { return d_(i);    }
  real& time ()     const  { return *t;       }


  const Vec(real)&    yvec() const { return y_;      }
  const Vec(real)& ydotvec() const { return ydot_;   }
  const Vec(real)&    dvec() const { return d_;      }

  virtual void makeItSimilar(Handle(DynamicState)& s);

  void print(Os& os) const;
  void scan(Is& is);


  CLASS_INFO
  DEF_VIRTUAL_CAST(CombinedDynamicState)

};
/*>DynamicState:*/

/*Class:DynamicState

NAME: DynamicState - Dynamic state, ODE

SYNTAX: @DynamicState

KEYWORDS: ODE, IVP, Dynamic State

DESCRIPTION:

  The class contains the state vector, state derivative, time and the discrite
  state vector. This is the eccential parts of the solution of a system
  of ordinary differential equations.

CONSTRUCTORS AND INITIALIZATION:

  The sizes given in the constructor or in the "redim" function are used
  to initialize the class. The first argument is the sizes of the state vector
  and the second the size of the discrite state vector.


MEMBER FUNCTIONS:
   $\begin{description}

    `\item[`"redim"`]` - redimensions the state vector and the discrite
       state vector. The first argument is the sizes of the state vector
       and the derivative, the second the size of the discrite state vector.

    `\item[`"size"`]` - the size of the state vector.

    `\item[`"noDiscrites"`]` - the size of the discrite state vector.

    `\item[`"operator="`]` - assignment, eg. "state = tmpstate;"
          puts the dynamic state equal to "tmpstate".

    `\item[`"print"`]` - print

    `\item[`"scan"`]` - reads the format used by print

    `\item[`"makeItSimilar"`]` - makes an imprint of the current object.


   $\end{description}

EXAMPLES:
|  DynamicState mystate(10,1)
|  for (int i=1; i<=10;i++)
|    mystate.y(1) = 1;
|

SEE ALSO:

  class "CombinedDyanicState" and class "ODEsolver"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */


#define ClassType DynamicState
#include <Handle.h>
#undef ClassType


#endif
