#include <EmsODEsolver.h>

INIT_CLASS_INFO(AdamsMoulton4)


//---------------------------------------------------------------------------
AdamsMoulton4::AdamsMoulton4(ODEsolver_prm& pm)
//--------------------------------------------------------------------------
{
  attach(pm.exp_sys());
  startmethod.rebind(new RungeKuttaF45(pm));
}



//---------------------------------------------------------------------------
 void AdamsMoulton4::attachStartMethod(ExplicitODEsolver& method)
//---------------------------------------------------------------------------
{
  if (! problem.ok())
    errorFP("AdamsMoulton4::attachStartMethod",
	    "There is no problem formualtion attached to the solver.\n"
	    "Please call attach before attachStartMethod ! ");

  if (method.getNoOfStepsMethod() != 1)
    {
       errorFP("AdamsMoulton4::attachStartMethod",
	       "The start method must be a single step method !");

    }
  if (method.getOrder() < 4)
    {
       warningFP("AdamsMoulton4::attachStartMethod",
	       "The start method is of lower order than four !");

    }
  startmethod.rebind(method);
  startmethod->attach(problem());

}



//---------------------------------------------------------------------------
void AdamsMoulton4:: newIntegrate()
//---------------------------------------------------------------------------
{

  ExplicitODEsolver::newIntegrate();

  // only reallocate if necessary:
  if (startmethod.ok())
    startmethod->attach(problem());
  else
    attachStartMethod(*new RungeKuttaF45);


  f0.redim(no_of_states);
  f1.redim(no_of_states);
  f2.redim(no_of_states);
  f3.redim(no_of_states);
  yp.redim(no_of_states);
  yc.redim(no_of_states);
  problem->state().makeItSimilar(s0);
  nsteps  = 3;
  ststeps = 0;

}



//---------------------------------------------------------------------------
void AdamsMoulton4:: advance (real dt)
//---------------------------------------------------------------------------
{
  int i;
  if (dt != last_dt) nsteps = 3;

  last_dt = dt;
  last_state() = current_state();

  if (nsteps > 0)
    {
      if      (nsteps == 3) { f0 = current_state().ydotvec(); }
      else if (nsteps == 2) { f1 = current_state().ydotvec(); }
      else if (nsteps == 1) { f2 = current_state().ydotvec(); }
      startmethod->advance(dt);
      nsteps--;
      ststeps++;
    }
  else
    {
      for (i=1; i <= no_of_discrites; i++)
	s0().d(i) = current_state().d(i);

      f3 = current_state().ydotvec();
      // predicotor :
      for (i = 1; i <= no_of_states; i++)
	{
	  yp(i) = current_state().y(i) + (dt/24.0) *
	    ( 55.0 * f3(i) - 59.0*f2(i) + 37.0*f1(i)
				      - 9.0* f0(i));
	  s0().y(i) = yp(i);
	}

      // Corrector

      f0 = f1; f1 = f2; f2 =f3;

      s0->time() = current_state().time()+dt;
      problemEvalStateDot(f3,s0());

      for (i = 1; i <= no_of_states; i++)
	{
	  yc(i) = current_state().y(i) + (dt/24.0) *
	    ( 9.0*f3(i) +19.0*f2(i) - 5.0*f1(i)
	      + f0(i));
	  current_state().y(i) = yc(i);
	}

      // update f3
      current_state().time() += dt;
      problemEvalStateDot(current_state());
      f3 = current_state().ydotvec();
    }
  error_computed = false;


}



//---------------------------------------------------------------------------
real AdamsMoulton4::advanceE(real dt)
//---------------------------------------------------------------------------
{
  real err;
  int i;

  if (dt != last_dt) nsteps = 3;

  last_state() = current_state();
  last_dt = dt;

  if (nsteps > 0)
    {
      if (nsteps == 3)      { f0 = current_state().ydotvec(); }
      else if (nsteps == 2) { f1 = current_state().ydotvec(); }
      else if (nsteps == 1) { f2 = current_state().ydotvec(); }

      err = startmethod->advanceE (dt);
      locerror = startmethod->errorVec();
      nsteps--;
      ststeps++;
    }
  else
    {
      for (i=1; i <= no_of_discrites; i++)
	s0().d(i) = current_state().d(i);

      f3 = current_state().ydotvec();

      // predicotor :
      for (i = 1; i <= no_of_states; i++)
	{
	  yp(i) = current_state().y(i) + (dt/24.0) *
	    ( 55.0 * f3(i) - 59.0*f2(i) + 37.0*f1(i)
				      - 9.0* f0(i));
	  s0().y(i) = yp(i);
	}

      // Corrector

      f0 = f1; f1 = f2; f2 =f3;

      s0->time() = current_state().time()+dt;
      problemEvalStateDot(f3,s0());

      for (i = 1; i <= no_of_states; i++)
	{
	  yc(i) = current_state().y(i) + (dt/24.0) *
	    ( 9.0*f3(i) +19.0*f2(i) - 5.0*f1(i)
	      + f0(i));
	  current_state().y(i) = yc(i);
	  locerror(i) = yc(i) - yp(i);
	}

      // update f3
      current_state().time() += dt;
      problemEvalStateDot(current_state());
      err = problem->evalErrorEstimate(locerror);
    }
  error_computed = true;
  return err;

}
