
#include <ErkODEsolver.h>

INIT_CLASS_INFO(Euler1)
INIT_CLASS_INFO(Euler2)
INIT_CLASS_INFO(RungeKutta23)
INIT_CLASS_INFO(RungeKutta4)
INIT_CLASS_INFO(RungeKuttaF45)
INIT_CLASS_INFO(RungeKuttaCK45)
INIT_CLASS_INFO(RungeKuttaDP45)


//---------------------------------------------------------------------------
Euler1::Euler1(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{
  attach(pm.exp_sys());
}



//----------------------------------------------------------------------------
void Euler1::newIntegrate()
//----------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

}


//---------------------------------------------------------------------------
void Euler1:: advance (real dt)
//---------------------------------------------------------------------------
{
  last_state() = current_state();
  last_dt = dt;

  // update state
  for (int i=1; i <= no_of_states; i++)
    current_state().y(i) = current_state().y(i) + dt * current_state().ydot(i);

  // update state derivative
  current_state().time() += dt;
  problemEvalStateDot(current_state());

  error_computed = false;
}



//---------------------------------------------------------------------------
Euler2::Euler2(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{
  attach(pm.exp_sys());
}



//----------------------------------------------------------------------------
void Euler2::newIntegrate()
//----------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);
  k2.redim(no_of_states);
}


//---------------------------------------------------------------------------
void Euler2:: advance (real dt)
//---------------------------------------------------------------------------
{
  int i;
  last_state() = current_state();
  last_dt = dt;

  for (i=1; i<= no_of_discrites; i++)
    s0().d(i) = current_state().d(i);

  // k1 = current_state->ydot

  for (i=1 ; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt * current_state().ydot(i);

  s0().time() = current_state().time()+dt;
  problemEvalStateDot(k2, s0());

  for (i=1 ; i <= no_of_states; i++)
    current_state().y(i) = current_state().y(i)+
      dt*0.5*( (current_state->ydot(i))
	       + k2(i) );

  // update state
  current_state().time() += dt;
  problemEvalStateDot(current_state());
  error_computed = false;

}


//---------------------------------------------------------------------------
real Euler2:: advanceE (real dt)
//---------------------------------------------------------------------------
{
  int i;
  advance(dt);
  error_computed = true;

  // k1 = last_state().ydot
  for (i=1 ; i <= no_of_states; i++)
    locerror(i) = 0.5*dt*((last_state().ydot(i)) - k2(i));

  return problem->evalErrorEstimate(locerror);

}



//---------------------------------------------------------------------------
RungeKutta23::RungeKutta23(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{
  attach(pm.exp_sys());
}



//----------------------------------------------------------------------------
void RungeKutta23::newIntegrate()
//----------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);

  k2.redim(no_of_states);
  k3.redim(no_of_states);

}


//---------------------------------------------------------------------------
void RungeKutta23:: advance (real dt)
//---------------------------------------------------------------------------
{
  int i;
  last_state()=current_state();
  last_dt = dt;
  for (i=1; i <= no_of_discrites; i++)
    s0().d(i) = current_state().d(i);

  for (i=1 ; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt * current_state().ydot(i);

  s0().time() = current_state().time()+dt;

  problemEvalStateDot(k2, s0());

  for (i=1 ; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i) +
      0.25*dt*((current_state().ydot(i))+k2(i));

  s0().time() = current_state().time() + 0.5*dt;

  problemEvalStateDot(k3,s0());

  for (i=1 ; i <= no_of_states; i++)
    current_state().y(i) = current_state().y(i)
      + dt*(1.0/6.0)*((current_state().ydot(i))
      + k2(i) + 4.0*k3(i));

  current_state().time() += dt;
  problemEvalStateDot(current_state());

  error_computed = false;

}


//---------------------------------------------------------------------------
real RungeKutta23:: advanceE (real dt)
//---------------------------------------------------------------------------
{
  int i;
  advance(dt);
  error_computed = true;

  for (i=1 ; i <= no_of_states; i++)
    locerror(i) = dt*((-1.0/3.0)*((last_state().ydot(i))+k2(i))
		      +(2.0/3.0)*k3(i));

  return problem->evalErrorEstimate(locerror);

}



//---------------------------------------------------------------------------
RungeKutta4::RungeKutta4(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{ attach(pm.exp_sys()); }



//---------------------------------------------------------------------------
void RungeKutta4:: newIntegrate()
//---------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);

  k2.redim(no_of_states);
  k3.redim(no_of_states);
  k4.redim(no_of_states);


}


//---------------------------------------------------------------------------
void RungeKutta4:: advance (real dt)
//---------------------------------------------------------------------------
{
  dt2 = dt/2.0;
  dt6 = dt/6.0;

  last_state() = current_state();
  last_dt = dt;
  int i;
  for (i=1; i<= no_of_discrites; i++)
    s0().d(i) = current_state().d(i);

  for (i = 1; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i)  +  dt2 * (current_state().ydot(i));

  s0().time()= current_state().time()+dt2;
  problemEvalStateDot(k2, s0());

  for (i = 1; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i)  +  dt2 * k2(i);

  problemEvalStateDot (k3, s0());

  for (i = 1; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*k3(i);

  s0().time() = current_state().time()+dt;
  problemEvalStateDot (k4, s0());

  for (i = 1; i <= no_of_states; i++)
    current_state().y(i) = current_state().y(i)  +
      dt6 * ( (current_state().ydot(i))+ 2.0 * k2(i) + 2.0 * k3(i) + k4(i) );

  error_computed = false;
  current_state().time() += dt;

  problemEvalStateDot(current_state());

}



//---------------------------------------------------------------------------
RungeKuttaF45::RungeKuttaF45(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{ attach(pm.exp_sys()); }


//---------------------------------------------------------------------------
void RungeKuttaF45:: newIntegrate()
//---------------------------------------------------------------------------
{

  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);
  k2.redim(no_of_states);
  k3.redim(no_of_states);
  k4.redim(no_of_states);
  k5.redim(no_of_states);
  k6.redim(no_of_states);

}



//---------------------------------------------------------------------------
void RungeKuttaF45:: advance (real dt)
//---------------------------------------------------------------------------
{

  last_state()=current_state();
  last_dt = dt;
  int i;
  for (i=1; i<= no_of_discrites; i++)
    s0().d(i) = current_state().d(i);

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + (dt/4.0) * (current_state->ydot(i));

  s0().time()=current_state().time()+dt/4.0;

  problemEvalStateDot (k2, s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*(
      (3.0/32.0)*(current_state->ydot(i)) + (9.0/32.0)*k2(i));

  s0().time() = current_state().time()+(3.0*dt)/8.0;
  problemEvalStateDot (k3,s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*(
      (1932.0/2197.0)*(current_state->ydot(i))
    - (7200.0/2197.0)*k2(i)
    + (7296.0/2197.0)*k3(i));

  s0().time() = current_state().time()+dt*12.0/13.0;
  problemEvalStateDot (k4,s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i)=current_state().y(i) + dt*(
      (439.0/216.0)*(current_state->ydot(i))
    -            8.0*k2(i)
    + (3680.0/513.0)*k3(i)
    - (845.0/4104.0)*k4(i));

  s0().time() = current_state().time()+dt;
  problemEvalStateDot (k5, s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*(
      (-8.0/27.0)*(current_state->ydot(i))
    +             2.0*k2(i)
    - (3544.0/2565.0)*k3(i)
    + (1859.0/4104.0)*k4(i)
    -     (11.0/40.0)*k5(i));

  s0().time()= current_state().time()+dt/2.0;
  problemEvalStateDot (k6, s0());


  for (i=1 ; i <= no_of_states; i++)
    {
      current_state().y(i) = current_state().y(i) +
	dt*((16.0/135.0)*(current_state->ydot(i)) + (6656.0/12825.0)*k3(i)
			+ k4(i)*(28561.0/56430.0) - k5(i)*(9.0/50.0)
			+ k6(i)*(2.0/55.0)) ;
    }
  current_state().time() += dt;

  problemEvalStateDot(current_state());

  error_computed = false;



}



//---------------------------------------------------------------------------
real  RungeKuttaF45::advanceE(real dt)
//---------------------------------------------------------------------------
{
  int i;

  advance(dt);
  error_computed = true;

  for (i=1 ; i <= no_of_states; i++)
    {
      locerror(i) = dt*( (last_state().ydot(i))/360.0
			 - (128.0/4275.0)*k3(i)
			 -(2197.0/75240.0)*k4(i)
			 + k5(i)/50.0 + (2.0/55.0)*k6(i));
    }

  return problem->evalErrorEstimate(locerror);

}


//------------------------------------------------

static const real a2 = 1.0/5.0;
static const real a3 = 3.0/10.0;
static const real a4 = 3.0/5.0;
static const real a5 = 1.0;
static const real a6 = 7.0/8.0;

static const real b21 = 1.0/5.0;
static const real b31 = 3.0/40.0;
static const real b32 = 9.0/40.0;
static const real b41 = 3.0/10.0;
static const real b42 = -9.0/10.0;
static const real b43 = 6.0/5.0;
static const real b51 = -11.0/54.0;
static const real b52 = 5.0/2.0;
static const real b53 = -70.0/27.0;
static const real b54 = 35.0/27.0;
static const real b61 = 1631.0/55296.0;
static const real b62 = 175.0/512.0;
static const real b63 = 575.0/13824.0;
static const real b64 = 44275.0/110592.0;
static const real b65 = 253.0/4096.0;

static const real c1 = 37.0/378.0;
static const real c2 = 0;
static const real c3 = 250.0/621.0;
static const real c4 = 125.0/594.0;
static const real c5 = 0;
static const real c6 = 512.0/1771.0;


static const real cc1 = c1 - 2825.0/27648.0;
static const real cc2 = 0;
static const real cc3 = c3 - 18575.0/48384.0;
static const real cc4 = c4 - 13525.0/55296.0;
static const real cc5 = c5 - 277.0/14336.0;
static const real cc6 = c6 - 0.25;


//---------------------------------------------------------------------------
RungeKuttaCK45::RungeKuttaCK45(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{ attach(pm.exp_sys()); }


//---------------------------------------------------------------------------
void RungeKuttaCK45:: newIntegrate()
//---------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);
  k2.redim(no_of_states);
  k3.redim(no_of_states);
  k4.redim(no_of_states);
  k5.redim(no_of_states);
  k6.redim(no_of_states);

}



//---------------------------------------------------------------------------
void RungeKuttaCK45:: advance (real dt)
//---------------------------------------------------------------------------
{ 
  int i;
  
  last_state()=current_state();
  last_dt = dt;

  problemEvalStateDot (k2, current_state());

  for (i=1; i<= no_of_discrites; i++)  s0().d(i) = current_state().d(i);

  for (i=1; i<=no_of_states; i++)
    {
    s0().y(i) = current_state().y(i) + (dt*b21*(current_state->ydot(i)));
//    cout << s0().y(i) << " " << current_state().y(i) << " " << current_state->ydot(i) << endl;
    }
  s0().time() = current_state().time()+dt*a2;
  problemEvalStateDot (k2, s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*( b31*(current_state->ydot(i))
					  + b32*k2(i));

// cout << s0().y(9) << " " << current_state->ydot(9) << " " << k2(9) << endl;

  s0().time()= current_state().time()+dt*a3;
  problemEvalStateDot (k3,s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*(   b41*(current_state->ydot(i))
					    + b42*k2(i)
					    + b43*k3(i));
//for (i=1; i<=no_of_states; i++)
// cout << s0().y(i) << " "; 
//cout << endl;

  s0().time() = current_state().time()+dt*a4;
  problemEvalStateDot (k4,s0());
  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) +
      dt*(b51*(current_state->ydot(i))
	  +b52*k2(i)
	  +b53*k3(i)
	  +b54*k4(i));


  s0().time() = current_state().time()+dt*a5;
  problemEvalStateDot (k5, s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i)
      + dt*(b61*(current_state->ydot(i))
	    +b62*k2(i)
	    +b63*k3(i)
	    +b64*k4(i)
	    +b65*k5(i));


  s0().time() = current_state().time()+dt*a6;
  problemEvalStateDot (k6, s0());

  for (i=1 ; i <= no_of_states; i++)
    {
      current_state().y(i) = current_state().y(i) +
	dt*( c1*(current_state->ydot(i))
	     + c3* k3(i) +c4*k4(i) +
	     + c6* k6(i)) ;
    }

  error_computed = false;
  current_state().time() += dt;

  problemEvalStateDot(current_state());

}

//---------------------------------------------------------------------------
real RungeKuttaCK45::advanceE(real dt)
//---------------------------------------------------------------------------
{
  int i;

//  cout << "inside RungeKuttaCK45::advanceE(real dt) \n";

//  cout << current_state().time() << " " <<  dt << endl; 
//  for (i=1 ; i <= no_of_states; i++)
//     cout << current_state().y(i) << " ";
//  cout << endl;

  
  advance(dt);

  error_computed = true;
  for (i=1 ; i <= no_of_states; i++)
    {
      locerror(i) = dt*( cc1*(last_state().ydot(i))+ cc3 *k3(i)
		     + cc4*k4(i) + cc5*k5(i) + cc6*k6(i) );
    }

//  cout << current_state().time() << " " <<  dt << endl; 
//  for (i=1 ; i <= no_of_states; i++)
//     cout << current_state().y(i) << " ";
//  cout << endl;


  return problem->evalErrorEstimate(locerror);

}





//------------------------------------------------
//Parameters Dormand-Prince

static const real ad2 = 1.0/5.0;
static const real ad3 = 3.0/10.0;
static const real ad4 = 4.0/5.0;
static const real ad5 = 8.0/9.0;
static const real ad6 = 1.0;
static const real ad7 = 1.0;
static const real bd21 = 1.0/5.0;
static const real bd31 = 3.0/40.0;
static const real bd32 = 9.0/40.0;
static const real bd41 = 44.0/45.0;
static const real bd42 = -56.0/15.0;
static const real bd43 = 32.0/9.0;
static const real bd51 = 19372.0/6561.0;
static const real bd52 = -25360.0/2187.0;
static const real bd53 = 64448.0/6561.0;
static const real bd54 = -212.0/729.0;
static const real bd61 = 9017.0/3168.0;
static const real bd62 = -355.0/33.0;
static const real bd63 = 46732.0/5247.0;
static const real bd64 = 49.0/176.0;
static const real bd65 = -5103.0/18656.0;

static const real bd71 = 35.0/384.0;
static const real bd72 = 0;
static const real bd73 = 500.0/1113.0;
static const real bd74 = 125.0/192.0;
static const real bd75 = -2187.0/6784.0;
static const real bd76 = 11.0/84.0;


static const real cd1 = bd71- 5179.0 / 57600.0;
static const real cd2 = 0;
static const real cd3 = bd73-7571.0/16695.0;
static const real cd4 = bd74 - 393.0/640.0;
static const real cd5 = bd75 - (- 92097.0 / 339200.0);
static const real cd6 = bd76 - 187.0 / 2100.0;
static const real cd7 = - 1.0/40.0;



//---------------------------------------------------------------------------
RungeKuttaDP45::RungeKuttaDP45(ODEsolver_prm& pm)
//---------------------------------------------------------------------------
{ attach(pm.exp_sys()); }


//---------------------------------------------------------------------------
void RungeKuttaDP45:: newIntegrate()
//---------------------------------------------------------------------------
{
  ExplicitODEsolver::newIntegrate();

  problem->state().makeItSimilar(s0);
  k2.redim(no_of_states);
  k3.redim(no_of_states);
  k4.redim(no_of_states);
  k5.redim(no_of_states);
  k6.redim(no_of_states);
  k7.redim(no_of_states);

}



//---------------------------------------------------------------------------
void RungeKuttaDP45:: advance (real dt)
//---------------------------------------------------------------------------
{
  int i;

  last_state() = current_state();
  last_dt    = dt;
  for (i=1; i<= no_of_discrites; i++) s0().d(i) = current_state().d(i);


  // k2
  for (i=1; i <= no_of_states; i++)
    s0().y(i) = current_state().y(i) + (dt*bd21*(current_state->ydot(i)));
  s0().time() = current_state().time()+dt*ad2;
  problemEvalStateDot (k2, s0());

  // k3
  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*( bd31*(current_state->ydot(i))
					  + bd32*k2(i));
  s0().time() = current_state().time()+dt*ad3;
  problemEvalStateDot (k3,s0());

  // k4
  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) + dt*(
      bd41*(current_state->ydot(i))
    + bd42*k2(i)
    + bd43*k3(i));
  s0().time() = current_state().time()+dt*ad4;
  problemEvalStateDot (k4,s0());

  // k5
  for (i=1; i<=no_of_states; i++)
    s0().y(i)=current_state().y(i) + dt*(
      bd51*(current_state->ydot(i))
    + bd52*k2(i)
    + bd53*k3(i)
    + bd54*k4(i));

  s0().time() = current_state().time()+dt*ad5;
  problemEvalStateDot (k5,s0());

  for (i=1; i<=no_of_states; i++)
    s0().y(i) = current_state().y(i) +
      dt*( bd61*(current_state->ydot(i))
	   + bd62*k2(i)
	   + bd63*k3(i)
	   + bd64*k4(i)
	   + bd65*k5(i));

  s0().time() = current_state().time()+dt*ad6;
  problemEvalStateDot (k6, s0());


  for (i=1; i<=no_of_states; i++)
    current_state().y(i) = current_state().y(i)
      + dt*( bd71*(current_state->ydot(i))
	     +bd72*k2(i)
	     +bd73*k3(i)
	     +bd74*k4(i)
	     +bd75*k5(i)
	     +bd76*k6(i));


  current_state().time() += dt;
  problemEvalStateDot(current_state());

  error_computed = false;


}

//---------------------------------------------------------------------------
real  RungeKuttaDP45::advanceE(real dt)
//---------------------------------------------------------------------------
{
  int i;
  advance(dt);
  error_computed = true;

  for (i=1 ; i <= no_of_states; i++)
    {
      locerror(i) =  dt*( cd1 *(last_state().ydot(i))
			  + cd2*k2(i) + cd3* k3(i)
			  + cd4*k4(i) + cd5*k5(i) + cd6* k6(i)
			  + cd7*current_state->ydot(i));
                            // k7 = current_state().ydot
    }


  return problem->evalErrorEstimate(locerror);

}



//--------------------------------------------------------------------------
void RungeKuttaDP45::interpolate( real teta, DynamicState& s,
				  bool update_ydot)
//---------------------------------------------------------------------------
{
  int i;
  if (teta<=0) { s = last_state();}
  else
  if (teta>=1) { s = current_state(); }
  else
  if (last_dt == 0) { s = current_state(); }
  else
    {

      real t = last_state().time() + last_dt*teta;

      real b1,b3,b4,b5,b6;

      b1 = teta*(1.0 + teta*(-1337.0/480.0 + teta*(1039.0/360.0 +
						   teta*(-1163.0/1152.0))));

      b3 = 100.0 * teta*teta*(1054.0/9275.0 + teta*(-4682.0/27825.0
	       				    + teta*(379.0/5565.0)))/3.0;
      b4 = -5.0*teta*teta*(27.0/40.0 + teta*(-9.0/5.0 + teta*(83.0/96.0)))/2.0;
      b5 = 18225.0*teta*teta*(-3.0/250.0 + teta*(22.0/375.0
						 +teta*(-37.0/600.0)))/848.0;
      b6 = -22.0*teta*teta*(-3.0/10.0 + teta*(29.0/30.0 +
					      teta*(-17.0/24.0)))/7.0;

      for (i = 1; i <= no_of_states; i++)
	{

	  s.y(i) = last_state().y(i) +
	    last_dt * ( b1 * (last_state().ydot(i))
			+ b3 * k3(i) + b4 * k4(i)
			+ b5* k5(i) + b6 * k6(i));

	}

      s.time() = t;
      for (i=1;i <= no_of_discrites; i++)
	s.d(i) = last_state().d(i);

      if (update_ydot)
	problemEvalStateDot(s);
    }
}
