#ifndef ErkODEsolver_h_IS_INCLUDED
#define ErkODEsolver_h_IS_INCLUDED

#include <ExplicitODEsolver.h>

//==========================================================

/*<Euler1:*/
class Euler1: public ExplicitODEsolver
{

  public:

  Euler1() {}
  Euler1(ODEsolver_prm& pm);

  ~Euler1() {}

  virtual void newIntegrate();

  virtual void advance  (real dt);

  int getOrder() const {return 1;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(Euler1)
  ASSIGNMENT_OPERATOR(Euler1)
  CLASS_INFO
  VIRTUAL_CAST(Euler1)

};
/*>Euler1:*/

/*Class:Euler1
NAME: Euler1 - numerical solution of ordinary differential equations
      by a forward Euler method.

SYNTAX: @Euler1

KEYWORDS:  solver, O(h) , IVP

DESCRIPTION:

   The class implements the method known as Euler method.

   The method is not recommended for any practical use as it is of
   low order and is not very stable.

CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
     initializes the class.


MEMBER FUNCTIONS:

  $\begin{description}

  `\item[`"advance"`]` - integrate one time step.

  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:
   class "RungeKutta4" and class "RungeKuttaF45"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */



/*<Euler2:*/
class Euler2: public ExplicitODEsolver
{
  protected:
Handle(DynamicState) s0;

  Vec(real)    k2;

  public:

  Euler2() {}
  Euler2(ODEsolver_prm& pm);

  ~Euler2() {}

  virtual void newIntegrate();

  virtual void advance  (real dt);
  virtual real advanceE( real dt);

  int getOrder() const { return 2; }
  int getNoOfStepsMethod() const { return 1; }


  COPY_CONSTRUCTOR(Euler2)
  ASSIGNMENT_OPERATOR(Euler2)
  CLASS_INFO
  VIRTUAL_CAST(Euler2)

};
/*>Euler2:*/

/*Class:Euler2
NAME: Euler2 - numerical solution of ordinary differential equations
      by the modified Euler method.

SYNTAX: @Euler2

KEYWORDS:  solver, O( `\(` h^2 `\)` ) , IVP

DESCRIPTION:

   The class implements the method known as a modified Euler method.


CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
    initializes the class.


MEMBER FUNCTIONS:

  $\begin{description}
  `\item[`"advance"`]` -
        integrate one time step.
  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:
   class "RungeKutta4" and class "RungeKuttaF45"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */


/*<RungeKutta23:*/
class RungeKutta23 : public ExplicitODEsolver
{
  protected:
Handle(DynamicState) s0;
  Vec(real)    k0;

  Vec(real)    k2;
  Vec(real)    k3;



  public:
  RungeKutta23() {}
  RungeKutta23(ODEsolver_prm& pm);

  ~RungeKutta23() {}

  virtual void newIntegrate();

  virtual void advance  ( real dt);
  virtual real advanceE( real dt);

  int getOrder() const {return 3;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(RungeKutta23)
  ASSIGNMENT_OPERATOR(RungeKutta23)
  CLASS_INFO

  VIRTUAL_CAST(RungeKutta23)
};
/*>RungeKutta23:*/

/*Class:RungeKutta23
NAME: RungeKutta23 - numerical solution of ordinary differential equations
      by a third order Runge Kutta method.
SYNTAX: @RungeKutta23

KEYWORDS:  solver, O( `\(` h^3 `\)` ) , IVP

DESCRIPTION:

   The class implements a third order Runge Kutta method.
   An error estimate is computed based on the difference of a third order
   Runge-Kutta method and a second order scheme, the second order scheme is
   computed using another combination of the function evaluations.

CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
    initializes the class.


MEMBER FUNCTIONS:

 $\begin{description}
  `\item[`"advance"`]` -
        integrate one time step.
  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:

   class "Euler1" and class "RungeKuttaF45"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge
End: */



/*<RungeKutta4:*/
class RungeKutta4 : public ExplicitODEsolver
{
  protected:
  real dt2, dt6;
Handle(DynamicState) s0;

  Vec(real)    k2;
  Vec(real)    k3;
  Vec(real)    k4;


  public:
  RungeKutta4() {}
  RungeKutta4(ODEsolver_prm& pm);

  ~RungeKutta4() {}

  virtual void newIntegrate();

  virtual void advance (real dt);

  int getOrder() const {return 4;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(RungeKutta4)
  ASSIGNMENT_OPERATOR(RungeKutta4)
  CLASS_INFO

  VIRTUAL_CAST(RungeKutta4)
};
/*>RungeKutta4:*/

/*Class:RungeKutta4
NAME: RungeKutta4 - numerical solution of ordinary differential equations
      by a fourth order Runge Kutta method.
SYNTAX: @RungeKutta4

KEYWORDS:  solver, O( `\(` h^4 `\)` ) , IVP

DESCRIPTION:

   The class implements the method known as 'THE RUNGE KUTTA METHOD', which is
   the classical fourth order Runge Kutta method.
   The class only implements the "advace" function and inherits the
   "advanceE" which is expensive. If an adaptiveLoop
   functionality is used the user should instead consult the classes containing
   an implementation of the error estimate (embedded methods).


CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
    initializes the class.


MEMBER FUNCTIONS:
  $\begin{description}
  `\item[`"advance"`]` -
        integrate one time step.
  $\end{description}


EXAMPLES:
   See class "Oscillator"

SEE ALSO:

   class "RungeKutta23" and class "RungeKuttaDP45"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge
End: */








/*<RungeKuttaF45:*/
class RungeKuttaF45 : public ExplicitODEsolver
{
  protected:
Handle(DynamicState) s0;

  Vec(real) k2;
  Vec(real) k3;
  Vec(real) k4;
  Vec(real) k5;
  Vec(real) k6;


  public:

   RungeKuttaF45() {}
   RungeKuttaF45(ODEsolver_prm& pm);

  ~RungeKuttaF45() {}

  virtual void newIntegrate();

  virtual void advance  ( real dt);
  virtual real advanceE( real dt);

  int getOrder() const {return 5;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(RungeKuttaF45);
  ASSIGNMENT_OPERATOR(RungeKuttaF45);
  CLASS_INFO
};
/*>RungeKuttaF45:*/

/*Class:RungeKuttaF45
NAME: RungeKutta45 - numerical solution of ordinary differential equations
      by a fifth order Runge Kutta method.

SYNTAX: @RungeKuttaF45

KEYWORDS:  solver, O( `\(` h^5 `\)` ) , IVP

DESCRIPTION:

   The class implements the method known as a Runge-Kutta-Fehlberg method.
   An error estimate is computed based on the difference of a fourth order
   Runge-Kutta method and a fifth order scheme, the fourth order scheme is
   computed using another combination of the function evaluations of
   the fifth order scheme. Similar embedded Runga-Kutta
   methods  are available by the class "RungeKuttaCK45" and
   "RungeKuttaDP45".


CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
    initializes the class.


MEMBER FUNCTIONS:

 $\begin{description}
 `\item[`"advance"`]` -
       integrate one time step.

  `\item[`"advanceE"`]` -
        integrate one time step and return with an estimate of the
        local error.

  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:

   class "Euler1" and class "RungeKutta4"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge
End: */



/*<RungeKuttaCK45:*/
class RungeKuttaCK45 : public ExplicitODEsolver
{
  protected:
Handle(DynamicState) s0;

  Vec(real) k2;
  Vec(real) k3;
  Vec(real) k4;
  Vec(real) k5;
  Vec(real) k6;

  public:

   RungeKuttaCK45() {}
   RungeKuttaCK45(ODEsolver_prm& pm);

  ~RungeKuttaCK45() {}

  virtual void newIntegrate();

  virtual void advance  ( real dt);
  virtual real advanceE( real dt);

  int getOrder() const {return 5;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(RungeKuttaCK45);
  ASSIGNMENT_OPERATOR(RungeKuttaCK45);
  CLASS_INFO
};
/*>RungeKuttaCK45:*/

/*Class:RungeKuttaCK45
NAME: RungeKuttaCK45 - numerical solution of ordinary differential equations
      by a fifth order Runge Kutta method.

SYNTAX: @RungeKuttaCK45

KEYWORDS:  solver, O(`\(`h^5`\)`) , IVP

DESCRIPTION:

   The class implements the method known as a fifth ordered Runge-Kutta method,
   with the Cash and Carp coefficients.
   An error estimate is computed based on the difference of a fourth order
   Runge-Kutta method and a fifth order scheme, the fourth order scheme is
   computed using another combination of the function evaluations of
   the fifth order scheme. Similar embedded Runga-Kutta
   methods are available by the class "RungeKuttaDP45" and
   "RungeKuttaF45"

CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
     initializes the class.


MEMBER FUNCTIONS:
  $\begin{description}

     `\item[`"advance"`]` -
     integrate one time step.

     `\item[`"advanceE"`]` -
       integrate one time step and return with an estimate of the
       local error.


  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:

   class "RungeKuttaDP45" and class "RungeKutta4"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge
End: */




/*<RungeKuttaDP45:*/
class RungeKuttaDP45 : public ExplicitODEsolver
{
  protected:
  Handle(DynamicState) s0;

  Vec(real) k2;
  Vec(real) k3;
  Vec(real) k4;
  Vec(real) k5;
  Vec(real) k6;
  Vec(real) k7;


  public:

   RungeKuttaDP45() {}
   RungeKuttaDP45(ODEsolver_prm& pm);

  ~RungeKuttaDP45() {}

  virtual void newIntegrate();

  virtual void advance  ( real dt);
  virtual real advanceE( real dt);

  // actually a contionous extension (extrapolation)
  virtual void interpolate(real teta,
			   DynamicState& s,
			   bool update_ydot=false);

  int getOrder() const {return 5;}
  int getNoOfStepsMethod() const { return 1; }

  COPY_CONSTRUCTOR(RungeKuttaDP45);
  ASSIGNMENT_OPERATOR(RungeKuttaDP45);
  CLASS_INFO
};
/*>RungeKuttaDP45:*/

/*Class:RungeKuttaDP45
NAME: RungeKuttaDP45 - numerical solution of ordinary differential equations
      by a fifth order Runge Kutta method.

SYNTAX: @RungeKuttaDP45

KEYWORDS:  solver, O(`\(`h^5`\)`) , IVP

DESCRIPTION:

   The class implements the method known as a fifth ordered Runge-Kutta method,
   with the Dormand and Prince coefficients.
   An error estimate is computed based on the difference of a fourth order
   Runge-Kutta method and a fifth order scheme, the fourth order scheme is
   computed using another combination of the function evaluations of
   the fifth order scheme. A similar embedded Runga-Kutta
   method is available by the class "RungeKuttaCK45" and "RungeKuttaF45".



CONSTRUCTORS AND INITIALIZATION:

    The empty constructor requires the attach member function to
    be called to  initialize the class correctly.
    The constructor taking a ODEsolver_prm as argument complitly
    initialize the class.


MEMBER FUNCTIONS:
  $\begin{description}

     `\item[`"advance"`]` -
     integrate one time step.

     `\item[`"advanceE"`]` -
       integrate one time step and return with an estimate of the
       local error.


  $\end{description}

EXAMPLES:
   See class "Oscillator"

SEE ALSO:

   class "Euler1" and class "RungeKutta4"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge
End: */



#endif
