#ifndef ExplicitODEeqUDC_h_IS_INCLUDED
#define ExplicitODEeqUDC_h_IS_INCLUDED

#include <DynSys.h>
#include <Vec_real.h>

/*<ExplicitODEeqUDC:*/
class ExplicitODEeqUDC:  public virtual DynSys
{

  public:

  virtual void evalStateDot(Vec(real)& ydot, const DynamicState& state) = 0;

  CLASS_INFO
  VIRTUAL_CAST(ExplicitODEeqUDC)

};

/*>ExplicitODEeqUDC:*/

/*Class:ExplicitODEeqUDC

NAME: ExplicitODEeqUDC - specification of an explicit system of
      ordinary differential equations, base class

SYNTAX: @ExplicitODEeqUDC

KEYWORDS:
   explicit-ode-system, base class, IVP

DESCRIPTION:

   The class contains a pure virtual function "evalStateDot" which shall
   define an explicit system of ordinary differential equations,
   by evaluating the derivative ydot of the system, given the dynamic state.
   This function and the functions "evalSize", "evalNoDisrite" and
   "setIV" inherited from the class "DynSys" must be
   implemented by a derived problem class.
   The class acts as input to the "ODEsolver" hierarchy.

CONSTRUCTORS AND INITIALIZATION:
   The only constructor is the empty constructor, since the class contains
   pure virtual functions the class can not be constructed by its own.

MEMBER FUNCTIONS:

  Also consult the documentation of the class "DynSys".

  $\begin{description}

    `\item[`"evalStateDot"`]` -
     defines the system of equations by evaluating the derivatives.



  $\end{description}


EXAMPLES:

| //  y''     =  2 - y*y /sqrt(sin(x)) , 1 <= x <= 2
| //  y (1)   =  sin(1)*sin(1)
| //  y'(1)   =  2sin(1)*cos(1)
|
| class testEq1: public ExplicitODEeqUDC, public ODE, public SimCase
| {
|  public:
|
|  testEq1 ()          {}
|  virtual ~testEq1 () {}
|
|  virtual void evalStateDot (Vec(real)& ydot, const DynamicState& s)
|  {
|    ydot(1) = s.y(2);
|    ydot(2) = 2.0 - (4.0*s.y(1)*s.y(1))/(sin(s.time())*sin(s.time()));
|  }
|
|  virtual int evalSize() const { return 2;}
|  virtual int evalNoDiscrite() const { return 0;}
|
|  virtual void setIV(DynamicState& s)
|  {
|    s.y(1)   = sin(1.0)*sin(1.0);
|    s.y(2)   = 2.0*sin(1.0)*cos(1.0);
|    s.time() = 1.0;
|  }
|
|  static void defineStatic (MenuSystem& menu, int level = MAIN);
|  {
|    menu.addItem(level,"solver", "OdeSolver" ,
|               "class name in OdeSolver hirarchy",
|               "RungeKuttaF45",
|               validationString(hierODEsolver())
|               );
|  }
|
|  virtual void define (MenuSystem& menu, int level = MAIN)
|  { defineStatic(menu,level); }
|
|  virtual void scan (MenuSystem& menu)
|  {
|    attachTimeStepper(*new TimeStepper);
|    tip->init(1,2,0.1);
|
|    //  allocate solver:
|    ODEsolver_prm solv_prm;
|    solv_prm.exp_sys.rebind(this); // explicit problem formulation
|    solv_prm.solver_tp = menu.get("solver");
|    attachSolver(*solv_prm.create());
|  }
|
|  virtual void saveODEresults() { }
|
|  virtual void solveProblem();
|  { timeLoopA(); }
|
|  virtual void resultReport() { }
|
|  COPY_CONSTRUCTOR(testEq1)
|  ASSIGNMENT_OPERATOR(testEq1)
|  CLASS_INFO
|
|};


SEE ALSO:

   class "ODEsolver" , class "ExplictODEsolver" , class "ODE"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */

#define ClassType ExplicitODEeqUDC
#include <Handle.h>
#undef ClassType



#endif
