
#ifndef ExplicitODEsolver_h_IS_INCLUDED
#define ExplicitODEsolver_h_IS_INCLUDED

#include <default_op.h>
#include <ODEsolver.h>
#include <ODEsolver_prm.h>
#include <ExplicitODEeqUDC.h>
#include <Vec_real.h>
#include <TimeStepper.h>


/*<ExplicitODEsolver:*/
class ExplicitODEsolver: public ODEsolver
{
  protected:

  Handle(ExplicitODEeqUDC)      problem;  // problem formulation
  DynamicState           internal_state;  // internal use
  int                              work;  // no. of derivative calculations,

  inline void problemEvalStateDot(Vec(real)& ydot,const DynamicState& state)
  { problem->evalStateDot(ydot,state); work++; }

  inline void problemEvalStateDot(const DynamicState& state)
  { problem->evalStateDot((Vec(real)&)(state.ydotvec()), state); work++; }


  public:

  ExplicitODEsolver() {}
  ExplicitODEsolver(DynSys& sys)  { attach(sys);}
 ~ExplicitODEsolver() {}

  virtual void attach(DynSys& problem);

  // initialize for integration
  virtual void newIntegrate();

  // use current state as IV and restart.
  virtual void restartIntegrate();

  // flag that next step is a retry
  virtual void  retryNextStep();

  // reset the work count
  virtual void resetWork()  { work = 0; }

  // how many function calls of problem->evalStateDot
  virtual int getWork () const { return work; }

  // dense output
  virtual void interpolate(real teta, DynamicState& s,
			   bool update_ydot=false);

  // how many previous steps does the method use
  virtual int getNoOfStepsMethod() const = 0;

  // using dt and 2*dt/2 steps to compute the error (expensive)
  virtual real advanceE(real dt);

  // advance must be implemented in a derived class

  CLASS_INFO
  VIRTUAL_CAST(ExplicitODEsolver)
};
/*>ExplicitODEsolver:*/

/*Class:ExplicitODEsolver

NAME: ExplicitODEsolver - base class for explicit ODE-solvers.

SYNTAX: @ExplicitODEsolver

KEYWORDS: ODE-solvers, IVP

DESCRIPTION:

   The class implements an interface for explicit ODE-solvers. The class
   is derived from the class "ODEsolver",
   specifying an interface for all solvers in the class hirarchy.

CONSTRUCTORS AND INITIALIZATION:

   The class is an abstract base class and can not be constructed
   by its own, consult the derived classes implementing an actual method.
   The member function "attach" can be uses to initialize the class properly.


MEMBER FUNCTIONS:

  $\begin{description}

  `\item[`"advance"`]` -
        integrate on time step, must be implemented by a derived class.

  `\item[`"advanceE"`]` -
         integrate one time step and return with an error estimate.
         The default version uses step size dt/2 and compare the result
         with the solution using step size dt. The differneces are
         scaled by
         $$ 1/ (2^n - 1)$
         where n is the order of the method.
         Derived methods will contain much more efficient calcultaions of
         an error estimate. The final estimate is computed by
         a call of the problem class member function "evalErrorEstimate".

   `\item[`"interpolate"`]` -
        dense output of the last integration interval   0 <= teta <= 1
        please note that the default is a third order cubic
        interpolation and is thus not very accurate.
        Please note that if teta is larger than one or less then zero the
        upper or lower part of the integration interval is returned.

  $\end{description}

EXAMPLES:

  See class "Oscillator"

SEE ALSO:

  class "RungeKutteF45" ,class "RungeKutteDP45" ,class "RungeKutteCK45" ,
  class "RungeKutte4", class "RungeKutte23", class "Euler2" ,class "Euler1"
  class "AdamsMoulton4".

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge


End: */

#define ClassType ExplicitODEsolver
#include <Handle.h>
#undef ClassType

#endif
