#include <ODE.h>

INIT_CLASS_INFO(ODE)


//----------------------------------------------------------------------------
bool ODE::ok ()
//----------------------------------------------------------------------------
{ return getbool(tip.ok() &&  solver.ok() && problem.ok()); }


//----------------------------------------------------------------------------
void ODE::attachTimeStepper (TimeStepper& tp)
//----------------------------------------------------------------------------
{ tip.rebind(tp); }



//----------------------------------------------------------------------------
void ODE::attachRootFinder (RootFinder& rootf)
//----------------------------------------------------------------------------
{ rootfinder.rebind(rootf); }



//----------------------------------------------------------------------------
void ODE::attachSolver (ODEsolver& s)
//----------------------------------------------------------------------------
{
  solver.rebind(s); problem.rebind(s.getProblem());
  problem->state().makeItSimilar(plotstate);
}




//----------------------------------------------------------------------------
void ODE::defineStatic (MenuSystem& menu, int level)
//----------------------------------------------------------------------------
{

  MenuSystem:: makeSubMenuHeader (menu, "ode", "ode", level, 'G');

  TimeStepper::defineStatic(menu,level+1);

  menu.addItem(level,"tolerance", "tol",
	       "error tolerance parameter",
	       "10e-06",
	       "S");

  menu.addItem(level, "plot spacing", "plotdt",
	       "spacing between plot values ",
	       "0.1",
	       "R");


}


//-----------------------------------------------------------------------------
void ODE::scan(MenuSystem& menu)
//-----------------------------------------------------------------------------
{
  attachTimeStepper(*(new TimeStepper));
  tip->scan(menu);

  plot_dt = menu.get("plot spacing").getReal();
  plot_time = tip->getCurrentTime();

  real tol = menu.get("tolerance").getReal();
  setTolerance(tol);

  String filename(casename);
  filename += ".dat";
  pfile.rebind(filename.chars(),NEWFILE);

}



//---------------------------------------------------------------------------
void ODE::saveODEresults()
//---------------------------------------------------------------------------
{
  if (plot_dt>0)
    {
      real low  = tip->getCurrentTime()-tip->getCurrentDelta();
      real high = tip->getCurrentTime();
      real teta;

      // initial value
      if (plot_time == tip->getLowerTime())
	{ pfile << plot_time << " " << solver->currentState().yvec() << "\n";
	plot_time += plot_dt;
	}
      else
	while (low <= plot_time && plot_time <= high)
	  {
	    teta = (plot_time - low)/(tip->getCurrentDelta());

	    solver->interpolate(teta,plotstate());
	    pfile  << plotstate->time() << " " << plotstate->yvec() << "\n";
	    plot_time += plot_dt;
	  }
    }

}



//----------------------------------------------------------------------------
void ODE::constraintSwitches(VecSimple(int)& index, const Vec(real)& g)
//----------------------------------------------------------------------------
{
  int i,j;
  int n = g.size();
  j = 0;
  if (n)
    {
      for (i=1 ; i <= n ; i++)
	if (g(i)<0) j++;

      if (j)
	{
	  index.redim(j);
	  j=0;
	  for (i=1 ; i <= n ; i++)
	    if (g(i)<0) {j++; index(j) = j;}
	}
    }
}



//----------------------------------------------------------------------------
bool ODE::checkConstraints(DynamicState& state,
			      Vec(real)& g)
//----------------------------------------------------------------------------
{
  int i;
  int n = problem->evalNoConstraints();
  bool retv = false;
  if (n)
    {
      problem->evalConstraints(g,state);
      for (i=1 ; i <= n ; i++)
	if (g(i)<0) { retv=true; break; }
    }

  return retv;
}


//----------------------------------------------------------------------------
void ODE::timeLoopS()
//----------------------------------------------------------------------------
{
  if (!ok())
    errorFP("ODE::timeLoopS","not ok !\n");


  real dt;

  if (tip->getCurrentTime() != solver-> currentState().time())
    errorFP("ODE::timeLoopS","state time and time stepper have\n"
	    "different initial time !");


  saveODEresults();

  while (tip->timeLeft())
    {
      dt = tip->getNewDelta();
      solver->advance(dt);
      tip->increaseTime();
      saveODEresults();
    }
}



//----------------------------------------------------------------------------
void ODE::timeLoopE()
//----------------------------------------------------------------------------
{
  if (!ok())
    errorFP("ODE::timeLoopE","not ok !\n");

  real dt;

  if (tip->getCurrentTime() != solver-> currentState().time())
    errorFP("ODE::timeLoopE","state time and time stepper have\n"
	    "different initial time !");

  saveODEresults();

  while (tip->timeLeft())
    {
      dt = tip->getNewDelta();
      solver->advanceE(dt);
      tip->increaseTime();
      saveODEresults();
    }
}




//----------------------------------------------------------------------------
void ODE::timeLoopA()
//----------------------------------------------------------------------------
{
  if (!ok())
    errorFP("ODE::timeLoopA","not ok !\n");


  real dt;
  accumerr = 0;

  if (tip->getCurrentTime() != solver->currentState().time())
    errorFP("ODE::timeLoopA","state time and time stepper have\n"
	    "different initial time !");


  saveODEresults();  // ???
  

  real t, y, slope, prev_slope = -100;  // skal hentes utenfra...
  real prev_t=0.0;
  real prev_y = solver->currentState().y(1);

    while (tip->timeLeft())
    {
      dt = tip->getNewDelta();
      err = solver->advanceE(dt);

      
      // cout << "Met: " << solver->ODEmethod() << " Err: " << err <<  " dt: " << dt << endl;
//      cout << "dt: " << dt << endl;

      if (err > tolerance*dt && ! tip->StepSizeUnderFlow())
	{
	  tip->modifyDelta(0.5*dt);
	  solver->retryNextStep();  // flag to the solver that we do
				    // retry of the last step
//	  cout << "Reducing dt to " << 0.5*dt << "\n";
//	  cout << "Local time is " << tip->getCurrentTime() << "\n";
	}
      else
	 {
/*
	   #ifdef SAFETY_CHECKS
	   if (err > tolerance*dt)
	       warningFP("ODE::timeLoopA",
		      "the tolerance criteria is not satisfied, "
		      "because the minimum step size is\n"
		      "reached ! You can recompile in optimized version and "
		      "this warning will \n"
                      "disapear, but the result might be inaccurate.\n"
		      "Consider to use a method of higher order ?\n");

	   #endif
	   */
	   tip->increaseTime();     // tip and state now have same time           	   accumerr += err;         // log the accumelated error estimate

	   saveODEresults();	// ???

	   // change next time step ?
	   if ( err < (0.8*dt*tolerance)/pow(2.0, real(solver->getOrder())))
	     {
	       tip->modifyDelta(2.0*dt);
//	       cout << "Increasing dt \n";
//	       cout << "Local time is " << tip->getCurrentTime() << "\n";
	     }

	   t = prev_t + dt;
	   y = solver->currentState().y(1);
	   slope = (y-prev_y)/(t-prev_t);
//           cout << "slope: " << slope;
	   if (slope < prev_slope) 
	     {
	       // cout << "peaked!\n";
             }
	   prev_t = t;
           prev_y = y;
           prev_slope = slope;
	 }

    }

}

#include<Ptv_real.h> 
//----------------------------------------------------------------------------
void ODE::timeLoopA_aux( Ptv(real)& auxx ) {
//----------------------------------------------------------------------------

  if (!ok())
    errorFP("ODE::timeLoopA","not ok !\n");

  real dt;

//  if (tip->getCurrentTime() != solver->currentState().time())
//    errorFP("ODE::timeLoopA","state time and time stepper have\n"
//	    "different initial time !");

  saveODEresults();  

  real prev_slope = auxx(2);  
  real new_t, new_y, slope;
  real prev_t=0.0;
  real prev_y = solver->currentState().y(1);



  accumerr = 0;
    while (tip->timeLeft())
    {
      dt = tip->getNewDelta();
      err = solver->advanceE(dt);

      
// cout << "Met: " << solver->ODEmethod() << " Err: " << err <<  " dt: " << dt << endl;
//      cout << "dt: " << dt << endl;

      if (err > tolerance*dt && ! tip->StepSizeUnderFlow())
	{
	  tip->modifyDelta(0.5*dt);
	  solver->retryNextStep();  // flag to the solver that we do
				    // retry of the last step
//	  cout << "Reducing dt to " << 0.5*dt << "\n";
//	  cout << "Local time is " << tip->getCurrentTime() << "\n";
	}
      else
	 {
/*
	   #ifdef SAFETY_CHECKS
	   if (err > tolerance*dt)
	       warningFP("ODE::timeLoopA",
		      "the tolerance criteria is not satisfied, "
		      "because the minimum step size is\n"
		      "reached ! You can recompile in optimized version and "
		      "this warning will \n"
                      "disapear, but the result might be inaccurate.\n"
		      "Consider to use a method of higher order ?\n");

	   #endif
	   */
	   tip->increaseTime();     // tip and state now have same time           	   accumerr += err;         // log the accumelated error estimate

	   saveODEresults();	// ???

	   // change next time step ?
	   if ( err < (0.8*dt*tolerance)/pow(2.0, real(solver->getOrder())))
	     {
	       tip->modifyDelta(2.0*dt);
//	       cout << "Increasing dt \n";
//	       cout << "Local time is " << tip->getCurrentTime() << "\n";
	     }

	   new_t = prev_t + dt;
	   new_y = solver->currentState().y(1);
	   slope = (new_y-prev_y)/(new_t-prev_t);
//           cout << "slope: " << slope;
	   if ((slope < prev_slope)&&(auxx(1)==0)) 
	     {
               auxx(1) = prev_t; 
//	       cout << "peaked!\n";
             }
	   prev_t = new_t;
           prev_y = new_y;
           prev_slope = slope;
           auxx(2) = prev_slope;
	 }

    }
}


//----------------------------------------------------------------------------
void ODE::timeLoopA_LR( Ptv(real)& auxx ) {
//----------------------------------------------------------------------------

  if (!ok())
    errorFP("ODE::timeLoopA","not ok !\n");

  real dt;

//  if (tip->getCurrentTime() != solver->currentState().time())
//    errorFP("ODE::timeLoopA","state time and time stepper have\n"
//	    "different initial time !");

  saveODEresults();  

  problem->setParameters(auxx); // updates the global time

  real prev_slope = auxx(2);  
  real new_t, new_y, slope;
  real prev_t=0.0; real prev_cai = solver->currentState().y(4);
  real prev_y = solver->currentState().y(1);


  accumerr = 0; tip->modifyDelta(0.125);
    while (tip->timeLeft())
    {
      dt = tip->getNewDelta(); 
      err = solver->advanceE(dt);

      
// cout << "Met: " << solver->ODEmethod() << " Err: " << err <<  " dt: " << dt << endl;
//      cout << "dt: " << dt << endl;

      if (err > tolerance*dt && ! tip->StepSizeUnderFlow())
	{
	  tip->modifyDelta(0.5*dt);
	  solver->retryNextStep();  // flag to the solver that we do
				    // retry of the last step
//	  cout << "Reducing dt to " << 0.5*dt << "\n";
//	  cout << "Local time is " << tip->getCurrentTime() << "\n";
	}
      else
	 {
/*
	   #ifdef SAFETY_CHECKS
	   if (err > tolerance*dt)
	       warningFP("ODE::timeLoopA",
		      "the tolerance criteria is not satisfied, "
		      "because the minimum step size is\n"
		      "reached ! You can recompile in optimized version and "
		      "this warning will \n"
                      "disapear, but the result might be inaccurate.\n"
		      "Consider to use a method of higher order ?\n");

	   #endif
	   */
	   tip->increaseTime();     // tip and state now have same time           	   accumerr += err;         // log the accumelated error estimate

	   saveODEresults();	// ???

	   // change next time step ?
	   if ( err < (0.8*dt*tolerance)/pow(2.0, real(solver->getOrder())))
	     {
	       tip->modifyDelta(2.0*dt);
//	       cout << "Increasing dt \n";
//	       cout << "Local time is " << tip->getCurrentTime() << "\n";
	     }

	   new_t = prev_t + dt;
	   new_y = solver->currentState().y(1);
	   slope = (new_y-prev_y)/(new_t-prev_t);
//           cout << "slope: " << slope; cout << " new_y: " << new_y;
//           cout << " global time passed: " << auxx(4) + new_t << endl;
	   if (((slope < prev_slope)&&(auxx(1)==-999.9)&&(slope>0.0)&&(new_y>-40.0))||((auxx(1)==-999.9)&&(new_y>40.0))) 
	     { // siste del ndlsning, 40 kan vre for hyt for FHN
               auxx(1) = prev_t + auxx(4); 
               auxx(3) = prev_cai; // rember calsium at peak
//	       cout << "peaked! slope: " << prev_slope; 
//	       cout << " y: " << prev_y;
//	       cout << " t: " << auxx(1) << endl;
	       problem->setParameters(auxx);
             }

	   if ((auxx(1)!=-999.9)&&((auxx(4)+new_t)>(auxx(1)+2.0))&&(auxx(5)==-999.9)) 
	     {
               auxx(5) = (solver->currentState().y(4) - auxx(3))*(2.0/(auxx(4)+new_t-auxx(1)));
	       cout << "accumulation completed: " << auxx(5) << endl;
               problem->setParameters(auxx);  // undvendig her?
             }

	   prev_t = new_t;
           prev_y = new_y;
           prev_slope = slope;
           prev_cai = solver->currentState().y(4);
           auxx(2) = prev_slope;
	 }

    }
}





//----------------------------------------------------------------------------
real ODE::advanceD(real dt)
//----------------------------------------------------------------------------
{

  real minswitchtime;
  real ts = solver-> currentState().time();

  err = solver->advanceE(dt);

  if (checkConstraints(solver->currentState(),g))
    {
       int i;
       constraintSwitches(index_switch, g);

       int no_of_switches = index_switch.size();
       if (no_of_switches == 1)
	 {
	   minswitchtime = rootfinder->rootfind(solver(),1);
	 }
       else
	 {
	   timeforswitch.redim(no_of_switches);

	   // loop over all switch times
           // and get the next switch time

           // set minswitchtime larger than the root
	   minswitchtime = ts+dt;
	   for (i=1; i<= no_of_switches;i++)
	     {
	       timeforswitch(i) = rootfinder->rootfind(solver(),
						       index_switch(i));
	       if (minswitchtime > timeforswitch(i))
		 minswitchtime = timeforswitch(i);
	     }

	   // check for several constraints at same time
	   switchesatsametime.reset();
	   for (i=1; i<= no_of_switches;i++)
	     {
	       if (timeforswitch(i)== minswitchtime)
		 switchesatsametime.add(i);
	     }
	   switchesatsametime.convert2vector(index_switch);
	 }

       real newdt = minswitchtime - ts;

       tip->modifyDelta(newdt);
       solver->retryNextStep();

       // advance to border of constraint
       err = solver->advanceE(newdt);

       // loop over switches at same time:
       for (i=1; i <= index_switch.size(); i++)
	 problem->handleRoot(solver->currentState(), index_switch(i));

       solver->restartIntegrate();
    }

  return err;
}


// adaptive loop for constrained system
//----------------------------------------------------------------------------
void ODE::timeLoopDA()
//----------------------------------------------------------------------------
{
  if (! (ok() && rootfinder.ok()))
    errorFP("ODE::timeLoopDA","not ok !\n");


  if (tip->getCurrentTime() != solver->currentState().time())
    errorFP("ODE::timeLoopDA","state time and time stepper have\n"
	    "different initial time !");

  g.redim(problem->evalNoConstraints());

  real dt; //, newdt;
  accumerr = 0;

  saveODEresults();


  while (tip->timeLeft())
    {
      dt = tip->getNewDelta();
      err = advanceD(dt);

      if (err > tolerance*dt && ! tip->StepSizeUnderFlow())
	{
	  tip->modifyDelta(0.5*dt);
	  solver->retryNextStep();  // flag to the solver that we do
				    // retry of the last step
	}
      else
	 {
          #ifdef SAFETY_CHECKS
	   if (err > tolerance*dt)
	     warningFP("ODE::timeLoopDA",
		       "the tolerance criteria is not satisfied, "
	               "because the minimum step size is\n"
                       " reached ! You can recompile in optimized version and "
	               "this warning will \n"
                       "disapear, but the result might be inaccurate.\n"
		       "Consider to use a method of higher order ?\n");
	   #endif
	   tip->increaseTime();     // tip and state now have same time
	   accumerr += err;         // log the accumelated error estimate

	   saveODEresults();

	   // change next time step ?
	   if ( err < (0.8*dt*tolerance)/pow(2.0, real(solver->getOrder())))
	     {
	       tip->modifyDelta(2.0*dt);
	     }
	 }

    }

}
