#ifndef ODE_h_IS_INCLUDED
#define ODE_h_IS_INCLUDED

#include <HandleId.h>
#include <btTypeInfo.h>   // btbtTypeInfo.h iflg namechanges2.7
#include <Dpmacros.h>
#include <DynamicState.h>
#include <TimeStepper.h>
#include <ODEsolver.h>
#include <RootFinder.h>
#include <SetOfNo_int.h>
#include <SimCase.h>

/*<ODE:*/
class ODE : public virtual HandleId
{
  private:

  Vec(real)                       g;
  VecSimple(int)       index_switch;
  VecSimple(real)     timeforswitch;
  SetOfNo(int)   switchesatsametime;


  protected:

  real                tolerance;  // tolerance compared to error per unit step
  real                 accumerr;  // the accumelated error estimate
  real                      err;  // computed error estimate in adaptive loop
  Handle(TimeStepper)       tip;  // time stepper
  Handle(RootFinder) rootfinder;  // the rootfinder for constrained systems
  Handle(ODEsolver)      solver;  // the solver
  Handle(DynSys)        problem;  // the problem
  // make a seperate class ODEstore ?
  Os                      pfile;  // store to file
  real                  plot_dt;  // plot spacing
  real                plot_time;  // the next plot time
  Handle(DynamicState) plotstate;  //  state for plotting

  real advanceD(real dt);
  void constraintSwitches(VecSimple(int)& index, const Vec(real)& result);


  public:
           ODE() { tolerance = 10e-06; }
  virtual ~ODE() {}

  // necessary attachment
  void attachTimeStepper (TimeStepper& tip);
  void attachSolver      (ODEsolver&   solver);

  void attachRootFinder  (RootFinder&  rootfinder);

  bool checkConstraints(DynamicState& state, Vec(real)& result);
  virtual void setTolerance(real tol) { tolerance = tol; }

  virtual void saveODEresults();

  virtual void timeLoopS  ();
  virtual void timeLoopE  ();
  virtual void timeLoopA  ();
  virtual void timeLoopA_aux  ( Ptv(real)& auxx );  
  virtual void timeLoopA_LR  ( Ptv(real)& auxx );
  virtual void timeLoopDA ();

  static void defineStatic (MenuSystem& menu, int level = MAIN);

  virtual void define (MenuSystem& menu, int level = MAIN)
  { defineStatic(menu,level); }

  virtual bool ok();

  virtual void scan (MenuSystem& menu);
  CLASS_INFO


};
/*>ODE:*/

/*Class:ODE
NAME: ODE - solving ordinary differetial equations.

SYNTAX: @ODE

KEYWORDS:
   ODE, base class, IVP

DESCRIPTION:

   The class contains various virtual functions for computing the
   solution of ordinary differential equations.

CONSTRUCTORS AND INITIALIZATION:

  The class only contains an empty constructor, further initialization
  is necessary and is accomplished by the "attachSolver" member function.
  If a constrained system is solved a root finder must be attached by the
  "attachRootFinder" member functon.

MEMBER FUNCTIONS:

  $\begin{description}


  `\item[`"timeLoopS"`]` -
        timeloop where the time stepper have full control of the
        the step sizes, the error estimate is not computed.

  `\item[`"timeLoopE"`]` -
        timeloop where the time stepper have full control of the
	the step sizes, the error estimate is computed and can hence be
	saved eg. for later visualization.

  `\item[`"timeLoopA"`]` - implements a simple adaptive
        loop. The first argument is the solution vector, the second
        a "TimeStepper" containing time parameters.
        The computed error estimate per step must be less than the given
        tolerance (error per unit step).
        If larger the step size is halfed, if the error per step is less than
        $$ 0.8*tolerance/ 2.0^{(n)}$
	the step size is doubled (n is the global order of the method).
        Please note that the computation
        of the error estimate can be controlled by the user by reimplementing
	the virtual functiom evalErrorEstimate of the class "DynSys".
        Ofcourse also this member function can be fully  reimplemented by
        a client class.


   `\item[`"timeLoopDA"`]` - implements a similar adaptive loop as
         "timeLoopA" but also checks for constraints. The member functions
	 "evalNoConstraints", "evalConstraints" and "handleRoot" must
         be given an implementation in the problem class. The class must
         also be supplied with a root finder from the "RootFinder" class
         hirarchy. The time loop checks for constraints at the end of
         each integration interval. The constraint function must therfore
         signal a constraint also a period after a constraint has been
         reached. This is done by the evalConstraint function which returns
	 with a vector, if some of the components are negative a constraint
         is reached. The root of the least of these constraints are
	 located by the root finder algorithm. Therafter the state is
         advanced to the border of the constriant, the "handleRoot"
	 member function shall switch the discrete states of the dynamic
         system so that the evaluation of the derivative becomes correct in the
         next integration step and the constraint function becomes
         positive. See class "DiscontEq1".




  $\end{description}


EXAMPLES:

   See "class Oscillator"

SEE ALSO:

   class "ODEsolver"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */



#define ClassType ODE
#include <Handle.h>
#undef ClassType

#endif
