#ifndef ODEsolver_h_IS_INCLUDED
#define ODEsolver_h_IS_INCLUDED

#include <btTypeInfo.h> // btbtTypeInfo.h iflg namechanges2.7
#include <DynamicState.h>
#include <VecSimplest_Handle_DynSys.h>

class ExplicitODEsolver;

/*<ODEsolver:*/
class ODEsolver: public virtual HandleId
{
  protected:

  Handle(DynSys)       dynamic_system;  // the dynamic system
  int                    no_of_states;  // size of state vector
  int                 no_of_discrites;  // size of discrite state vector
  Handle(DynamicState)  current_state;  // the current state
  Handle(DynamicState)     last_state;  // state before the last advace
  real                        last_dt;  // last time step
  bool                 error_computed;  // is the error estimate computed
  Vec(real)                  locerror;  // estimate of local error
  int                      retrycount;  // no of retries

  public:

  ODEsolver() {}
  ODEsolver(DynSys& sys)
  { attach(sys); }

  virtual ~ODEsolver() {}

  DynamicState&     prevState ()       { return last_state();    }
  DynamicState&     currentState ()    { return current_state(); }
  real              lastDt ()          { return last_dt;  }

  const Vec(real)&  errorVec () const;
  bool              isErrorVecComputed () const { return error_computed; }
  DynSys&           getProblem()       { return dynamic_system(); }

  // attach dynamical system and initialize for integration
  virtual void attach(DynSys& sys)
  {
    dynamic_system.rebind(sys);
    newIntegrate();
  }

  // starts a new integration
  virtual void newIntegrate();

  // use current state as IV and restart.
  virtual void restartIntegrate();

   // advance
  virtual void advance  (real dt) = 0;

  // advance and compute error estimate
  virtual real advanceE (real dt) = 0;

  virtual void retryNextStep()
  { current_state() = last_state(); last_dt = 0; retrycount++;

  }

  // dense output of the last integration interval   0 <= teta <= 1
  virtual void interpolate(real teta,
			   DynamicState& s,
			   bool update_ydot = false ) = 0;


  virtual void resetWork()  = 0;
  virtual  int getOrder () const = 0; // global order of method
  virtual  int getWork  () const = 0;
  virtual  int getRetries  () const { return retrycount; }


  String   ODEmethod() { return String(getInfo()->name()); }

  virtual bool ok() { return dynamic_system.ok(); }

  CLASS_INFO
  DEF_VIRTUAL_CAST(ExplicitODEsolver)

};
/*>ODEsolver:*/

/*Class:ODEsolver
NAME: ODEsolver - base class for solvers for ordinary differetial equations.

SYNTAX: @ODEsolver

KEYWORDS:
   ODEsolver, base class, IVP

DESCRIPTION:
   The class is an abstract base class defining the interface for
   all solvers in the hirarchy. Specially the class contains a pointer
   to a dynamical system. The advance functions act on this system and
   advance the solution of the dynamical system a given time step.
   The solution of the dynamical system is given by the dynamic state
   contained by the "DynSys" class.

CONSTRUCTORS AND INITIALIZATION:

MEMBER FUNCTIONS:
  $\begin{description}

  `\item[`"attach"`]` -
         the dynamic system is attached to the solver, proper initialization
         is performed.

  `\item[`"newIntegrate"`]` -
        initialize before integration.

  `\item[`"restartIntegrate"`]` -
        use current state as initial value and restart integration.

  `\item[`"retryNextStep"`]` -
        flag to the solver that the next step is a retry.

  `\item[`"advance"`]` -
        integrate on time step, must be implemented by a derived class.

   `\item[`"advanceE"`]` -
        integrate on time step and returns with an error estimate,
	must be implemented by a derived class.

  `\item[`"errorVec"`]` -
        returns a reference to the vector containg the computed estimates
        of the local error. Note that "advanceE" must be called
        prior to the call of "errorVec". The function is useful if the
        estimated error is to be stored during a timeLoop.
        To manipulate the error vector  the "evalErrorEstimate" function of
        the problem class should be used.

  $\end{description}

EXAMPLES:

   See "class Oscillator"

SEE ALSO:

   class "ExplicitODEsolver", class "ODE"

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */



#define ClassType ODEsolver
#include <Handle.h>
#undef ClassType

#endif
