#ifndef RootFinder_h_IS_INCLUDED
#define RootFinder_h_IS_INCLUDED

#include <DynamicState.h>
#include <DynSys.h>
#include <ODEsolver.h>

class ODE;
class ODEsolver;
/*<RootFinder:*/
class RootFinder: public HandleId
{
  protected:

  int itt;                       // no of iterations
  real within;                   // tolerance parameter
  int maxitt;                    // maximum iterations

  public :

  RootFinder() { itt = 0; maxitt = 100; within = 10e-16; }

  virtual ~RootFinder() {}

  int getIterations() const { return itt; }

  virtual real rootfind(ODEsolver& solver, int index) = 0;

  CLASS_INFO

};
/*>RootFinder:*/


/*Class:RootFinder

NAME: RootFinder - root finder, abstract base class.

SYNTAX: @RootFinder

KEYWORDS: root finder, ODE, constraints

DESCRIPTION:

  The class is an abstract base class for classes implementing
  a root finding algorithm for constrained ODE-equations.
  The algorithm shall locate the time of such an switch.
  The root finder  use the interpolator
  supplied with the method to locate the root.

CONSTRUCTORS AND INITIALIZATION:

  The class must be initialized by a derived class.

MEMBER FUNCTIONS:

  $\begin{description}

  `\item[`"getIterations"`]` - returns the number of iterations used to
   compute a root.


  `\item[`"rootfind"`]` - find the root, there are two arguments.
     First the current solver, which is assumed to have attempted
     an integration step and a constriant is reached. The root of the
     constraint function is then calculated.
     Secondly the index refers to which constraint we search for as we might
     have several constraints.

 $\end{description}

EXAMPLES:

  See class "Disconteq1"

SEE ALSO:

   See class "Bisection", class "RegulaFalsi".

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */


#define ClassType RootFinder
#include <Handle.h>
#undef ClassType


/*<Bisection:*/
class Bisection : public RootFinder
{
  Handle(DynamicState) scratch;
  Vec(real) results;

  public :

  virtual real rootfind(ODEsolver& solver ,int index);

  CLASS_INFO

};
/*>Bisection:*/

/*Class:Bisection

NAME: Bisection - root finder, the bisection method

SYNTAX: @Bisection

KEYWORDS: root finder, bisection, ODE, constraints

DESCRIPTION:

  The class implements the bisection root finding algorithm.

CONSTRUCTORS AND INITIALIZATION:

  The class is initialized by the empty constructor.

MEMBER FUNCTIONS:

    Consult the class "RootFinder".

EXAMPLES:

  See class "Disconteq1"

SEE ALSO:

   See class "RegulaFalsi", class "RootFinder".

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */


/*<RegulaFalsi:*/
class RegulaFalsi : public RootFinder
{
  Handle(DynamicState) scratch;
  Vec(real) results;

  public:
  virtual real rootfind(ODEsolver& solver , int index);

  CLASS_INFO

};
/*>RegulaFalsi:*/


/*Class:RegulaFalsi

NAME: RegulaFalsi - root finder, the RegulaFalsi method

SYNTAX: @RegulaFalsi

KEYWORDS: root finder, RegulaFalsi, ODE, constraints

DESCRIPTION:
  The class implements the RegulaFalsi root finding algorithm.

CONSTRUCTORS AND INITIALIZATION:
  The class is initialized by the empty constructor.

MEMBER FUNCTIONS:
    Consult the class "RootFinder".

EXAMPLES:

  See class "Disconteq1"

SEE ALSO:

   See class "RegulaFalsi", class "RootFinder".

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */



#endif
