#ifndef TimeStepper_h_IS_INCLUDED
#define TimeStepper_h_IS_INCLUDED

#include <SimCase.h>

/* SIMPLE TIMESTEPPING */


/*<TimeStepper:*/
class TimeStepper: public virtual HandleId
{

  protected:

  real           time;   // current time
  real          delta;   // current time stepping
  real  min_step_size;   // minimum size of a step, default 10e-14
  int            step;   // no of increments
  real     lower_time;   // start of time interval
  real     upper_time;   // end of time interval
  real      prev_time;
  int underflow_count;   // no of steps with min_step_size
  public:

  TimeStepper()  { init(0,0,0); }

  TimeStepper(real lt,real ut,real idt)
  { init(lt,ut,idt); }

  void init(real lt,real ut,real it);

  inline real getCurrentTime()  const  { return time;  }
  inline int  getCurrentStep()  const  { return step;  }
  inline real getCurrentDelta() const  { return delta; }
  inline int  getNoUnderFlowSteps() const  { return underflow_count; }

  inline real getUpperTime() const { return upper_time; }
  inline real getLowerTime() const { return lower_time; }

  inline void setNewTime(real t)  { time = t; }
  inline void setNewUpperTime(real ut);
  inline void setMinStepSize(real minimum) { min_step_size = minimum; }

  inline bool StepSizeUnderFlow() const
  { return getbool (delta <= min_step_size); }

  inline real getNewDelta();

  inline void modifyDelta(real dt);

  inline  bool timeLeft() const
  { return getbool (time < upper_time); }

  inline  void increaseTime()
  {  prev_time = time;  time += getNewDelta();
     if (StepSizeUnderFlow()) underflow_count++;
     step++;
  }

  static void defineStatic (MenuSystem& menu, int level = MAIN);

  virtual void define (MenuSystem& menu, int level = MAIN)
  { defineStatic(menu,level); }

  virtual void scan (MenuSystem& menu);

  CLASS_INFO

};
/*>TimeStepper:*/

/*Class:TimeStepper

NAME: TimeStepper - time stepping, time parameters.

SYNTAX: @TimeStepper

KEYWORDS: Time stepping, ODE

DESCRIPTION:

  The class contains parameters for time stepping.

CONSTRUCTORS AND INITIALIZATION:

  The class is constructed by specifying the time interval and an
  initial time step.

MEMBER FUNCTIONS:

   Most member functions should be self-explanatory.

EXAMPLES:
  |  // example of a simple time loop, tip is of type TimeStepper
  |
  |   while (tip.timeLeft())
  |    {
  |      dt = tip.getNewDelta();
  |      solver.advance(dt);
  |      tip.increaseTime();
  |    }


SEE ALSO:

   See class "ODE".

DEVELOPED BY:

   SINTEF Applied Mathematics, Oslo, Norway.

AUTHOR:

   Johan Berge

End: */




//------------------------------------------------------------------------
inline real TimeStepper::getNewDelta()
  //------------------------------------------------------------------------
{
  if (timeLeft())
    {
      if (time+delta > upper_time)
	modifyDelta(upper_time - time);
    }
  return delta;
}



//------------------------------------------------------------------------
inline void TimeStepper::modifyDelta(real dt)
//------------------------------------------------------------------------
{
  delta = dt;
  if (delta < min_step_size)
    {
      delta = min_step_size;

    }
}



//------------------------------------------------------------------------
inline void  TimeStepper::setNewUpperTime(real ut)
//------------------------------------------------------------------------
{
  if (time > ut)
    errorFP("TimeStepper::setNewUpperTime()", "new upper time is less"
            "than current time !");

  upper_time  = ut;
}


#define ClassType TimeStepper
#include <Handle.h>
#undef ClassType


#endif
