
/*****************************************************************************/
/*                                                                           */
/* (c) Copyright 1994, 95, 96                                                */
/*     Numerical Objects AS, Oslo, Norway.                                   */
/*     All rights reserved.                                                  */
/*                                                                           */
/*     See http://www.nobjects.com/ for further details.                     */
/*                                                                           */
/*****************************************************************************/

//#include <Copyright.h>

// $Id: GridFEAdT.C,v 1.1.1.1 2003/11/25 16:35:31 glennli Exp $

/*
  Class for representation of a hierarchical finite element grid.
  Code implementation of member functions.
*/

#include <GridFEAdT.h>
#include <MatSparse_real.h>
#include <CPUclock.h>

/************ Handle class **************/

#define  ClassType GridFEAdT
#include <Handle_ClassType_C.h>             // note! version without inline functions
#undef   ClassType                // Handle_ClassType_h.h is included in GridFEAdT.h

INIT_CLASS_INFO(GridFEAdT)

//-----------------------------------------------------------------------------
GridFEAdT:: GridFEAdT ()
//-----------------------------------------------------------------------------
{
// initialize the GridFE data
  nsd = nno = nel = nne = maxnne = 0;
  elm_tp = "Dummy element";
  onemat = true;
  uniform_mesh = false;
  same_elm_tp = true;
  current_node = 0;
  associated_bfg = NULL;

  initGridFEAdT();  // initialize the GridFEAdT specific part
}

//-----------------------------------------------------------------------------
void GridFEAdT:: initGridFEAdT ()
//-----------------------------------------------------------------------------
{
  next_grid = NULL;
  neighbor_computed = new_nodes_are_set = false;
  finest_grid = coarsest_grid =           true; 
  parent_info = children_info =           false;
  regular_uniform_ref =                   false;
  refinement_method = 1;
  reg_bisect_short = 2; // choose shortest edge alternative, used in 2D
  use_node_for_equal_test = true;

  current_elm_number = current_node_number = 0;
  bound_box_computed = false;
  FET3.attach(this); // at most one of FET3 and FET4 are going to be used
  FET4.attach(this); // for the same grid
  lengthsscratch.redim(6); // global arrays used in getEdgeLengthOrder in 3D
  orderscratch.redim(6);
}

//-----------------------------------------------------------------------------
GridFEAdT:: GridFEAdT (const GridFE& grid)
//-----------------------------------------------------------------------------
{
  if (!grid.ok())
    fatalerrorFP("GridFEAdT:: GridFEAdT", "Empty grid is given.");

  if (this == &grid) 
    return;

  // purify (operator= calls redim which tests this->nsd==nsd_ etc., hence
  // these variables must be initialized!)
  nsd = nno = nel = nne = maxnne = 0;

  *this = grid;
}

//-----------------------------------------------------------------------------
GridFEAdT:: GridFEAdT (const GridFEAdT& grid)
//-----------------------------------------------------------------------------
{
  if (!grid.ok())
    fatalerrorFP("GridFEAdT:: GridFEAdT", "Empty grid is given.");

  if (this == &grid)
    return;

  nsd = nno = nel = nne = maxnne = 0;

  *this = grid; 

}

//-----------------------------------------------------------------------------
GridFEAdT:: ~GridFEAdT()
//-----------------------------------------------------------------------------
{ 
//  if (verbose >= 0)
    s_o << "Deleting a GridFEAdT object with " << nno << " nodes.\n";
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: ok () const
//-----------------------------------------------------------------------------
{
// Check the GridFE parts ...

  bool test = GridFE::ok();

// continue to check the GridFEAdT specific parts...

// consistency of the bool flags, hierarchy etc..

  return test;  // if no errors are found, this statement will be reached
}

//-----------------------------------------------------------------------------
void GridFEAdT:: operator = (const GridFE& grid)
//-----------------------------------------------------------------------------
{
  if (!grid.ok())
    fatalerrorFP("GridFEAdT:: operator =","Empty grid is given as argument.");

  String etyp = grid.getElmType(1);
  if( grid.allElmsOfSameType() == false &&
      "ElmT4n3D" != etyp && "ElmT10n3D" != etyp &&
      "ElmT3n2D" != etyp &&  "ElmT6n2D" != etyp && 
      "ElmB2n1D" != etyp &&  "ElmB3n1D" != etyp )
      fatalerrorFP("GridFEAdT:: operator =(const GridFE&)\n",
		   "Only elements ElmB2n1D, ElmB3n1D, ElmT3n2D, ElmT6n2D, "
		   "ElmT4n3D and ElmT10n3D are implemented\n"
		   "so far, this is element type %s.\n"
		   "It might be possible to use Prepro::box2triangle "
		   "to change from ElmB4n2D and ElmB8n3D elements?",
		   etyp.chars());

  GridFE::operator = (grid); // copy GridFE part

  initGridFEAdT ();  // initiate GridFEAdT specific data
}


//-----------------------------------------------------------------------------
void GridFEAdT:: operator = (const GridFEAdT& grid)
//-----------------------------------------------------------------------------
{ 
  if (!grid.ok())
    fatalerrorFP("GridFEAdT:: operator =","Empty grid is given as argument.");

  GridFE::operator = (grid);

  initGridFEAdT ();

// copy GridFEAdT specific data

  neighbor_computed = grid.neighbor_computed;
  new_nodes_are_set = grid.new_nodes_are_set;
  finest_grid = grid.finest_grid;
  coarsest_grid = grid.coarsest_grid;
  parent_info = grid.parent_info;
  children_info = grid.children_info;
  regular_uniform_ref = grid.regular_uniform_ref;
  refinement_method = grid.refinement_method;
  reg_bisect_short = grid.reg_bisect_short;
  use_node_for_equal_test = grid.use_node_for_equal_test;

  current_elm_number = grid.current_elm_number;
  current_node_number = grid.current_node_number;
  bound_box_computed = grid.bound_box_computed;

  if( bound_box_computed ){
    bounding_box.redim(nel,2*nsd);
    bounding_box = grid.bounding_box;
  }
  if( neighbor_computed ){
    elm_neigh.redim( nel, nsd+1 );
    elm_neigh = grid.elm_neigh;
    if( grid.edge_face2.size() || grid.edge_face3.size() ){
      edge_face2.redim(grid.edge_face2.size());
      edge_face3.redim(grid.edge_face3.size());
      edge_face2 = grid.edge_face2;
      edge_face3 = grid.edge_face3;
    }
  }

  if( new_nodes_are_set ){
    new_nodes.redim( nel, grid.new_nodes.columns() );
    new_nodes = grid.new_nodes;
  }
  if( children_info ){
    pos_child.redim( nel+1 );
    pos_child = grid.pos_child;
  }
  if( parent_info ){
    parent.redim( nel );
    parent = grid.parent;
  }

  if( prev_grid.ok())
    prev_grid.rebind(grid.prev_grid());

  next_grid = grid.next_grid;

}

//-----------------------------------------------------------------------------
//Handle(GridFEAdT) GridFEAdT:: makeCopyOfHierarchy( const int firstlevel, 
//						     const int lastlevel )
GridFEAdT* GridFEAdT:: makeCopyOfHierarchy( const int firstlevel, 
					      const int lastlevel )
//-----------------------------------------------------------------------------
{//DbP("GridFEAdT:: makeCopyOfHierarchy")
// Create a copy of a part of the grid hierarchy.
// firstlevel is the coarsest level in "this" which is copied
// and which will be the coarsest grid in the copy.
// lastlevel is the finest level which is copied
// and which will be the finest grid in the copy.
//
// One application of this member function is to create a sequence
// of hierarchical grids which have some common coarse grids.

  Handle(GridFEAdT) oldgrid;
  oldgrid.rebind(this);

  if( firstlevel < 1 ){
    errorFP("GridFEAdT:: makeCopyOfHierarchy\n",
	    "Requested first level %d is to coarse, must be > 0.", firstlevel);
    return oldgrid.getPtr();
  }

  if( lastlevel > getFinestGridLevelNo () ){
    errorFP("GridFEAdT:: makeCopyOfHierarchy\n",
	    "Requested last level %d is to fine, only %d levels exists.",
	    lastlevel,getFinestGridLevelNo());
    return oldgrid.getPtr();
  }

  if( firstlevel > lastlevel ){
    errorFP("GridFEAdT:: makeCopyOfHierarchy\n",
	    "First level is larger than last level.");
    return oldgrid.getPtr();
  }

  int i;
////DbP("GridFEAdT:: makeCopyOfHierarchy")
  
  oldgrid.rebind( getCoarsestGrid() );

// before the first level
  for( i=1; i<firstlevel; i++ )
    oldgrid.rebind(oldgrid().getChildGrid());
//DbP("GridFEAdT:: makeCopyOfHierarchy")

  GridFEAdT* g;
  GridFEAdT* gg;

  for( i=firstlevel; i<=lastlevel; i++ ){
//cout << i << endl;
//DbP("GridFEAdT:: makeCopyOfHierarchy")

    g = new GridFEAdT(oldgrid()); // copy one of the grids 
//DbP("GridFEAdT:: makeCopyOfHierarchy")

    if ( i == firstlevel ){
      gg = g; // connect to new grid
      gg->removeParentGrid(); // make sure this is the coarsest grid 
    }
    else {
      gg->setChildGrid( g ); // set child pointer
      g->setParentGrid( gg ); // set parent pointer
      gg = g; // connect to new grid
    } 

    if ( i == lastlevel )
      gg->removeChildGrid(); // now the finest grid in the new hierarchy 
    else
      oldgrid.rebind( oldgrid().getChildGrid() );// go one step up in hierarchy
  }
  return gg;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getGridLevelNo()
//-----------------------------------------------------------------------------
{ //recursive call
  return ( isThisCoarsestGrid() ) ? 1 : getParentGrid()->getGridLevelNo() + 1;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getFinestGridLevelNo()
//-----------------------------------------------------------------------------
{
  return getFinestGrid()->getGridLevelNo();
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: getFinestGrid()
//-----------------------------------------------------------------------------
{ //recursive call
  return ( isThisFinestGrid() ) ? this : getChildGrid()->getFinestGrid();
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: getCoarsestGrid()
//-----------------------------------------------------------------------------
{ //recursive call
  return ( isThisCoarsestGrid() ) ? this : getParentGrid()->getCoarsestGrid();
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeNeighborInfo ()
//-----------------------------------------------------------------------------
// Removal of neighbor data, i.e., elm_neigh and new_nodes
{
  elm_neigh.redim(0,0);
  new_nodes.redim(0,0);
  edge_face2.redim(0);
  edge_face3.redim(0);
  setNeighborsComputed( false );
  setNewNodesSet( false );
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeNewNodeInfo()
//-----------------------------------------------------------------------------
// Removal of new_nodes data showing the node numbering  or marking of edges
{
  new_nodes.redim(0,0);
  setNewNodesSet( false );
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeChildrenInfo()
//-----------------------------------------------------------------------------
// Removal of the information about where the children element are.
{
  pos_child.redim(0);
  setChildrenInfo( false );
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeChildGrid()
//-----------------------------------------------------------------------------
// Removal of pointer to the child grid and 
// info about where the children elements are.
{
  next_grid = NULL;

//  this->getFinestGrid()->removeParentGrid();
  
  setFinestGrid( true ); // This grid is now finest
  removeChildrenInfo();    // without child grid no need for children info
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeParentInfo()
//-----------------------------------------------------------------------------
// removal of info on where the parent elements are
{
  parent.redim(0);
  setParentInfo( false );
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeParentGrid()
//-----------------------------------------------------------------------------
// removal of parent info and prev_grid
{
  prev_grid.detach();
  setCoarsestGrid( true ); // if no parent then this grid is coarsest
  removeParentInfo(); // No need for info about parent elms if no parent grid
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeHierInfo()
//-----------------------------------------------------------------------------
// removal of children, parent, prev_ and next_grid
{
  removeChildGrid();
  removeParentGrid();
}

//-----------------------------------------------------------------------------
void GridFEAdT:: initRefine()
//-----------------------------------------------------------------------------
{
  if( "ElmT4n3D" != elm_tp && "ElmT10n3D" != elm_tp &&
      "ElmT3n2D" != elm_tp && "ElmT6n2D" != elm_tp && 
      "ElmB2n1D" != elm_tp && "ElmB3n1D" != elm_tp )
      fatalerrorFP("GridFEAdT:: initRefine\n",
		   "Only elements ElmB2n1D, ElmB3n1D, ElmT3n2D, ElmT6n2D, "
//		   " ElmT4n3D and ElmT10n3D are implemented so far,\n"
		   " and ElmT4n3D are implemented so far,\n"
		   " this is element type %s",elm_tp.chars());
  if( ! isNeighborsComputed() ) // is the element-element info computed?
    initNeighbor();            // required for numbering of new nodes
  else if (nsd==3) {//in 3D it is necessary to make sure that n2e info is comp.
    additional_info.initNeighbor(*this,true,false,false);
    NeighborFE& neighbor = additional_info.neighbor;
    if( neighbor.nodeSize() == 0 )
      neighbor.init(*this,true,false,false);
  }
  if( ! isThisFinestGrid() ) {
//    warningFP("GridFEAdT:: initRefine", 
//	      "\nA refined grid already exists, will replace it.");
    removeChildGrid();
  }
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: refine(const VecSimple(int)& mark_elms)
//-----------------------------------------------------------------------------
{//DbP("");
  initRefine();
//DbP("");
  markElements( mark_elms );
//DbP("");
  if( refine() ) 
    return getChildGrid();
  else
    errorFP("GridFEAdT:: refine", "Refinement failed.");
  return NULL;
//DbP("");
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: refine(const VecSimple(bool)& mark_elms,
				const int ref_method )
//-----------------------------------------------------------------------------
{
  initRefine();
  setRefinementMethod( ref_method );
  markElementsBool( mark_elms );
  if( refine() ) 
    return getChildGrid();
  else
    errorFP("GridFEAdT:: refine", "Refinement failed.");
  return NULL;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: refineUniformly( const int ref_method, bool regref)
//-----------------------------------------------------------------------------
// refine uniformly, if regref is true and all edges in 2D and 3D will
// be refined, then regular subdivision is chosen
{ CPUclock cpu; cpu.initTime();
  initRefine(); //cout<<"\n after initrefine = " << cpu.getInterval()<<endl;
  setRefinementMethod( ref_method );
  markAllElements(); //cout<<"\n after markall = " << cpu.getInterval()<<endl;
  if( nsd == 3 && ref_method == 6 || nsd == 2 && ref_method == 3 )
    setRegularUniformRef(regref); // bisection or regular refinement?
  if( refine() ){ // this is where the actual refinement takes place
    setRegularUniformRef(false);// reset value for next refinement
//    cout<<"\n after refine = " << cpu.getInterval()<<endl;
   return getChildGrid();
  }
  else
    errorFP("GridFEAdT:: refineUniformly", "refinement failed.");
  return NULL;
}
//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: refineLongEdges( const real break_ratio )
//-----------------------------------------------------------------------------
// Let maxlen be the longest edge in the whole grid
// refine all edges which are longer or equal than break_ratio of maxlen
// if ElmT6n2D or ElmT10n3D only the vertex nodes will be used.
{
  int e;
  VecSimple(int) markings(nel);
  if ( nsd==3 ) {
    FET4.refill(1);
    real maxlen = FET4.maxLength(); // find longest edge in grid
    for( e=2; e<=nel; e++ ) {
      FET4.refill(e);
      maxlen = max(maxlen,FET4.maxLength());
    } 
    real b2 = sqr(break_ratio*maxlen); // use square for efficiency
    for( e=1; e<=nel; e++ ) {
      FET4.refill(e);
      int p = 0;
      if( FET4.l1sq() >= b2 ) p++;
      if( FET4.l2sq() >= b2 ) p++;
      if( FET4.l3sq() >= b2 ) p++;
      if( FET4.l4sq() >= b2 ) p++;
      if( FET4.l5sq() >= b2 ) p++;
      if( FET4.l6sq() >= b2 ) p++;
      markings(e) = p; // at least p edges in element e will be subdivided
    }
  }
  else if ( nsd==2 ) {
    FET3.refill(1);
    real maxlen = FET3.maxLength(); 
    for( e=2; e<=nel; e++ ) {
      FET3.refill(e);
      maxlen = max(maxlen,FET3.maxLength());
    } 
    real b2 = sqr(break_ratio*maxlen);
    for( e=1; e<=nel; e++ ){
      FET3.refill(e);
      int p = 0;
      if( FET3.l1sq() >= b2 ) p++;
      if( FET3.l2sq() >= b2 ) p++;
      if( FET3.l3sq() >= b2 ) p++;
      markings(e) = p; // at least p edges in element e will be subdivided
    }
  }
  else
    errorFP("GridFEAdT:: refineLongEdges","dimension %d not implemented",nsd);
  return refine(markings); // refine with "mixed method"
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: refine()
//-----------------------------------------------------------------------------
// return true if refinement successful
{
  switch ( nsd ) {
  case 1: return refine1D ();
  case 2: return refine2D ();
  case 3: return refine3D ();
  default: fatalerrorFP("GridFEAdT:: refine"," nsd = %d not implemented",nsd);
  }
  return false;
}
//-----------------------------------------------------------------------------
void GridFEAdT:: initRefinedGrid(GridFEAdT* newgrid,int new_nno, int new_nel)
//-----------------------------------------------------------------------------
// dimensioning the new grid: new GridFEAdT, children, parent
// common for 1D,2D and 3D linear and quadratic
{
  newgrid->redim( nsd, new_nno, new_nel, maxnne, getNoBoInds(), nne, onemat );
  newgrid->setElmType( elm_tp );
  newgrid->setNonUniformMesh();
  newgrid->setRefinementMethod( getRefinementMethod() );
  newgrid->setRegBisectOrShort( getRegBisectOrShort() );// only used in 2D
  newgrid->setNodeForEqualTest( useNodeForEqualTest() ); // only used in 3D
// set hierarchy relations of old grid and refined grid
  newgrid->setParentGrid( this );
  setChildGrid ( newgrid );
  setFinestGrid( false );
  newgrid->setFinestGrid( true );
  newgrid->setCoarsestGrid( false );
  setChildrenInfo( true );
  newgrid->setParentInfo( true );
  pos_child.redim( nel+1 ); // pos_child and parent will be filled later
  newgrid->parent.redim( new_nel );
}

//-----------------------------------------------------------------------------
void GridFEAdT:: checkReadyForRefine()
//-----------------------------------------------------------------------------
// Check that the grid is ready for refinement
{
  if ( ! isNeighborsComputed() )
    fatalerrorFP("GridFEAdT:: checkReadyForRefine", 
		 "Neighbor data is not computed.");
  if ( ! isThisFinestGrid() ) 
    fatalerrorFP("GridFEAdT:: checkReadyForRefine", 
		 "Refined grid already exists.");
  if( ! isNewNodesSet() ) 
    fatalerrorFP("GridFEAdT:: checkReadyForRefine",
		 "Edges must be marked before refinement.");
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: refine1D()
//-----------------------------------------------------------------------------
// return true if refinement successful
{
  checkReadyForRefine (); // check if the grid is ok

  if( "ElmB2n1D" != elm_tp && "ElmB3n1D" != elm_tp )
      fatalerrorFP("GridFEAdT:: refine",
		   "Not implemented yet, element type = %s",elm_tp.chars() );

  int b,e,i;

// Find the number of nodes and elements in the refined grid.
// The new_nodes gives the number of new nodes in interval

  int new_nel = nel;
  for( e=1; e<=nel; e++ ) // should work for both linear and quadratic elms
    new_nel += new_nodes( e, 1 );

  setCurrNodeNo(nno);

  int new_nno = getCurrNodeNo() + new_nel - nel; 
  if( "ElmB3n1D" == elm_tp ) 
    new_nno = getCurrNodeNo() + 2*(new_nel - nel); 

// dimensioning the new grid: new GridFEAdT, children, parent
  GridFEAdT* newgrid = new GridFEAdT;
  initRefinedGrid( newgrid, new_nno, new_nel );

// copy the name of the boundary indicators

  for (i = 1; i <= getNoBoInds(); i++)
    newgrid->putBoIndName( getBoIndName(i), i );

// copy the old nodes which will be present in the refined grid
// and copy boundary indicators

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

  for( i=1; i<=nno; i++ )
    newgrid->coor(i,1) = coor(i,1); 
  if ( getNoBoInds() ) 
    for( i=1; i<=nno; i++ )
      for( b=1; b<=getNoBoInds(); b++ )
	if( ind(i,b) == '1' ) 
	  nind(i,b) = '1';

// Start the numbering of the elements in the new grid so that the first 
// element is given element number 1.
  setCurrElmNo(0); 

  int e1,e2,e3,e4,e5; //numbers for new elements
  int m1,m2,m3,m4,m5,m6,m7,m8;    //numbers for new nodes
  int n1,n2,n12;      //node numbers in old grid
  real x1,x2,x12;     //coordinates in old grid

  if( "ElmB2n1D" == elm_tp ) {
    for( e=1; e <= nel; e++ ) {//Refine the elements independently and set data
      pos_child( e ) = getCurrElmNo()+1; 
      
      n1 = loc2glob( e, 1 );  
      n2 = loc2glob( e, 2 );
      x1 = coor( n1, 1 );  
      x2 = coor( n2, 1 ); 
      
      int number_of_new_elements = new_nodes( e, 1 ) + 1;
      switch ( number_of_new_elements ) {
//----------------------------------------------------------------
      case 1: // no refinement: just copy the info to the new element
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
	newgrid->putLoc2Glob12(n1, n2, e1);
	break;
//----------------------------------------------------------------
      case 2: // the element is divided into two elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();
	newgrid->putLoc2Glob12( n1, m1, e1 );
	newgrid->putLoc2Glob12( m1, n2, e2 );
	newgrid->coor(m1,1) = (x1 + x2)/2;
	for( b=1; b<=getNoBoInds(); b++ )
	  if( ind(n1,b) == '1' && ind(n2,b) == '1') // if both ends are set...
	    nind(m1,b) = '1';  // then set midpoint
	break;
//----------------------------------------------------------------
      case 3: // the element is divided into three elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();	m2 = incCurrNodeNo();
	newgrid->putLoc2Glob12( n1, m1, e1 );
	newgrid->putLoc2Glob12( m1, m2, e2 );
	newgrid->putLoc2Glob12( m2, n2, e3 );
	newgrid->coor(m1,1) = x1*0.6666666666666666 + x2*0.3333333333333333;
	newgrid->coor(m2,1) = x1*0.3333333333333333 + x2*0.6666666666666666;
	for( b=1; b<=getNoBoInds(); b++ )
	  if( ind(n1,b) == '1' && ind(n2,b) == '1')
	    nind(m1,b) = nind(m2,b) = '1';
	break;
//----------------------------------------------------------------
      case 4: // the element is divided into four elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();  m3 = incCurrNodeNo();
	newgrid->putLoc2Glob12( n1, m1, e1 );
	newgrid->putLoc2Glob12( m1, m2, e2 );
	newgrid->putLoc2Glob12( m2, m3, e3 );
	newgrid->putLoc2Glob12( m3, n2, e4 );
	newgrid->coor(m1,1) = x1*0.75 + x2*0.25;
	newgrid->coor(m2,1) = (x1 + x2)*0.5;
	newgrid->coor(m3,1) = x1*0.25 + x2*0.75;
	for( b=1; b<=getNoBoInds(); b++ )
	  if( ind(n1,b) == '1' && ind(n2,b) == '1')
	    nind(m1,b) = nind(m2,b) = nind(m3,b) = '1';
	break;
//----------------------------------------------------------------
      case 5: // the element is divided into five elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
	newgrid->putParent( e5 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();
	m3 = incCurrNodeNo();  m4 = incCurrNodeNo();
	newgrid->putLoc2Glob12( n1, m1, e1 );
	newgrid->putLoc2Glob12( m1, m2, e2 );
	newgrid->putLoc2Glob12( m2, m3, e3 );
	newgrid->putLoc2Glob12( m3, m4, e4 );
	newgrid->putLoc2Glob12( m4, n2, e5 );
	newgrid->coor(m1,1) = x1*0.8 + x2*0.2;
	newgrid->coor(m2,1) = x1*0.6 + x2*0.4;
	newgrid->coor(m3,1) = x1*0.4 + x2*0.6;
	newgrid->coor(m4,1) = x1*0.2 + x2*0.8;
	for( b=1; b<=getNoBoInds(); b++ )
	  if( ind(n1,b) == '1' && ind(n2,b) == '1')
	    nind(m1,b) = nind(m2,b) = nind(m3,b) = nind(m4,b) = '1';
	break;
//------------------------------------------------------
      default: 
//------------------------------------------------------
	fatalerrorFP("GridFEAdT:: refine1D",
		     "There must be 1-5 new elements.");
      } //end of switch
    }
  }
  else if( "ElmB3n1D" == elm_tp ) {
    for( e=1; e <= nel; e++ ) {//Refine the elements independently and set data
      pos_child( e ) = getCurrElmNo()+1; 
      
      n1  = loc2glob( e, 1 ); 
      n12 = loc2glob( e, 2 ); 
      n2  = loc2glob( e, 3 );
      x1  = coor( n1, 1 );  
      x12 = coor( n12,1 );
      x2  = coor( n2, 1 );

      int number_of_new_elements = new_nodes( e, 1 ) + 1;
      switch ( number_of_new_elements ) {
//----------------------------------------------------------------
      case 1: // no refinement: just copy the info to the new element
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
	newgrid->putLoc2Glob123( n1, n12, n2, e1 );
	break;
//----------------------------------------------------------------
      case 2: // the element is divided into two elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n1,  m1, n12, e1 );
	newgrid->putLoc2Glob123( n12, m2, n2,  e2 );
	newgrid->coor(m1,1) = (x1 + x12)/2;
	newgrid->coor(m2,1) = (x2 + x12)/2;
	for( b=1; b<=getNoBoInds(); b++ ){
	  if( ind(n1,b) == '1' && ind(n12,b) == '1') //if both ends are set...
	    nind(m1,b) = '1';
	  if( ind(n12,b) == '1' && ind(n2,b) == '1') //if both ends are set...
	    nind(m2,b) = '1';
	}
	break;
//----------------------------------------------------------------
      case 3: // the element is divided into three elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();
	m3 = incCurrNodeNo();  m4 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n1, m1,  m2, e1 );
	newgrid->putLoc2Glob123( m2, n12, m3, e2 );
	newgrid->putLoc2Glob123( m3, m4,  n2, e3 );
	x1 = coor(n1,1);  x2 = coor(n2,1); 
	newgrid->coor(m1,1) = x1*0.6666666666666666 + x12*0.3333333333333333;
	newgrid->coor(m2,1) = x1*0.3333333333333333 + x12*0.6666666666666666;
	newgrid->coor(m3,1) = x2*0.3333333333333333 + x12*0.6666666666666666;
	newgrid->coor(m4,1) = x2*0.6666666666666666 + x12*0.3333333333333333;
	for( b=1; b<=getNoBoInds(); b++ ) {
	  if( ind(n1,b) == '1' && ind(n12,b) == '1')
	    nind(m1,b) = nind(m2,b) = '1';
	  if( ind(n12,b) == '1' && ind(n2,b) == '1')
	    nind(m3,b) = nind(m4,b) = '1';
	}
	break;
//----------------------------------------------------------------
      case 4: // the element is divided into four elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();  m3 = incCurrNodeNo();
	m4 = incCurrNodeNo();  m5 = incCurrNodeNo();  m6 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n1,  m1, m2,  e1 );
	newgrid->putLoc2Glob123( m2,  m3, n12, e2 );
	newgrid->putLoc2Glob123( n12, m4, m5,  e3 );
	newgrid->putLoc2Glob123( m5,  m6, n2,  e4 );
	newgrid->coor(m1,1) = x1*0.75 + x12*0.25;
	newgrid->coor(m2,1) = x1*0.5  + x12*0.5;
	newgrid->coor(m3,1) = x1*0.25 + x12*0.75;
	newgrid->coor(m4,1) = x2*0.25 + x12*0.75;
	newgrid->coor(m5,1) = x2*0.5  + x12*0.5;
	newgrid->coor(m6,1) = x2*0.75 + x12*0.25;
	for( b=1; b<=getNoBoInds(); b++ ){
	  if( ind(n1,b) == '1' && ind(n12,b) == '1')
	    nind(m1,b) = nind(m2,b) = nind(m3,b) = '1';
	  if( ind(n12,b) == '1' && ind(n2,b) == '1')
	    nind(m4,b) = nind(m5,b) = nind(m6,b) = '1';
	}
	break;
//----------------------------------------------------------------
      case 5: // the element is divided into five elements
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
	newgrid->putParent( e5 = incCurrElmNo(), e );
	m1 = incCurrNodeNo();  m2 = incCurrNodeNo();
	m3 = incCurrNodeNo();  m4 = incCurrNodeNo();
	m5 = incCurrNodeNo();  m6 = incCurrNodeNo();
	m7 = incCurrNodeNo();  m8 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n1, m1,  m2, e1 );
	newgrid->putLoc2Glob123( m2, m3,  m4, e2 );
	newgrid->putLoc2Glob123( m4, n12, m5, e3 );
	newgrid->putLoc2Glob123( m5, m6,  m7, e4 );
	newgrid->putLoc2Glob123( m7, m8,  n2, e5 );
	newgrid->coor(m1,1) = x1*0.8 + x12*0.2;
	newgrid->coor(m2,1) = x1*0.6 + x12*0.4;
	newgrid->coor(m3,1) = x1*0.4 + x12*0.6;
	newgrid->coor(m4,1) = x1*0.2 + x12*0.8;
	newgrid->coor(m5,1) = x2*0.2 + x12*0.8;
	newgrid->coor(m6,1) = x2*0.4 + x12*0.6;
	newgrid->coor(m7,1) = x2*0.6 + x12*0.4;
	newgrid->coor(m8,1) = x2*0.8 + x12*0.2;
	for( b=1; b<=getNoBoInds(); b++ ){
	  if( ind(n1,b) == '1' && ind(n12,b) == '1')
	    nind(m1,b) = nind(m2,b) = nind(m3,b) = nind(m4,b) = '1';
	  if( ind(n12,b) == '1' && ind(n2,b) == '1')
	    nind(m5,b) = nind(m6,b) = nind(m7,b) = nind(m8,b) = '1';
	}
	break;
//------------------------------------------------------
      default: 
//------------------------------------------------------
	fatalerrorFP("GridFEAdT:: refine1D",
		     "There must be 1-5 new elements.");
      } //end of switch
    }
  }
  else
    fatalerrorFP("GridFEAdT:: refine1D","Not implemented element");

  if( new_nno != getCurrNodeNo() )
    fatalerrorFP("GridFEAdT:: refine1D","new_nno != getCurrNodeNo()");

  pos_child( nel+1 ) = getCurrElmNo()+1; 

// set the subdomain types for the new grid:
// copy the info from the parent elements

  if( ! onemat )
    for(e=1;e<=new_nel;e++)
      newgrid->material_type(e) = material_type(newgrid->getParent_eff(e));

  removeNewNodeInfo(); // remove the data showing the marked edges
  newgrid->setParentInfo( true );
  setChildrenInfo( true );

#ifdef SAFETY_CHECKS
  if( ! newgrid->checkElementOrientation() )
    s_o<<"Wrong orientation."<<endl;
  if( ! checkChildrenInfo() ) 
    s_o<<"children info failed."<<endl;
  if( ! newgrid->checkParentInfo() ) 
    s_o<<"parent info failed."<<endl;
#endif
  return true; //everything went fine
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: refine2D()
//-----------------------------------------------------------------------------
// return true if refinement successful
{
  checkReadyForRefine (); // check if the grid is ok
  if( "ElmT3n2D" != elm_tp && "ElmT6n2D" != elm_tp )
    fatalerrorFP("GridFEAdT:: refine",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  int b,e,i;

// Find the number of nodes and elements in the refined grid.
// The new nodes have been numbered before.
// A triangle which have n marked/numbered edges will be refined
// into n+1 elements.

// The new_nodes structure is 0 where no edge refinement is taking place.

  int new_nel = nel;
  for( e=1; e<=nel; e++ )
    if( new_nodes( e, 1 ) ) { 
      new_nel++;
      if( new_nodes( e, 2 ) ) new_nel++;
      if( new_nodes( e, 3 ) ) new_nel++;
    }

// current node number has been incremented during the numbering process
  int new_nno = getCurrNodeNo();
  if( "ElmT6n2D" == elm_tp ) // interior nodes are added below
    new_nno += new_nel - nel;

// dimensioning the new grid: new GridFEAdT, children, parent

  GridFEAdT* newgrid = new GridFEAdT;
  initRefinedGrid( newgrid, new_nno, new_nel );

// copy the name of the boundary indicators

  for (i = 1; i <= getNoBoInds(); i++) newgrid->putBoIndName( getBoIndName(i), i );

// copy the old nodes which will be present in the refined grid
// and copy boundary indicators

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

  for( i=1; i<=nno; i++ ){ // copy all the nodes from old grid
    newgrid->coor(i,1) = coor(i,1); 
    newgrid->coor(i,2) = coor(i,2); 
  }

  if ( getNoBoInds() ) 
    for( i=1; i<=nno; i++ )
      for( b=1; b<=getNoBoInds(); b++ )
	nind(i,b) = ind(i,b);

// Go through all elements and compute position of the new nodes
// at the midpoints of edges which have been assigned a node number.
// Also set indicators for these nodes

  int n1,n2,n3,n4,n5,n6;
  if( "ElmT3n2D" == elm_tp )
    for( e=1; e<=nel; e++ )
      if ( (n4 = new_nodes( e, 1 )) ){
	n1 =  loc2glob( e, 1 );
	n2 =  loc2glob( e, 2 );
	n3 =  loc2glob( e, 3 );
	n5 = new_nodes( e, 2 );
	n6 = new_nodes( e, 3 );
	if( ! newgrid->coor(n4,1) ){ // to avoid computing twice
	  newgrid->coor(n4,1) = (coor(n1,1)+coor(n2,1))/2;
	  newgrid->coor(n4,2) = (coor(n1,2)+coor(n2,2))/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( ind(n1,b) == '1' && ind(n2,b) == '1') //if both ends are set...
	      nind(n4,b) = '1';
	}
	if( ((n5) && ! newgrid->coor(n5,1)) ){ 
	  newgrid->coor(n5,1) = (coor(n2,1)+coor(n3,1))/2;
	  newgrid->coor(n5,2) = (coor(n2,2)+coor(n3,2))/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( ind(n2,b) == '1' && ind(n3,b) == '1' )
	      nind(n5,b) = '1';
	}
	if( ((n6) && ! newgrid->coor(n6,1)) ){
	  newgrid->coor(n6,1) = (coor(n3,1)+coor(n1,1))/2;
	  newgrid->coor(n6,2) = (coor(n3,2)+coor(n1,2))/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( ind(n3,b) == '1' && ind(n1,b) == '1' )
	      nind(n6,b) = '1';
	}
      }

// Start the numbering of the elements in the new grid so that the first 
// element is given element number 1.
  setCurrElmNo(0); 
  bool T6 = ( "ElmT6n2D" == elm_tp ) ? true : false;

  int e1,e2,e3,e4; //numbers for new elements
  int m1,m2,m3;
  int mm1,mm2,mm3, nn1,nn2,nn3; // used for quadratic elements

// what should be done if 3 edges are marked:
  bool REGULAR, TEST_SHORTEST;
  if( getRegularUniformRef () || reg_bisect_short==0 ){
    REGULAR = true; // always regular refinement
    TEST_SHORTEST = false; 
  }
  else {
    if( reg_bisect_short == 1 ){
      REGULAR = false; // always bisection refinement
      TEST_SHORTEST = false;
    }
    else // case reg_bisect_short == 2
      TEST_SHORTEST = true;// REGULAR will be determined by shortest edge
  }

  real l1,l2;
  int number_of_new_elements;
  if( ! T6 ) {  
    for( e=1; e <= nel; e++ ) { // Refine elements independently and set data
      pos_child( e ) = getCurrElmNo()+1; 
      number_of_new_elements = 1;
      if( (m1 = new_nodes( e, 1 )) ) {
	number_of_new_elements++;
	if( (m2 = new_nodes( e, 2 )) ) // m2 and m3 only if m1 is set
	  number_of_new_elements++;
	if( (m3 = new_nodes( e, 3 )) ) 
	  number_of_new_elements++;
      }
      switch ( number_of_new_elements ) {
//----------------------------------------------------------------
      case 1: // no refinement: just copy the info to the new element
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
	newgrid->putLoc2Glob123(loc2glob(e,1),loc2glob(e,2),loc2glob(e,3), e1);
	break;
//----------------------------------------------------------------
      case 2:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	n3 = loc2glob( e, 3 );
	newgrid->putLoc2Glob123( n3, loc2glob(e,1), m1, e1 );
	newgrid->putLoc2Glob123( loc2glob(e,2), n3, m1, e2 );
	break;
//----------------------------------------------------------------
      case 3:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	n3 =  loc2glob( e, 3 );
	if( m2 ) { // edge 1 and edge 2 marked
	  newgrid->putLoc2Glob123( n3, loc2glob(e,1), m1, e1 );
	  newgrid->putLoc2Glob123( n3, m1, m2, e2 );
	  newgrid->putLoc2Glob123( m1, loc2glob(e,2), m2, e3 );
	}
	else { // edge 1 and edge 3 marked
	  newgrid->putLoc2Glob123( loc2glob(e,2), n3, m1, e1 );
	  newgrid->putLoc2Glob123( m1, n3, m3, e2 );
	  newgrid->putLoc2Glob123( loc2glob(e,1), m1, m3, e3 );
        }
	break;
//----------------------------------------------------------------
      case 4:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
// it is assumed that edge 1 is the longest edge, these triangles are common
	n3 = loc2glob( e, 3 );
	newgrid->putLoc2Glob123( loc2glob( e, 1 ), m1, m3, e1 );
	newgrid->putLoc2Glob123( m1, loc2glob( e, 2 ), m2, e2 );
	if ( TEST_SHORTEST ) { // choose the edge which is shortest
          // an alternative is to check if largest angle is obtuse
	  l1 = sqr(newgrid->coor(m2,1) - newgrid->coor(m3,1)) +
	    sqr(newgrid->coor(m2,2) - newgrid->coor(m3,2));
	  l2 = sqr(newgrid->coor(m1,1) - newgrid->coor(n3,1)) +
	    sqr(newgrid->coor(m1,2) - newgrid->coor(n3,2));
	  REGULAR = ( l1*0.999999 <= l2 ) ? true : false;
	  // the factor 0.9999999 is to avoid arbitrariness in the equal case
	}
	if( REGULAR ) { // regular refinement
	  newgrid->putLoc2Glob123( m3, m2, n3, e3 );
	  newgrid->putLoc2Glob123( m2, m3, m1, e4 );
	} else { // three edges drawn to midpoint of longest edge
	  newgrid->putLoc2Glob123( m1, n3, m3, e3 );
	  newgrid->putLoc2Glob123( n3, m1, m2, e4 );
	}
	break;
//------------------------------------------------------
      default: 
//------------------------------------------------------
	fatalerrorFP("GridFEAdT:: refine","There must be 1-4 new elements.");
      } //end of switch
    }
  }
  else { // T6 is true,  quadratic elements
    for( e=1; e <= nel; e++ ) { // Refine elements independently and set data
      pos_child( e ) = getCurrElmNo()+1; 
      number_of_new_elements = 1;
      if( (nn1 = new_nodes( e, 1 )) ) {
	number_of_new_elements++;
	if( (nn2 = new_nodes( e, 2 )) ) 
	  number_of_new_elements++;
	if( (nn3 = new_nodes( e, 3 )) ) 
	  number_of_new_elements++;
      }
      switch ( number_of_new_elements ) {
//----------------------------------------------------------------
      case 1: // no refinement: just copy the info to the new element
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
	newgrid->putLoc2Glob123(loc2glob(e,1),loc2glob(e,2),loc2glob(e,3), e1);
	newgrid->putLoc2Glob456(loc2glob(e,4),loc2glob(e,5),loc2glob(e,6), e1);
	break;
//----------------------------------------------------------------
      case 2:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	m1 = loc2glob( e, 4 );
	n3 = loc2glob( e, 3 );
	mm1 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n3, loc2glob(e,1), m1, e1 );
	newgrid->putLoc2Glob456( loc2glob(e,6), nn1, mm1, e1 );
	newgrid->putLoc2Glob123( loc2glob(e,2), n3, m1, e2 );
	newgrid->putLoc2Glob456( loc2glob(e,5), mm1, new_nodes(e,4), e2 );
	break;
//----------------------------------------------------------------
      case 3:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	n3 =  loc2glob( e, 3 );
	mm1 = incCurrNodeNo();
	m1 = loc2glob( e, 4 );
	if( nn2 ) { // edge 1 and edge 2 are marked
	  m2 = loc2glob( e, 5 );
	  mm2 = incCurrNodeNo();
	  newgrid->putLoc2Glob123( n3, loc2glob(e,1), m1, e1 );
	  newgrid->putLoc2Glob456( loc2glob(e,6) , nn1, mm2, e1 );
	  newgrid->putLoc2Glob123( n3, m1, m2, e2 );
	  newgrid->putLoc2Glob456( mm2, mm1, new_nodes(e,5), e2 );
	  newgrid->putLoc2Glob123( m1, loc2glob(e,2), m2, e3 );
	  newgrid->putLoc2Glob456( new_nodes(e,4), nn2, mm1, e3 );
	}
	else { // edge 1 and edge 3 are marked
	  m3 = loc2glob( e, 6 );
	  mm3 = incCurrNodeNo();
	  newgrid->putLoc2Glob123( loc2glob(e,2), n3, m1, e1 );
	  newgrid->putLoc2Glob456( loc2glob(e,5) , mm1, new_nodes(e,4), e1 );
	  newgrid->putLoc2Glob123( m1, n3, m3, e2 );
	  newgrid->putLoc2Glob456( mm1, nn3, mm3, e2 );
	  newgrid->putLoc2Glob123( loc2glob(e,1), m1, m3, e3 );
	  newgrid->putLoc2Glob456( nn1, mm3, new_nodes(e,6), e3 );
        }
	break;
//----------------------------------------------------------------
      case 4:
//----------------------------------------------------------------
	newgrid->putParent( e1 = incCurrElmNo(), e );
	newgrid->putParent( e2 = incCurrElmNo(), e );
	newgrid->putParent( e3 = incCurrElmNo(), e );
	newgrid->putParent( e4 = incCurrElmNo(), e );
// assume that 1 is the longest edge
	n1 = loc2glob( e, 1 );	n2 = loc2glob( e, 2 );	n3 = loc2glob( e, 3 );
	m1 = loc2glob( e, 4 );	m2 = loc2glob( e, 5 );	m3 = loc2glob( e, 6 );
	mm1 = incCurrNodeNo();	mm2 = incCurrNodeNo();	mm3 = incCurrNodeNo();
	newgrid->putLoc2Glob123( n1, m1, m3, e1 );
	newgrid->putLoc2Glob456( nn1, mm1, new_nodes(e,6), e1 );
	newgrid->putLoc2Glob123( m1, n2, m2, e2 );
	newgrid->putLoc2Glob456( new_nodes(e,4), nn2, mm2, e2 );
	if ( TEST_SHORTEST ) {// choose shortest edge alternative 
	  l1 = sqr(newgrid->coor(m2,1) - newgrid->coor(m3,1)) +
	    sqr(newgrid->coor(m2,2) - newgrid->coor(m3,2));
	  l2 = sqr(newgrid->coor(m1,1) - newgrid->coor(n3,1)) +
	    sqr(newgrid->coor(m1,2) - newgrid->coor(n3,2));
	  REGULAR = ( l1*0.999999 <= l2 ) ? true : false;
	}
	if( REGULAR ) { //regular if shorter distance
	  newgrid->putLoc2Glob123( m3, m2, n3, e3 );
	  newgrid->putLoc2Glob456( mm3, new_nodes( e, 5 ), nn3, e3 );
	  newgrid->putLoc2Glob123( m2, m3, m1, e4 );
	  newgrid->putLoc2Glob456( mm3, mm1, mm2, e4 );
	} else { // three edges drawn to midpoint of longest edge
	  newgrid->putLoc2Glob123( m1, n3, m3, e3 );
	  newgrid->putLoc2Glob456( mm3, nn3, mm1, e3 );
	  newgrid->putLoc2Glob123( n3, m1, m2, e4 );
	  newgrid->putLoc2Glob456( mm3, mm2, new_nodes( e, 5 ), e4 );
	}
	break;
//------------------------------------------------------
      default: 
//------------------------------------------------------
	fatalerrorFP("GridFEAdT:: refine","There must be 1-4 new elements.");
      } //end of switch
    }
  } // end of quadratic elements (T6)

  pos_child( nel+1 ) = getCurrElmNo()+1; 

// set subdomain types for the new grid copy the info from the parent elements
  if( ! onemat )
    for(e=1;e<=new_nel;e++)
      newgrid->material_type(e) = material_type(newgrid->getParent_eff(e));

  real x1,y1,x2,y2,x3,y3;

// newgrid->coor is initialized to 0

// compute positions and boundary indicators for the new midpoint nodes
  if( T6 ) // already done for T3 case
    for( e=1; e<=new_nel; e++) {
      if( getNoChildren( newgrid->getParent_eff( e ) ) > 1 ) {
	// the element has at least one new node
	n1 = newgrid->loc2glob( e, 1 );
	n2 = newgrid->loc2glob( e, 2 );
	n3 = newgrid->loc2glob( e, 3 );
	n4 = newgrid->loc2glob( e, 4 );
	n5 = newgrid->loc2glob( e, 5 );
	n6 = newgrid->loc2glob( e, 6 );
	x1 = newgrid->coor(n1,1); y1 = newgrid->coor(n1,2);
	x2 = newgrid->coor(n2,1); y2 = newgrid->coor(n2,2);
	x3 = newgrid->coor(n3,1); y3 = newgrid->coor(n3,2);
	if( ! newgrid->coor(n4,1) ){ // to avoid computing things twice
	  newgrid->coor(n4,1) = (x1 + x2)/2;
	  newgrid->coor(n4,2) = (y1 + y2)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n1,b) == '1'&& nind(n2,b) == '1' )
	      nind(n4,b) = '1';
	}
	if( ! newgrid->coor(n5,1) ){
	  newgrid->coor(n5,1) = (x2 + x3)/2;
	  newgrid->coor(n5,2) = (y2 + y3)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n2,b) == '1' && nind(n3,b) == '1' )
	      nind(n5,b) = '1';
	}
	if( ! newgrid->coor(n6,1) ){
	  newgrid->coor(n6,1) = (x3 + x1)/2;
	  newgrid->coor(n6,2) = (y3 + y1)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n3,b) == '1' && nind(n1,b) == '1' )
	      nind(n6,b) = '1';
	}
      }
    }

  removeNewNodeInfo(); // remove the data showing the marked edges
  newgrid->setParentInfo( true );
  setChildrenInfo( true );
#ifdef SAFETY_CHECKS
  if( ! newgrid->checkElementOrientation() )
    s_o<<"Wrong orientation."<<endl;
  if( !checkChildrenInfo() ) 
    s_o<<"children info failed."<<endl;
  if( !newgrid->checkParentInfo() ) 
    s_o<<"parent info failed."<<endl;
#endif
  return true; //everything went fine
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: refine3D()
//-----------------------------------------------------------------------------
{
// return true if refinement successful

  checkReadyForRefine ();  // check if the grid is ok
  if( "ElmT4n3D" != elm_tp && "ElmT10n3D" != elm_tp )
    fatalerrorFP("GridFEAdT:: refine3D",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  int b,e,i;

// Find the number of nodes and elements in the refined grid.
// The new nodes have been numbered before.
// A triangle which have n marked/numbered edges will be refined
// into n+2 elements if edge 6 is refined otherwise into n+1 elements

// The new_nodes structure is 0 where no edge refinement is taking place.

  int new_nel = nel;
  for( e=1; e<=nel; e++ )
    if( new_nodes( e, 1 ) ) { // if edge 2-6 are refined then so are edge 1
      new_nel++;
      if( new_nodes( e, 2 ) )  new_nel++;
      if( new_nodes( e, 3 ) )  new_nel++;
      if( new_nodes( e, 4 ) )  new_nel++;
      if( new_nodes( e, 5 ) )  new_nel++;
      if( new_nodes( e, 6 ) ){ new_nel++; new_nel++; }
    }

// current node number was incremented during the node numbering process
  int new_nno = getCurrNodeNo(); 

  GridFEAdT* newgrid = new GridFEAdT;
  initRefinedGrid( newgrid, new_nno, new_nel ); // redim and initialization

// copy the name of the boundary indicators

  for (i = 1; i <= getNoBoInds(); i++) newgrid->putBoIndName( getBoIndName(i), i );

// copy the old nodes and indicators which will be present in refined grid

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

  for( i=1; i<=nno; i++ ){
    newgrid->coor(i,1) = coor(i,1); 
    newgrid->coor(i,2) = coor(i,2); 
    newgrid->coor(i,3) = coor(i,3); 
  }

  if ( getNoBoInds() )
    for( i=1; i<=nno; i++ )
      for( b=1; b<=getNoBoInds(); b++ )
	nind(i,b) = ind(i,b);

  int e1,e2,e3,e4,e5,e6,e7,e8; // numbers for new elements
  int n1,n2,n3,n4,n5,n6,n7,n8,n9,n10; // numbers for nodes in element
  int mid,m1,m2,node2,node3;

  setCurrElmNo(0); // first element is given element number 1.

  bool T10 = ( "ElmT10n3D" == elm_tp ) ? true : false;

  if (T10) errorFP("GridFEAdT::refine3D","ElmT10n3D not yet implemented");

#ifdef SAFETY_CHECKS
  VecSimple(int) edgeorder;
  edgeorder.redim(6);
#endif
  int refcase;
  for( e=1; e <= nel; e++ ) { // Refine the elements independently and set data
    pos_child( e ) = getCurrElmNo()+1; 
    n1 = loc2glob(e,1);  // these are needed for all cases
    n2 = loc2glob(e,2);
    n3 = loc2glob(e,3);
    n4 = loc2glob(e,4);
    n5 = new_nodes( e, 1 ); // always set if any refinement

    refcase = 1;
    if( n5 ){ // if not n5 then none of the others
      n6  = new_nodes( e, 2 );
      n7  = new_nodes( e, 3 );
      n8  = new_nodes( e, 4 );
      n9  = new_nodes( e, 5 );
      n10 = new_nodes( e, 6 );
      for(i=1;i<=6;i++){
	switch ( i ) { // set nodes on edge 1-6
	case 1: mid=n5;  m1=n1;  m2=n2; break;
	case 2: mid=n6;  m1=n2;  m2=n3; break;
	case 3: mid=n7;  m1=n3;  m2=n1; break;
	case 4: mid=n8;  m1=n1;  m2=n4; break;
	case 5: mid=n9;  m1=n2;  m2=n4; break;
	case 6: mid=n10; m1=n3;  m2=n4; break;
	}
	if( mid ) refcase++;  
	if( mid && ! newgrid->coor(mid,1) ){ // to avoid recomputing things
// Compute position and indicators of the new nodes at midpoints of edges 
	  newgrid->coor(mid,1) = (coor(m1,1) + coor(m2,1))/2;
	  newgrid->coor(mid,2) = (coor(m1,2) + coor(m2,2))/2;
	  newgrid->coor(mid,3) = (coor(m1,3) + coor(m2,3))/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if(ind(m1,b) == '1' && ind(m2,b) == '1') // if both ends are set...
	      nind(mid,b) = '1'; // then set midpoint
	}
      }
    }
#ifdef SAFETY_CHECKS  // some tests in nonoptimized version
    getEdgeLengthOrder( edgeorder, e );
    if( edgeorder(1) != 1 ) 
      errorFP("GridFEAdT::refine3D","1 must be longest");
    if( refcase > 1 && !n5 ) 
      errorFP("GridFEAdT::refine3D","n5 must be set");
#endif

    switch ( refcase ) {
//----------------------------------------------------------------
    case 1 : // no refinement: just copy the info to the new element
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
      newgrid->putLoc2Glob1234( n1, n2, n3, n4  , e1 );
      break;
//----------------------------------------------------------------
    case 2 :
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
      newgrid->putParent( e2 = incCurrElmNo(), e ); // e is parent to e2
//	cout<<e<<" 5"<<endl;
      switch ( edge_face2(e) ) {
      case 2:  newgrid->putLoc2Glob1234( n3, n2, n4, n5  , e1 );  break;
      case 5:  newgrid->putLoc2Glob1234( n2, n4, n3, n5  , e1 );  break;
      case 6:  newgrid->putLoc2Glob1234( n4, n3, n2, n5  , e1 );  break;
      default:  errorFP("GridFEAdT::refine3D","wrong number in edge_face2");
      }
      switch ( edge_face3(e) ) {
      case 3:  newgrid->putLoc2Glob1234( n1, n3, n4, n5  , e2 );  break;
      case 4:  newgrid->putLoc2Glob1234( n4, n1, n3, n5  , e2 );  break;
      case 6:  newgrid->putLoc2Glob1234( n3, n4, n1, n5  , e2 );  break;
      default:  errorFP("GridFEAdT::refine3D","wrong number in edge_face3"); 
      }
      break;
//----------------------------------------------------------------
    case 3 :
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
      newgrid->putParent( e2 = incCurrElmNo(), e ); // e is parent to e2
      newgrid->putParent( e3 = incCurrElmNo(), e ); // e is parent to e3
      for( node2=2; node2<=6; node2++ ) 
	if( new_nodes( e, node2) )  
	  break;
      switch ( node2 ) {
      case 2:
//	  cout<<e<<" 56"<<endl;
	switch ( edge_face3(e) ) {
	case 3:  newgrid->putLoc2Glob1234( n1, n3, n4, n5  , e1 );  break;
	case 4:  newgrid->putLoc2Glob1234( n4, n1, n3, n5  , e1 );  break;
	case 6:  newgrid->putLoc2Glob1234( n3, n4, n1, n5  , e1 );  break;
	default:  errorFP("GridFEAdT::refine3D","wrong number in edge_face3");
	}
	newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e2 );
	newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e3 );
	break;
      case 3:
//	  cout<<e<<" 57"<<endl;
	switch ( edge_face2(e) ) {
	case 2:  newgrid->putLoc2Glob1234( n3, n2, n4, n5  , e1 );  break;
	case 5:  newgrid->putLoc2Glob1234( n2, n4, n3, n5  , e1 );  break;
	case 6:  newgrid->putLoc2Glob1234( n4, n3, n2, n5  , e1 );  break;
	default:  errorFP("GridFEAdT::refine3D","wrong number in edge_face2");
	}
	newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e2 );
	newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e3 );
	break;
      case 4:
//	  cout<<e<<" 58"<<endl;
	switch ( edge_face2(e) ) {
	case 2:  newgrid->putLoc2Glob1234( n3, n2, n4, n5  , e1 );  break;
	case 5:  newgrid->putLoc2Glob1234( n2, n4, n3, n5  , e1 );  break;
	case 6:  newgrid->putLoc2Glob1234( n4, n3, n2, n5  , e1 );  break;
	default:  errorFP("GridFEAdT::refine3D","wrong number in edge_face2");
	}
	newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e2 );
	newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e3 );
	break;
      case 5:
//	  cout<<e<<" 59"<<endl;
	switch ( edge_face3(e) ) {
	case 3:  newgrid->putLoc2Glob1234( n1, n3, n4, n5  , e1 );  break;
	case 4:  newgrid->putLoc2Glob1234( n4, n1, n3, n5  , e1 );  break;
	case 6:  newgrid->putLoc2Glob1234( n3, n4, n1, n5  , e1 );  break;
	default: errorFP("GridFEAdT::refine3D","wrong number in edge_face3");
	}
	newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e2 );
	newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e3 );
	break;
      case 6:
//	  cout<<e<<" 510"<<endl;
	newgrid->putParent( e4 = incCurrElmNo(), e ); // e is parent to e4
	newgrid->putLoc2Glob1234( n1, n3, n10, n5  , e1 );
	newgrid->putLoc2Glob1234( n2, n4, n10, n5  , e2 );
	newgrid->putLoc2Glob1234( n3, n2, n10, n5  , e3 );
	newgrid->putLoc2Glob1234( n4, n1, n10, n5  , e4 );
	break;
      default: errorFP("GridFEAdT::refine3D",
		       "case 3 did not find second node");
      }
      break;
//----------------------------------------------------------------
    case 4 : // find the two edges except edge 1 to be refined
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e ); // e is parent to e1
      newgrid->putParent( e2 = incCurrElmNo(), e ); // e is parent to e2
      newgrid->putParent( e3 = incCurrElmNo(), e ); // e is parent to e3
      newgrid->putParent( e4 = incCurrElmNo(), e ); // e is parent to e4
      if( n10 )
	newgrid->putParent( e5 = incCurrElmNo(), e ); // e is parent to e5
// first find if second node is n6 n7 n8 or n9
// there after find third node which comes after second node
      for( node2=2; node2<=6; node2++ ) 
	if( new_nodes( e, node2 ) )  
	  break;
      switch ( node2 ) {
      case 2:
	for( node3=3; node3<=6; node3++ ) 
	  if( new_nodes(e,node3) )  
	    break;
	switch ( node3 ) {
	case 3: // case 5 6 7 
//	    cout<<e<<" 567"<<endl;
	  newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e1 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e2 );
	  newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e3 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e4 );
	  break;
	case 4:// case 5 6 8
//	    cout<<e<<" 568"<<endl;
	  newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e1 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e2 );
	  newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e3 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e4 );
	  break;
	case 5:// case 5 6 9
//	    cout<<e<<" 569"<<endl;
	  switch ( edge_face3(e) ) {
	  case 3:  newgrid->putLoc2Glob1234( n1, n3, n4, n5  , e1 );  break;
	  case 4:  newgrid->putLoc2Glob1234( n4, n1, n3, n5  , e1 );  break;
	  case 6:  newgrid->putLoc2Glob1234( n3, n4, n1, n5  , e1 );  break;
	  default:  errorFP("GridFEAdT::refine3D","wrong num in edge_face3");
	  }
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e2 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e3 );
	    newgrid->putLoc2Glob1234( n6, n4, n5, n9  , e4 );
	  } else {
	    newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e3 );
	    newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e4 );
	  }
	  break;
	case 6: // case 5 6 10
//	    cout<<e<<" 5610"<<endl;
	  newgrid->putLoc2Glob1234( n1, n3, n10, n5  , e1 ); // common elms
	  newgrid->putLoc2Glob1234( n4, n1, n10, n5  , e2 );
	  newgrid->putLoc2Glob1234( n5, n3, n10, n6  , e3 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e4 );
	    newgrid->putLoc2Glob1234( n4, n6, n5, n10 , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n10  , e4 );
	    newgrid->putLoc2Glob1234( n10, n2, n5, n6  , e5 );
	  }
	  break;
	default:
	  errorFP("GridFEAdT:: refine3D","node 5 6,did not find third node");
	}
	break;
      case 3:
	for( node3=4; node3<=6; node3++ ) 
	  if( new_nodes( e, node3 ) )  
	    break;
	switch(node3){
	case 4: // case 5 7 8
//	    cout<<e<<" 578"<<endl;
	  switch ( edge_face2(e) ) {
	  case 2:  newgrid->putLoc2Glob1234( n3, n2, n4, n5  , e1 );  break;
	  case 5:  newgrid->putLoc2Glob1234( n2, n4, n3, n5  , e1 );  break;
	  case 6:  newgrid->putLoc2Glob1234( n4, n3, n2, n5  , e1 );  break;
	  default:  
	    errorFP("GridFEAdT::refine3D","wrong number in edge_face2");
	  }
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e2 ); // common elms
	  if ( edge_face3(e) == 3 ) {
	    newgrid->putLoc2Glob1234( n4, n7, n5, n8  , e3 );
	    newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e4 );
	  } else {
	    newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e3 );
	    newgrid->putLoc2Glob1234( n8, n3, n5, n7  , e4 );
	  }
	  break;
	case 5: // case 5 7 9
//	    cout<<e<<" 579"<<endl;
	  newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e1 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e2 );
	  newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e3 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e4 );
	  break;
	case 6: // case 5 7 10
//	    cout<<e<<" 5710"<<endl;
	  newgrid->putLoc2Glob1234( n2, n4, n10, n5  , e1 );
	  newgrid->putLoc2Glob1234( n3, n2, n10, n5  , e2 );
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10  , e3 );
	  if ( edge_face3(e) == 3 ) {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e4 );
	    newgrid->putLoc2Glob1234( n7, n4, n5, n10 , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n10  , e4 );
	    newgrid->putLoc2Glob1234( n1, n10, n5, n7  , e5 );
	  }
	  break;
	default:
	  errorFP("GridFEAdT:: refine3D","node 5 7,did not find third node");
	}
	break;
      case 4:
	if( n9 ) { // case 5 8 9 
//	    cout<<e<<" 589"<<endl;
	  newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e1 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e2 );
	  newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e3 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e4 );
	} else {	  // case 5 8 10
//	    cout<<e<<" 5810"<<endl;
	  newgrid->putLoc2Glob1234( n2, n4, n10, n5  , e1 );
	  newgrid->putLoc2Glob1234( n3, n2, n10, n5  , e2 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8  , e3 );
	  if ( edge_face3(e) == 4 ) {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e4 );
	    newgrid->putLoc2Glob1234( n3, n8, n5, n10 , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n10 , e4 );
	    newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e5 );
	  }
	}
	break;
      case 5: // case 5 9 10
//	  cout<<e<<" 5910"<<endl;
	newgrid->putLoc2Glob1234( n1, n3, n10, n5  , e1 );
	newgrid->putLoc2Glob1234( n4, n1, n10, n5  , e2 );
	newgrid->putLoc2Glob1234( n4, n5, n10, n9  , e3 );
	if( edge_face2(e) == 5 ) {
	  newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e4 );
	  newgrid->putLoc2Glob1234( n9, n3, n5, n10 , e5 );
	} else {
	  newgrid->putLoc2Glob1234( n2, n3, n5, n10 , e4 );
	  newgrid->putLoc2Glob1234( n10, n2, n9, n5 , e5 );
	}
	break;
      case 6:
      default: errorFP("GridFEAdT::refine3D",
		       "case 4 did not find second node");
      }
      break;
//----------------------------------------------------------------
    case 5: // find the two edges which are not refined
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e );
      newgrid->putParent( e2 = incCurrElmNo(), e );
      newgrid->putParent( e3 = incCurrElmNo(), e );
      newgrid->putParent( e4 = incCurrElmNo(), e );
      newgrid->putParent( e5 = incCurrElmNo(), e );
      if ( n10 )
	newgrid->putParent( e6 = incCurrElmNo(), e ); // e is parent to e6
      for( node2=2; node2<=6; node2++ ) 
	if( ! new_nodes(e,node2) )  
	  break;
      switch ( node2 ) {
      case 2:
	for( node3=3; node3<=6; node3++ ) 
	  if( ! new_nodes( e, node3 ) ) 
	    break;
	switch ( node3 ) {
	case 3: // not node 6 and 7 => case 5 8 9 10
//	    cout<<e<<" 58910"<<endl;
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8  , e1 );
	  if ( edge_face3(e) == 4 ) {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e2 );
	    newgrid->putLoc2Glob1234( n3, n8, n5, n10 , e3 );
	  } else {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n10 , e2 );
	    newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e3 );
	  }
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9  , e4 );
	  if( edge_face2(e) == 5 ) {
	    newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e5 );
	    newgrid->putLoc2Glob1234( n9, n3, n5, n10 , e6 );
	  } else {
	    newgrid->putLoc2Glob1234( n2, n3, n5, n10 , e5 );
	    newgrid->putLoc2Glob1234( n10, n2, n9, n5 , e6 );
	  }
	  break;
	case 4: // not node 6 and 8 => case 5 7 9 10
//	    cout<<e<<" 57910"<<endl;
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10  , e1 );
	  if ( edge_face3(e) == 3 ) {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e2 );
	    newgrid->putLoc2Glob1234( n4, n5, n7, n10 , e3 );
	  } else {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n10  , e2 );
	    newgrid->putLoc2Glob1234( n1, n10, n5, n7  , e3 );
	  }
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9  , e4 );
	  if( edge_face2(e) == 5 ) {
	    newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e5 );
	    newgrid->putLoc2Glob1234( n3, n5, n9, n10 , e6 );
	  } else {
	    newgrid->putLoc2Glob1234( n2, n3, n5, n10 , e5 );
	    newgrid->putLoc2Glob1234( n10, n2, n9, n5 , e6 );
	  }
	  break;
	case 5: // not node 6 and 9 => case 5 7 8 10
//	    cout<<e<<" 57810"<<endl;
	  newgrid->putLoc2Glob1234( n2, n4, n10, n5  , e1 );
	  newgrid->putLoc2Glob1234( n3, n2, n10, n5  , e2 );
	  switch ( edge_face3(e) ) {
	  case 3:
	    newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e3 );
	    newgrid->putLoc2Glob1234( n7, n4, n5, n10 , e4 );
	    newgrid->putLoc2Glob1234( n4, n7, n5, n8  , e5 );
	    newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e6 );
	    break;
	  case 4:
	    newgrid->putLoc2Glob1234( n8, n3, n5, n7  , e3 );
	    newgrid->putLoc2Glob1234( n3, n8, n5, n10 , e4 );
	    newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e5 );
	    newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e6 );
	    break;
	  case 6:
	    newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e3 );
	    newgrid->putLoc2Glob1234( n1, n10, n5, n7 , e4 );
	    newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e5 );
	    newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e6 );
	    break;
	  default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face3");
	  }
	  break;
	case 6: // not node 6 and 10 => case 5 7 8 9
//	    cout<<e<<" 5789"<<endl;
	  newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e1 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e2 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e3 );// common elms
	  if ( edge_face3(e) == 3 ) {
	    newgrid->putLoc2Glob1234( n5, n4, n7, n8  , e4 );
	    newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e4 );
	    newgrid->putLoc2Glob1234( n8, n3, n5, n7  , e5 );
	  }
	  break;
	default:
	  errorFP("GridFEAdT:: refine3D","node 6,did not find second node");
	}
	break;
      case 3:
	for( node3=4; node3<=6; node3++ ) 
	  if( ! new_nodes( e, node3 ) ) 
	    break;
	switch ( node3 ) {
	case 4: // not node 7 and 8 => case 5 6 9 10
//	    cout<<e<<" 56910"<<endl;
	  newgrid->putLoc2Glob1234( n1, n3, n10, n5  , e1 );
	  newgrid->putLoc2Glob1234( n4, n1, n10, n5  , e2 );
	  switch ( edge_face2(e) ) {
	  case 2:
	    newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e3 );
	    newgrid->putLoc2Glob1234( n6, n4, n5, n9  , e4 );
	    newgrid->putLoc2Glob1234( n4, n6, n5, n10 , e5 );
	    newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e6 );
	    break;
	  case 5:
	    newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e3 );
	    newgrid->putLoc2Glob1234( n9, n3, n5, n10 , e4 );
	    newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e5 );
	    newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e6 );
	    break;
	  case 6:
	    newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e3 );
	    newgrid->putLoc2Glob1234( n10, n2, n5, n6 , e4 );
	    newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e5 );
	    newgrid->putLoc2Glob1234( n2, n10, n5, n9 , e6 );
	    break;
	  default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face2");
	  }
	  break;
	case 5: // not node 7 and 9 => case 5 6 8 10
//	    cout<<e<<" 56810"<<endl;
	  newgrid->putLoc2Glob1234( n5, n3, n10, n6  , e1 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e2 );
	    newgrid->putLoc2Glob1234( n4, n6, n5, n10 , e3 );
	  } else {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n10  , e2 );
	    newgrid->putLoc2Glob1234( n10, n2, n5, n6  , e3 );
	  }
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8  , e4 );
	  if ( edge_face3(e) == 4 ) {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e5 );
	    newgrid->putLoc2Glob1234( n3, n8, n5, n10 , e6 );
	  } else {
	    newgrid->putLoc2Glob1234( n3, n1, n5, n10 , e5 );
	    newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e6 );
	  }
	  break;
	case 6: // not node 7 and 10 => case 5 6 8 9
//	    cout<<e<<" 5689"<<endl;
	  newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e1 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e2 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e3 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e4 );
	    newgrid->putLoc2Glob1234( n4, n5, n6, n9  , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e4 );
	    newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e5 );
	  }
	  break;
	default:
	  errorFP("GridFEAdT:: refine3D","node 6,did not find second node");
	}
	break;
      case 4:
	if( ! n9 ) { // not node 8 and 9 => case 5 6 7 10
//	    cout<<e<<" 56710"<<endl;
	  newgrid->putLoc2Glob1234( n5, n3, n10, n6  , e1 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e2 );
	    newgrid->putLoc2Glob1234( n4, n6, n5, n10 , e3 );
	  } else {
	    newgrid->putLoc2Glob1234( n4, n2, n5, n10  , e2 );
	    newgrid->putLoc2Glob1234( n10, n2, n5, n6  , e3 );
	  }
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10  , e4 );
	  if ( edge_face3(e) == 3 ) {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e5 );
	    newgrid->putLoc2Glob1234( n7, n4, n5, n10 , e6 );
	  } else {
	    newgrid->putLoc2Glob1234( n1, n4, n5, n10  , e5 );
	    newgrid->putLoc2Glob1234( n1, n10, n5, n7  , e6 );
	  }
	} else { // not node 8 and 10 => case 5 6 7 9
//	  cout<<e<<" 5679"<<endl;
	  newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e1 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e2 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e3 );
	  if ( edge_face2(e) == 2 ) {
	    newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e4 );
	    newgrid->putLoc2Glob1234( n6, n4, n5, n9  , e5 );
	  } else {
	    newgrid->putLoc2Glob1234( n5, n3, n9, n6  , e4 );
	    newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e5 );
	  }
	}
	break;
      case 5: // not node 9 and 10 => case 5 6 7 8
//	  cout<<e<<" 5678"<<endl;
	newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e1 );
	newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e2 );
	newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e3 );
	if ( edge_face3(e) == 3 ) {
	  newgrid->putLoc2Glob1234( n4, n7, n5, n8  , e4 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e5 );
	} else {
	  newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e4 );
	  newgrid->putLoc2Glob1234( n3, n5, n8, n7  , e5 );
	}
	break;
      case 6:
      default: errorFP("GridFEAdT::refine3D",
		       "case 5 did not find second node");
      }
      break;
//----------------------------------------------------------------
    case 6: // find the edge which not is going to be refined
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e );
      newgrid->putParent( e2 = incCurrElmNo(), e );
      newgrid->putParent( e3 = incCurrElmNo(), e );
      newgrid->putParent( e4 = incCurrElmNo(), e );
      newgrid->putParent( e5 = incCurrElmNo(), e );
      newgrid->putParent( e6 = incCurrElmNo(), e ); 
      if ( n10 )
	newgrid->putParent( e7 = incCurrElmNo(), e );
      for( node2=2; node2<=6; node2++ ) 
	if( ! new_nodes( e, node2 ) ) 
	  break;
      switch ( node2 ) {
      case 2: // not node 6 => case 5 7 8 9 10
//	  cout<<e<<" 578910"<<endl;
	newgrid->putLoc2Glob1234( n4, n5, n10, n9  , e1 );
	if( edge_face2(e) == 5 ) {
	  newgrid->putLoc2Glob1234( n2, n3, n5, n9  , e2 );
	  newgrid->putLoc2Glob1234( n3, n5, n9, n10 , e3 );
        } else {
	  newgrid->putLoc2Glob1234( n2, n3, n5, n10 , e2 );
	  newgrid->putLoc2Glob1234( n10, n2, n9, n5 , e3 );
	}
	switch ( edge_face3(e) ) {
	case 3:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e4 );
	  newgrid->putLoc2Glob1234( n4, n5, n7, n10 , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n7, n8  , e6 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e7 );
	  break;
	case 4:
	  newgrid->putLoc2Glob1234( n8, n3, n5, n7  , e4 );
	  newgrid->putLoc2Glob1234( n3, n8, n5, n10 , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e6 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e7 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e4 );
	  newgrid->putLoc2Glob1234( n1, n10, n5, n7 , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e6 );
	  newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e7 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face3");
	}
	break;
      case 3: // not node 7 => case 5 6 8 9 10
//	  cout<<e<<" 568910"<<endl;
	newgrid->putLoc2Glob1234( n5, n4, n10, n8  , e1 );
	if ( edge_face3(e) == 4 ) {
	  newgrid->putLoc2Glob1234( n3, n1, n5, n8  , e2 );
	  newgrid->putLoc2Glob1234( n5, n3, n8, n10 , e3 );
	} else {
	  newgrid->putLoc2Glob1234( n3, n1, n5, n10 , e2 );
	  newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e3 );
	}
	switch ( edge_face2(e) ) {
	case 2:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e4 );
	  newgrid->putLoc2Glob1234( n4, n5, n6, n9  , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n6, n10 , e6 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e7 );
	  break;
	case 5:
	  newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e4 );
	  newgrid->putLoc2Glob1234( n9, n3, n5, n10 , e5 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e6 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e7 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e4 );
	  newgrid->putLoc2Glob1234( n10, n2, n5, n6 , e5 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e6 );
	  newgrid->putLoc2Glob1234( n2, n10, n5, n9 , e7 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face2");
	}
	break;
      case 4: // not node 8 => case 5 6 7 9 10
//	  cout<<e<<" 567910"<<endl;
	newgrid->putLoc2Glob1234( n5, n3, n7, n10  , e1 );
	if ( edge_face3(e) == 3 ) {
	  newgrid->putLoc2Glob1234( n1, n4, n5, n7  , e2 );
	  newgrid->putLoc2Glob1234( n7, n4, n5, n10 , e3 );
	} else {
	  newgrid->putLoc2Glob1234( n1, n4, n5, n10  , e2 );
	  newgrid->putLoc2Glob1234( n1, n10, n5, n7  , e3 );
	}
	switch ( edge_face2(e) ) {
	case 2:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e4 );
	  newgrid->putLoc2Glob1234( n6, n4, n5, n9  , e5 );
	  newgrid->putLoc2Glob1234( n4, n6, n5, n10 , e6 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e7 );
	  break;
	case 5:
	  newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e4 );
	  newgrid->putLoc2Glob1234( n9, n3, n5, n10 , e5 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e6 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e7 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e4 );
	  newgrid->putLoc2Glob1234( n10, n2, n5, n6 , e5 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e6 );
	  newgrid->putLoc2Glob1234( n2, n10, n5, n9 , e7 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face2");
	}
	break;
      case 5:   // not node 9 => case 5 6 7 8 10
//	  cout<<e<<" 567810"<<endl;
	newgrid->putLoc2Glob1234( n5, n3, n10, n6  , e1 );
	if ( edge_face2(e) == 2 ) {
	  newgrid->putLoc2Glob1234( n4, n2, n5, n6  , e2 );
	  newgrid->putLoc2Glob1234( n5, n4, n6, n10 , e3 );
	} else {
	  newgrid->putLoc2Glob1234( n4, n2, n5, n10  , e2 );
	  newgrid->putLoc2Glob1234( n10, n2, n5, n6  , e3 );
	}
	switch ( edge_face3(e) ) {
	case 3:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e4 );
	  newgrid->putLoc2Glob1234( n7, n4, n5, n10 , e5 );
	  newgrid->putLoc2Glob1234( n4, n7, n5, n8  , e6 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e7 );
	  break;
	case 4:
	  newgrid->putLoc2Glob1234( n3, n5, n8, n7  , e4 );
	  newgrid->putLoc2Glob1234( n5, n3, n8, n10 , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e6 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e7 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e4 );
	  newgrid->putLoc2Glob1234( n1, n10, n5, n7 , e5 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e6 );
	  newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e7 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face3");
	}
	break;
      case 6: // not node 10 => case 5 6 7 8 9
//	  cout<<e<<" 56789"<<endl;
	newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e1 );
	if ( edge_face2(e) == 2 ) {
	  newgrid->putLoc2Glob1234( n3, n4, n5, n6  , e2 );
	  newgrid->putLoc2Glob1234( n6, n4, n5, n9  , e3 );
	} else {
	  newgrid->putLoc2Glob1234( n3, n9, n5, n6  , e2 );
	  newgrid->putLoc2Glob1234( n3, n4, n5, n9  , e3 );
	}
	newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e4 );
	if ( edge_face3(e) == 3 ) {
	  newgrid->putLoc2Glob1234( n4, n7, n5, n8  , e5 );
	  newgrid->putLoc2Glob1234( n4, n3, n5, n7  , e6 );
	} else {
	  newgrid->putLoc2Glob1234( n4, n3, n5, n8  , e5 );
	  newgrid->putLoc2Glob1234( n8, n3, n5, n7  , e6 );
	}
	break;
      default: errorFP("GridFEAdT::refine3D",
		       "case 6 did not find second node");
      }
      break;
//----------------------------------------------------------------
    case 7 : // refine the element into 8 elements
//----------------------------------------------------------------
      newgrid->putParent( e1 = incCurrElmNo(), e );
      newgrid->putParent( e2 = incCurrElmNo(), e );
      newgrid->putParent( e3 = incCurrElmNo(), e );
      newgrid->putParent( e4 = incCurrElmNo(), e );
      newgrid->putParent( e5 = incCurrElmNo(), e );
      newgrid->putParent( e6 = incCurrElmNo(), e );
      newgrid->putParent( e7 = incCurrElmNo(), e );
      newgrid->putParent( e8 = incCurrElmNo(), e );
      
      if ( getRegularUniformRef() ) {
	// regular method cut of four tetrahedra, choice of diagonal
	// only used for uniform refinement with method 6
	newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e1 );
	newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e2 );
	newgrid->putLoc2Glob1234( n7, n6, n3, n10 , e3 );
	newgrid->putLoc2Glob1234( n8, n9, n10, n4 , e4 );
	  
	switch ( getShortestDiagonal(e,newgrid->coor) ) {
	case 1: //	case five2ten:
	  newgrid->putLoc2Glob1234( n6, n7, n5, n10 , e5 );
	  newgrid->putLoc2Glob1234( n5, n8, n10, n7 , e6 );
	  newgrid->putLoc2Glob1234( n8, n9, n5, n10 , e7 );
	  newgrid->putLoc2Glob1234( n5, n6, n10, n9 , e8 );
//	    cout<<e<<" 5-10"<<endl;
	  break;
	case 2: //	case six2eight:
	  newgrid->putLoc2Glob1234( n9, n8, n6, n5  , e5 );
	  newgrid->putLoc2Glob1234( n9, n8, n10, n6 , e6 );
	  newgrid->putLoc2Glob1234( n7, n6, n10, n8 , e7 );
	  newgrid->putLoc2Glob1234( n7, n6, n8, n5  , e8 );
//	    cout<<e<<" 6-8"<<endl;
	  break;
	case 3: //	case seven2nine:
	  newgrid->putLoc2Glob1234( n7, n6, n9, n5  , e5 );
	  newgrid->putLoc2Glob1234( n7, n6, n10, n9 , e6 );
	  newgrid->putLoc2Glob1234( n9, n8, n10, n7 , e7 );
	  newgrid->putLoc2Glob1234( n9, n8, n7, n5  , e8 );
//	    cout<<e<<" 7-9"<<endl;
	  break;
	default: errorFP("GridFEAdT:: refine3D","wrong diagonal");
	}
      } else { // bisection method
	switch ( edge_face2(e) ) {
	case 2:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e1 );
	  newgrid->putLoc2Glob1234( n4, n5, n6, n9  , e2 );
	  newgrid->putLoc2Glob1234( n5, n4, n6, n10 , e3 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e4 );
	  break;
	case 5:
	  newgrid->putLoc2Glob1234( n5, n3, n9, n6  , e1 );
	  newgrid->putLoc2Glob1234( n3, n5, n9, n10 , e2 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e3 );
	  newgrid->putLoc2Glob1234( n5, n2, n6, n9  , e4 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n3, n5, n6, n10 , e1 );
	  newgrid->putLoc2Glob1234( n10, n2, n5, n6 , e2 );
	  newgrid->putLoc2Glob1234( n4, n5, n10, n9 , e3 );
	  newgrid->putLoc2Glob1234( n2, n10, n5, n9 , e4 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face2");
	}
	switch ( edge_face3(e) ) {
	case 3:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e5 );
	  newgrid->putLoc2Glob1234( n4, n5, n7, n10 , e6 );
	  newgrid->putLoc2Glob1234( n5, n4, n7, n8  , e7 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e8 );
	  break;
	case 4:
	  newgrid->putLoc2Glob1234( n3, n5, n8, n7  , e5 );
	  newgrid->putLoc2Glob1234( n5, n3, n8, n10 , e6 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e7 );
	  newgrid->putLoc2Glob1234( n1, n5, n7, n8  , e8 );
	  break;
	case 6:
	  newgrid->putLoc2Glob1234( n5, n3, n7, n10 , e5 );
	  newgrid->putLoc2Glob1234( n1, n10, n5, n7 , e6 );
	  newgrid->putLoc2Glob1234( n5, n4, n10, n8 , e7 );
	  newgrid->putLoc2Glob1234( n10, n1, n5, n8 , e8 );
	  break;
	default: errorFP("GridFEAdT:: refine3D","illegal no in edge_face3");
	}
      }
      break; 
//------------------------------------------------------
    default : 
//------------------------------------------------------
      fatalerrorFP("GridFEAdT:: refine3D",
		   "\nMust be 1-7 new elements, now there are %d.",refcase);
    } //end of switch
  }

  pos_child( nel + 1 ) = getCurrElmNo() + 1; 

// copy the subdomain types for the new grid from the parent elements:
// material_type is redimed i GridFE.redim if onemat false
  if( ! onemat )
    for(e=1;e<=new_nel;e++)
      newgrid->material_type(e) = material_type(newgrid->getParent_eff(e));
  real x1,y1,z1, x2,y2,z2, x3,y3,z3, x4,y4,z4;
  if ( T10 ) {
    for( e=1; e<=new_nel; e++ ) {
// compute positions and boundary indicators for the new midpoint nodes
      if( getNoChildren( newgrid->getParent_eff( e ) ) > 1 ){
	// the element has at least one new node
	n1 = newgrid->loc2glob( e, 1 );
	n2 = newgrid->loc2glob( e, 2 );
	n3 = newgrid->loc2glob( e, 3 );
	n4 = newgrid->loc2glob( e, 4 );
	n5 = newgrid->loc2glob( e, 5 );
	n6 = newgrid->loc2glob( e, 6 );
	n7 = newgrid->loc2glob( e, 7 );
	n8 = newgrid->loc2glob( e, 8 );
	n9 = newgrid->loc2glob( e, 9 );
	n10= newgrid->loc2glob( e, 10);
	
	x1 = newgrid->coor(n1,1); 
	y1 = newgrid->coor(n1,2);
	z1 = newgrid->coor(n1,3);
	x2 = newgrid->coor(n2,1);
	y2 = newgrid->coor(n2,2);
	z2 = newgrid->coor(n2,3);
	x3 = newgrid->coor(n3,1);
	y3 = newgrid->coor(n3,2);
	z3 = newgrid->coor(n3,3);
	x4 = newgrid->coor(n4,1);
	y4 = newgrid->coor(n4,2); 
	z4 = newgrid->coor(n4,3);

	if( ! newgrid->coor(n5,1) ) { // to avoid computing things twice
	  newgrid->coor(n5,1) = (x1 + x2)/2;
	  newgrid->coor(n5,2) = (y1 + y2)/2;
	  newgrid->coor(n5,3) = (z1 + z2)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n1,b) == '1' && nind(n2,b) == '1' ) // ends set..
	      nind(n5,b) = '1'; // then set midpoint
	}
	if( ! newgrid->coor(n6,1) ) { 
	  newgrid->coor(n6,1) = (x2 + x3)/2;
	  newgrid->coor(n6,2) = (y2 + y3)/2;
	  newgrid->coor(n6,3) = (z2 + z3)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n2,b) == '1' && nind(n3,b) == '1' )
	      nind(n6,b) = '1';
	}
	if( ! newgrid->coor(n7,1) ) {
	  newgrid->coor(n7,1) = (x3 + x1)/2;
	  newgrid->coor(n7,2) = (y3 + y1)/2;
	  newgrid->coor(n7,3) = (z3 + z1)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n3,b) == '1' && nind(n1,b) == '1' )
	      nind(n7,b) = '1';
	}
	if( ! newgrid->coor(n8,1) ) {
	  newgrid->coor(n8,1) = (x4 + x1)/2;
	  newgrid->coor(n8,2) = (y4 + y1)/2;
	  newgrid->coor(n8,3) = (z4 + z1)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n4,b) == '1' && nind(n1,b) == '1' )
	      nind(n8,b) = '1';
	}
	if( ! newgrid->coor(n9,1) ) {
	  newgrid->coor(n9,1) = (x4 + x2)/2;
	  newgrid->coor(n9,2) = (y4 + y2)/2;
	  newgrid->coor(n9,3) = (z4 + z2)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n4,b) == '1' && nind(n2,b) == '1' )
	      nind(n9,b) = '1';
	}
	if( ! newgrid->coor(n10,1) ) {
	  newgrid->coor(n10,1) = (x4 + x3)/2;
	  newgrid->coor(n10,2) = (y4 + y3)/2;
	  newgrid->coor(n10,3) = (z4 + z3)/2;
	  for( b=1; b<=getNoBoInds(); b++ )
	    if( nind(n4,b) == '1' && nind(n3,b) == '1' )
	      nind(n10,b) = '1';
	}
      }
    }
  }

  removeNewNodeInfo(); // remove the data showing the marked edges
//  edge_face2.redim(0);
//  edge_face3.redim(0);
  newgrid->setParentInfo( true );
  setChildrenInfo( true );

#ifdef SAFETY_CHECKS
  if( ! newgrid->checkElementOrientation() )
    s_o<<"Wrong orientation."<<endl;
  if( !checkChildrenInfo() ) 
    s_o<<"children info failed."<<endl;
  if( !newgrid->checkParentInfo() ) 
    s_o<<"parent info failed."<<endl;
#endif
    
  return true; // everything went fine
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeInteriorIndicators()
//-----------------------------------------------------------------------------
{
  if( ! isNeighborsComputed() ) // Is the element-element info computed?
    initNeighbor(); 

  VecSimple(int) inner( nno );
  inner.fill( 1 ); // set boundary nodes to zero in the inner vector

  int i,b,e;
  switch ( nsd ) {
  case 1:
    for( e=1; e<=nel; e++ ) {
      if( ! getElmNeighbor(e,1) ) inner(loc2glob(e,1)) = 0;
      if( ! getElmNeighbor(e,2) ) inner(loc2glob(e,2)) = 0;
    }
    break;
  case 2:
    for( e=1; e<=nel; e++ ) {
      if( ! getElmNeighbor(e,1) )
	inner(loc2glob(e,2)) = inner(loc2glob(e,1)) = 0;
      if( ! getElmNeighbor(e,2) )
	inner(loc2glob(e,3)) = inner(loc2glob(e,2)) = 0;
      if( ! getElmNeighbor(e,3) )
	inner(loc2glob(e,3)) = inner(loc2glob(e,1)) = 0;
    }
    if( "ElmT6n2D" == elm_tp )
      for( e=1; e<=nel; e++ ) {
	if( ! getElmNeighbor(e,1) ) inner(loc2glob(e,4)) = 0;
	if( ! getElmNeighbor(e,2) ) inner(loc2glob(e,5)) = 0;
	if( ! getElmNeighbor(e,3) ) inner(loc2glob(e,6)) = 0;
      }
    break;
  case 3:
    for( e=1; e<=nel; e++ ) {
      if( ! getElmNeighbor(e,1) )
	inner(loc2glob(e,4)) = inner(loc2glob(e,2)) = inner(loc2glob(e,1)) = 0;
      if( ! getElmNeighbor(e,2) )
	inner(loc2glob(e,4)) = inner(loc2glob(e,3)) = inner(loc2glob(e,2)) = 0;
      if( ! getElmNeighbor(e,3) )
	inner(loc2glob(e,4)) = inner(loc2glob(e,3)) = inner(loc2glob(e,1)) = 0;
      if( ! getElmNeighbor(e,4) )
	inner(loc2glob(e,3)) = inner(loc2glob(e,2)) = inner(loc2glob(e,1)) = 0;
    }
    if( "ElmT10n3D" == elm_tp ) 
      for( e=1; e<=nel; e++ ) {
	if( ! getElmNeighbor(e,1) )
	  inner(loc2glob(e,9)) = inner(loc2glob(e,8)) = inner(loc2glob(e,5))=0;
	if( ! getElmNeighbor(e,2) )
	  inner(loc2glob(e,10))= inner(loc2glob(e,9)) = inner(loc2glob(e,6))=0;
	if( ! getElmNeighbor(e,3) )
	  inner(loc2glob(e,10))= inner(loc2glob(e,8)) = inner(loc2glob(e,7))=0;
	if( ! getElmNeighbor(e,4) )
	  inner(loc2glob(e,7)) = inner(loc2glob(e,6)) = inner(loc2glob(e,5))=0;
      }
    break;
  default: 
    errorFP("GridFEAdT:: removeInteriorIndicators","nsd=%d",nsd);
  }

  MatSimple(char)& indicators = boundaryData().indicatorAccess();
// check that size of indicators really is (nno,nbind) ?

  for ( i=1; i <= nno; i++ )
    if( inner(i) )
      for ( b = 1; b <= getNoBoInds(); b++)
	indicators(i,b) = '0';
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: isElmNeighborConsistent ()
//-----------------------------------------------------------------------------
// Check if neighbor element numbers are consistent, i.e., a neighbor to
// an element must have the element as its neighbor. return true if consistent 
{
// ksa: could extend this routine to check that the nodes situated on a 
// face/edge also exists in the neighbor element. Not yet done.
  if( ! isNeighborsComputed() ){ // is the neighbor info created?
    errorFP("GridFEAdT:: isElmNeighborConsistent",
	      "Neighbor info is not computed");
    return false;
  }
  int e,n,elm;
  int nfaces = nsd+1; // number of sides 2 in 1D, 3 in 2D and 4 in 3D
  for( e=1; e<=nel; e++ )
    for( n=1; n<=nfaces; n++ )
      if((elm = elm_neigh( e, n ))) // elm is zero if boundary i.e. no neighbor
	if( ! isElmNeighbor( elm, e ) ) {
	  errorFP("GridFEAdT:: isElmNeighborConsistent",
		  "\nElement %d and face/edge/node %d, neighbor is %d\n"
		  "parents are %d and %d."
		  ,e,n,elm,getParent(e),getParent(elm));
	  return false; 
	}
  return true;
}

//-----------------------------------------------------------------------------
void GridFEAdT::initNeighbor (bool alwaysSortEdgesIn2D)
//-----------------------------------------------------------------------------
// Initiate the elm_neigh structure so that it contains the neighbors elements.
{
  elm_neigh.redim( nel, nsd+1 ); // dimensioning of neighbor data structure    
  elm_neigh.fill(0); // 0 means boundary edge

  switch ( nsd ) {
  case 1: initNeighbor1D(); break;
  case 2: initNeighbor2D( alwaysSortEdgesIn2D ); break;
  case 3: initNeighbor3D(); break;
  default: errorFP("GridFEAdT::initNeighbor","nsd = %d not implemented", nsd);
  }
  setNeighborsComputed( true );

#ifdef SAFETY_CHECKS
  if( ! isElmNeighborConsistent() )
    fatalerrorFP("GridFEAdT:: initNeighbor", "Not consistent.");
  if( !checkElementOrientation() )
    fatalerrorFP("GridFEAdT:: initNeighbor3D", "Wrong orientation.");
#endif
}

//-----------------------------------------------------------------------------
void GridFEAdT::initNeighbor1D ()
//-----------------------------------------------------------------------------
// Initiate the elm_neigh structure so that it contains the neighbors elements.
{
  int e,k,ne;

// Find the neighbor element which have one node in common
// with a given side of element e. If no element is found this
// is a boundary edge.


// should check a standard case 

// elm_neigh  1 2 3 ...  nel-1 nel
// --------------------------------   
//            0 1 2 ... nel-2 nel-1
//            2 3 4 ...  nel    0 

// Very efficient way to compute element-element info.
// The n2e vector is initially zero. In the loop over
// the elements each edge is accessed at most twice.
// The first time, the number of the neighbor element
// is put in n2e, the second time the elm_neigh is set
// correctly. The edges which only are accessed once are 
// left as zeros in elm_neigh as they should.

  VecSimple(int) n2e(nno);
  n2e.fill( 0 );

    // element-element neighbor info.:

  int secondnode = ("ElmB2n1D" == elm_tp) ? 2:3;

  for( e=1; e<=nel; e++ ){
    ne = n2e( k = loc2glob( e, 1 ) );
    if( ne ) {
      putElmNeighbor( e, 2, ne );
      putElmNeighbor( ne, 1, e );
    }
    else
      n2e( k ) = e;
    ne = n2e( k = loc2glob( e, secondnode ) );
    if( ne ) {
      putElmNeighbor( e, 1, ne );
      putElmNeighbor( ne, 2, e );
    }
    else
      n2e( k ) = e;
  }
}


//-----------------------------------------------------------------------------
void GridFEAdT::initNeighbor2D ( bool alwaysSortEdgesIn2D )
//-----------------------------------------------------------------------------
// Initiate the elm_neigh structure so that it contains the neighbors elements.
{
  CPUclock cpu;
  cpu.initTime();

  int n,e,j,k,n1,n2,ne;

// Reorder the element-node info so that the first edge in the
// element is longest, (the edge between node 1 and 2.
// warning: the ordering may not be correct after that
// the nodes are moved, e.g., using the move or smoothing functions
// Normally if no nodes has been moved in a refined grid it is not
// necessary to sort the order, they are already correct. Thereof
// the parameter alwaysSortEdgesIn2D

  if( isThisCoarsestGrid () || alwaysSortEdgesIn2D ) { 
    if ( "ElmT3n2D" == elm_tp ) {
      for ( e=1; e<=nel; e++ ) {
	n = getLongestEdgeNo( e );
	if( n == 1 ) // if true then correct order
	  continue;
	if( n == 2 ) 
	  putLoc2Glob123( loc2glob(e,2), loc2glob(e,3), loc2glob(e,1), e);
	else     
	  putLoc2Glob123( loc2glob(e,3), loc2glob(e,1), loc2glob(e,2), e);
      }
    } else { // Quadratic elements
      for ( e=1; e<=nel; e++ ){
	n = getLongestEdgeNo( e );
	if( n == 1 ) // if true then correct order
	  continue;
	if( n == 2 ) {
	  putLoc2Glob123( loc2glob(e,2), loc2glob(e,3), loc2glob(e,1), e);
	  putLoc2Glob456( loc2glob(e,5), loc2glob(e,6), loc2glob(e,4), e);
	} else {
	  putLoc2Glob123( loc2glob(e,3), loc2glob(e,1), loc2glob(e,2), e);
	  putLoc2Glob456( loc2glob(e,6), loc2glob(e,4), loc2glob(e,5), e);
	}
      }
    }
  }
//cout<<"\ninit  after edge ordering, cpu-time = " << cpu.getInterval()<<endl;

  if( "ElmT3n2D" == elm_tp ) {

// Find the neighbor elements which have two nodes in common with a given 
// side of element e. If no element is found this is a boundary edge.

    // first compute node-to-element information
    // Key observation: An element can be added to a given node at most once
    // hence no sort is needed for the node-to-element (n2e) case.
    VecSimple(int) countn2e( nno );
    VecSimple(int) n2ejcol(nel*3); // each element is added to exactly 3 nodes
    VecSimple(int) n2eirow(nno+1);
    countn2e.fill(0);
    for ( e=1; e<=nel; e++ ) { // count the number of elements a node is in
      countn2e( loc2glob(e,1) )++;
      countn2e( loc2glob(e,2) )++;
      countn2e( loc2glob(e,3) )++;
    }
// set up the n2eirow which tells us where next node begins
    int prev = n2eirow(1) = 1;
    for (n=1; n<=nno; n++ )  prev = n2eirow( n+1 ) = countn2e( n ) + prev;
    countn2e.fill(0);
    for ( e=1; e<=nel; e++ ){
      ne = loc2glob(e,1);
      n2ejcol ( n2eirow( ne ) + countn2e( ne )++ ) = e;
      ne = loc2glob(e,2);
      n2ejcol ( n2eirow( ne ) + countn2e( ne )++ ) = e;
      ne = loc2glob(e,3);
      n2ejcol ( n2eirow( ne ) + countn2e( ne )++ ) = e;
    }
//cout << "\ninit after computing n2e, cpu-time = " << cpu.getInterval()<<endl;

    // compute element-element neighbor info.:

// it is faster to go backwards due to the fact that the elements in n2e
// are ordered starting with lowest element number and increases.
// Therefore the neighbor element will be found earlier.

    for( e=nel; e>=1; e-- ) 
      for( j=1; j<=3; j++ ){
	if( ! getElmNeighbor(e,j) ) { // if not yet set
	  n1 = loc2glob( e, j ); // first node
	  n2 = loc2glob( e, j==3 ? 1 : j+1 );
	  int kstop = n2eirow(n1+1);
	  for( k=n2eirow(n1); k<kstop; k++ ) { // search for neighbor element
	    ne = n2ejcol(k);
	    if ( ne != e ) { // find second common node
	      if ( loc2glob (ne, 1) == n2 ) {
		putElmNeighbor(  e, j, ne );
		putElmNeighbor( ne, 1,  e );
		k = kstop; // to break the k-loop
	      }
	      else if ( loc2glob (ne, 2) == n2 ) {
		putElmNeighbor(  e, j, ne );
		putElmNeighbor( ne, 2,  e );
		k = kstop; // to break the k-loop
	      }
	      else if ( loc2glob (ne, 3) == n2 ) {
		putElmNeighbor(  e, j, ne );
		putElmNeighbor( ne, 3,  e );
		k = kstop; // to break the k-loop
	      }
	    }
	  }
	}
      }
//  cout << "\ninit finished e2e, cpu-time = " << cpu.getInterval()<<endl;
  } else {  // if( "ElmT6n2D" == elm_tp )

// Very efficient way to compute element-element info for quadratic
// elements, based on that the nodes on mid of edges is only shared
// by two elements.
// The n2e vector is initially zero. In the loop over the elements each 
// edge is accessed at most twice. The first time, the number of the 
// neighbor element is put in n2e, the second time the elm_neigh is set
// correctly. The edges accessed only once are left as zeros in elm_neigh 
// as they should, (they are boundary edges).

    VecSimple(int) n2e(nno);
    n2e.fill( 0 );
    for( e=1; e<=nel; e++ )
      for( j=4; j<=6; j++ )
	if( (ne = n2e( k = loc2glob( e, j ) )) ) {
	  putElmNeighbor( e, j-3, ne );
	  if( loc2glob( ne, 4 ) == k )       putElmNeighbor( ne, 1, e );
	  else if( loc2glob( ne, 5 ) == k )  putElmNeighbor( ne, 2, e );
	  else if( loc2glob( ne, 6 ) == k )  putElmNeighbor( ne, 3, e );
	}
	else
	  n2e( k ) = e;
  }
}


//-----------------------------------------------------------------------------
int GridFEAdT:: firstOf256 ( VecSimple(int)& order )
//-----------------------------------------------------------------------------
{ 
// order contains order of longest edges where edge 1 is longest
// find which of edges 2, 5 and 6 are longest
  int ed = order(2);
  if( ed==2 || ed==5 || ed==6 ) return ed;
  ed = order(3);
  if( ed==2 || ed==5 || ed==6 ) return ed;
  ed = order(4);
  if( ed==2 || ed==5 || ed==6 ) return ed;
  errorFP("GridFEAdT:: firstOf256","should have returned");
  order.print(s_o);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: firstOf346( VecSimple(int)& order )
//-----------------------------------------------------------------------------
{
// order contains order of longest edges where edge 1 is longest
// find which of edges 3, 4 and 6 are longest
  int ed = order(2);
  if( ed==3 || ed==4 || ed==6 ) return ed;
  ed = order(3);
  if( ed==3 || ed==4 || ed==6 ) return ed;
  ed = order(4);
  if( ed==3 || ed==4 || ed==6 ) return ed;
  errorFP("GridFEAdT:: firstOf346","should have returned");
  order.print(s_o);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: firstOf256( VecSimple(int)& order,
			     const int ed2, const int ed5, const int ed6)
//-----------------------------------------------------------------------------
{
  int ed = order(2);
  if( ed==ed2 ) return 2;
  if( ed==ed5 ) return 5;
  if( ed==ed6 ) return 6;
  ed = order(3);
  if( ed==ed2 ) return 2;
  if( ed==ed5 ) return 5;
  if( ed==ed6 ) return 6;
  ed = order(4);
  if( ed==ed2 ) return 2;
  if( ed==ed5 ) return 5;
  if( ed==ed6 ) return 6;
  errorFP("GridFEAdT:: firstOf256","should have returned");
  order.print(s_o);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: firstOf346( VecSimple(int)& order,
			     const int ed3, const int ed4, const int ed6)
//-----------------------------------------------------------------------------
{
  int ed = order(2);
  if( ed==ed3 ) return 3;
  if( ed==ed4 ) return 4;
  if( ed==ed6 ) return 6;
  ed = order(3);
  if( ed==ed3 ) return 3;
  if( ed==ed4 ) return 4;
  if( ed==ed6 ) return 6;
  ed = order(4);
  if( ed==ed3 ) return 3;
  if( ed==ed4 ) return 4;
  if( ed==ed6 ) return 6;
  errorFP("GridFEAdT:: firstOf346","should have returned");
  order.print(s_o);
  return 0;
}

//------------------------------------------------------------------------
void GridFEAdT:: initNeighbor3D ()
//------------------------------------------------------------------------
// Initiate the elm_neigh structure so that it contains the numbers of
// the neighbors elements on the other side of the element faces.
{
// Reorder the element-node info so that the first edge is longest,
// warning: this could be changed by the move function

  CPUclock cpu;
  cpu.initTime();

  int n,e,n1,n2,n3,n4,n5,n6,n7,n8,n9,n10,ne;

  bool T10 = ( "ElmT10n3D" == elm_tp ) ? true : false;

  edge_face2.redim(nel);
  edge_face3.redim(nel);
  lengthsscratch.redim(6); // global arrays used in getEdgeLengthOrder
  orderscratch.redim(6);

  for ( e=1; e<=nel; e++ ) {
    getEdgeLengthOrder( orderscratch, e );
    n = orderscratch(1);
    if( n != 1) { // if n is 1 then correct order else reorder
      n1 = loc2glob( e, 1 );  n2 = loc2glob( e, 2 );
      n3 = loc2glob( e, 3 );  n4 = loc2glob( e, 4 );
      if ( T10 ) {
	n5 = loc2glob( e, 5 );	n6 = loc2glob( e, 6 );
	n7 = loc2glob( e, 7 );	n8 = loc2glob( e, 8 );
	n9 = loc2glob( e, 9 );	n10= loc2glob( e, 10);
      }
    }
    switch ( n ) { // two alternatives, which to choose?
    case 1:
      edge_face2(e) = firstOf256(orderscratch);
      edge_face3(e) = firstOf346(orderscratch);
      break;
    case 2:
      putLoc2Glob1234 ( n2, n3, n1, n4, e );   // now prev edge 2 is first
      if(T10)
	putLoc2Glob5678910 (n6, n7, n5, n9, n10, n8, e );
      edge_face2(e) = firstOf256(orderscratch,3,6,4);
      edge_face3(e) = firstOf346(orderscratch,1,5,4);
      break;
    case 3: 
      putLoc2Glob1234 ( n3, n1, n2, n4, e );	
      if(T10)
	putLoc2Glob5678910 ( n7, n5, n6, n10, n8, n9, e );
      edge_face2(e) = firstOf256(orderscratch,1,4,5);
      edge_face3(e) = firstOf346(orderscratch,2,6,5);
      break;
    case 4: 
      putLoc2Glob1234 ( n1, n4, n2, n3, e );	
      if(T10)
	putLoc2Glob5678910 ( n8, n9, n5, n7, n10, n6, e );
      edge_face2(e) = firstOf256(orderscratch,5,6,2);
      edge_face3(e) = firstOf346(orderscratch,1,3,2);
      break; 
    case 5: 
      putLoc2Glob1234 ( n2, n4, n3, n1, e );	
      if(T10)
	putLoc2Glob5678910 ( n9, n10, n6, n5, n8, n7, e );
      edge_face2(e) = firstOf256(orderscratch,6,4,3);
      edge_face3(e) = firstOf346(orderscratch,2,1,3);
      break; 
    case 6: 
      putLoc2Glob1234 ( n3, n4, n1, n2, e );	
      if(T10)
	putLoc2Glob5678910 ( n10, n8, n7, n6, n9, n5, e );
      edge_face2(e) = firstOf256(orderscratch,4,5,1);
      edge_face3(e) = firstOf346(orderscratch,3,2,1);
      break;
    default: errorFP("GridFEAdT::initNeighbor3D","illegal value %d",n);
    }
  }
//  cout << "\ninit0   cpu-time = " << cpu.getInterval() <<endl;

// Given face of element e, find the neighbor elements which have 
// three nodes in common with the face. If no such element is found this
// is a boundary face.

  additional_info.initNeighbor(*this,true,false,false);
  NeighborFE& neighbor = additional_info.neighbor;
  if( neighbor.nodeSize() == 0 )
    neighbor.init(*this,true,false,false);

//  cout << "\ninit1   cpu-time = " << cpu.getInterval() <<endl;

  if( "ElmT4n3D" == elm_tp ){

// face1 has nodes 1 2 4
// face2 has nodes 2 3 4
// face3 has nodes 1 3 4
// face4 has nodes 1 2 3

    int ne1,ne2;
    for( e=nel; e>=1; e-- ){ // more efficient with decreasing elements
      n1 = loc2glob( e, 1 );
      n2 = loc2glob( e, 2 );
      n3 = loc2glob( e, 3 );
      n4 = loc2glob( e, 4 );
      if( ! getElmNeighbor(e,1) )
	if( ! getElmNeighbor(e,4) ){
	  getElmsAtFaces(n1,n2,n4,n3,e,ne1,ne2);// case face 1 and 4 not done
	  if( ne1 ){
	    n = getFaceNo(n1,n2,n4,ne1);
	    putElmNeighbor(  e, 1, ne1 );
	    putElmNeighbor( ne1, n,  e );
	  }
	  if( ne2 ){
	    n = getFaceNo(n1,n2,n3,ne2);
	    putElmNeighbor(  e, 4, ne2 );
	    putElmNeighbor( ne2, n,  e );
	  }
	}
	else {
	  ne = getElmAtFace(n1,n2,n4,e); // case face 1
	  if( ne ){
	    n = getFaceNo(n1,n2,n4,ne);
	    putElmNeighbor(  e, 1, ne );
	    putElmNeighbor( ne, n,  e );
	  }
	}
      else if( !getElmNeighbor(e,4) ) {
	ne = getElmAtFace(n1,n2,n3,e);    // case face 4
	if( ne ){
	  n = getFaceNo(n1,n2,n3,ne);
	  putElmNeighbor(  e, 4, ne );
	  putElmNeighbor( ne, n,  e );
	}
      }

      if( !getElmNeighbor(e,2) )
	if( !getElmNeighbor(e,3) ){
	  getElmsAtFaces(n3,n4,n2,n1,e,ne1,ne2);
	  if( ne1 ){
	    n = getFaceNo(n3,n4,n2,ne1);
	    putElmNeighbor(  e, 2, ne1 );
	    putElmNeighbor( ne1, n,  e );
	  }
	  if( ne2 ){
	    n = getFaceNo(n3,n4,n1,ne2);
	    putElmNeighbor(  e, 3, ne2 );
	    putElmNeighbor( ne2, n,  e );
	  }
	}
	else {
	  ne = getElmAtFace(n2,n3,n4,e);
	  if( ne ){
	    n = getFaceNo(n2,n3,n4,ne);
	    putElmNeighbor(  e, 2, ne );
	    putElmNeighbor( ne, n,  e );
	  }
	}
      else if( !getElmNeighbor(e,3) ){
	ne = getElmAtFace(n1,n3,n4,e);
	if( ne ){
	  n = getFaceNo(n1,n3,n4,ne);
	  putElmNeighbor(  e, 3, ne );
	  putElmNeighbor( ne, n,  e );
	}
      }
    }
  }

//  cout << "\ninit2   cpu-time = " << cpu.getInterval() <<endl;

  if( "ElmT10n3D" == elm_tp ){

// face1 has nodes 1 2 4 and midnodes 5 8 9
// face2 has nodes 2 3 4 and midnodes 6 9 10
// face3 has nodes 1 3 4 and midnodes 7 8 10
// face4 has nodes 1 2 3 and midnodes 5 6 7

    for( e=nel; e>=1; e-- ){ 
      n1 = loc2glob( e, 1 );
      n2 = loc2glob( e, 2 );
      n3 = loc2glob( e, 3 );
      n4 = loc2glob( e, 4 );
      bool safe=false;
      if( !getElmNeighbor(e,1) ) {     // case face 1 not yet done
	ne = getElmAtFace(loc2glob(e,5),loc2glob(e,8),e); // find  element
	if( ne ){                      // ne is 0 if it is a boundary face
	  n = getFaceNo(n1,n2,n4,ne,safe);  // find the face number
	  putElmNeighbor(  e, 1, ne ); // put found elm in neighbor array
	  putElmNeighbor( ne, n,  e );
	}
      }
      if( !getElmNeighbor(e,2) ) {     // case face 2 not yet done
	ne = getElmAtFace(loc2glob(e,6),loc2glob(e,9),e); 
	if( ne ){                     
	  n = getFaceNo(n2,n3,n4,ne,safe); 
	  putElmNeighbor(  e, 2, ne );
	  putElmNeighbor( ne, n,  e );
	}
      }
      if( !getElmNeighbor(e,3) ) {     // case face 3 not yet done
	ne = getElmAtFace(loc2glob(e,7),loc2glob(e,8),e);
	if( ne ){                     
	  n = getFaceNo(n1,n3,n4,ne,safe); 
	  putElmNeighbor(  e, 3, ne );
	  putElmNeighbor( ne, n,  e );
	}
      }
      if( !getElmNeighbor(e,4) ) {     // case face 4 not yet done
	ne = getElmAtFace(loc2glob(e,5),loc2glob(e,6),e);   
	if( ne ){                     
	  n = getFaceNo(n1,n2,n3,ne,safe); 
	  putElmNeighbor(  e, 4, ne );
	  putElmNeighbor( ne, n,  e );
	}
      }
    }
  }
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: isElmNeighbor( int e, int elnei )
//-----------------------------------------------------------------------------
// Does element e have element elnei as a neighbor?
{
  if( !isNeighborsComputed() ) { // Is the element-element info computed?
    warningFP("GridFEAdT::isElmNeighbor","Element neighbors not computed");
    return false; 
  }
  if( !e ) {
    warningFP("GridFEAdT::isElmNeighbor","Element number is 0");
    return false; 
  }
  int nof = nsd+1; 
  for( int i=1; i<=nof; i++ )
    if( getElmNeighbor(e,i) == elnei ) return true;
  return false; // no hit was found
}

//-----------------------------------------------------------------------------
void GridFEAdT:: initNumbering() 
//-----------------------------------------------------------------------------
{
  if( !isNeighborsComputed() ) // Is the element-element info computed?
    initNeighbor();  // if not compute it
  if( "ElmT4n3D" == elm_tp )
    new_nodes.redim( nel, 6 );
  else if( "ElmT3n2D" == elm_tp)
    new_nodes.redim( nel, 3 );
  else if( "ElmT6n2D" == elm_tp )
    new_nodes.redim( nel, 6 );
  else if( "ElmT10n3D" == elm_tp )
    new_nodes.redim( nel, 31 ); // 6 old 12 on edges 12 on faces 1 internal
  else if( "ElmB2n1D" == elm_tp || "ElmB3n1D" == elm_tp)
    new_nodes.redim( nel, 1 ); 
  else 
    fatalerrorFP("GridFEAdT:: initNumbering"," %s not implemented", 
		 elm_tp.chars());
  new_nodes.fill(0);
  setNewNodesSet( true );
  setCurrNodeNo(nno); 
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: markElementsBool( const VecSimple(bool)& ele )
//-----------------------------------------------------------------------------
// ele is true for the elements to be refined
{
  initNumbering();
  int e;
  if ( nsd==1 ) { //in 1D node numbering is simple, all new nodes inside elm
    int refm = getRefinementMethod();
    for( e=1; e<=nel; e++ ) if( ele(e) ) new_nodes(e,1) = refm;
    return true;
  }

  stackcounter=1;
  if( ele.size() != nel )
    errorFP("GridFEAdT:: markElementsBool",
	    "wrong size of vector (%d)",ele.size());
  else // in 2D and 3D use recursive method
    for( e=1; e<=nel; e++ ) if( ele( e ) ) markAllEdgesRec( e );

  elmstack.redim(0);
  ednostack.redim(0);

  extraNumberingForT10 (); // add nodes on faces

  return true;
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: markAllElements()
//-----------------------------------------------------------------------------
// ele is true for the elements to be refined
{
  initNumbering();

  if ( nsd==1 ) {
    new_nodes.fill( getRefinementMethod() );
    return true;
  }

  int e;
  stackcounter=1;
  if( nsd==2 && getRefinementMethod() == 3 )// all edges refined
    for( e=1; e<=nel; e++ ){
      if( ! new_nodes(e,1) ) numberingNewNodes( e, 1 );
      if( ! new_nodes(e,2) ) numberingNewNodes( e, 2 );
      if( ! new_nodes(e,3) ) numberingNewNodes( e, 3 );
    }
  else if( nsd==3 && getRefinementMethod() == 6 )// all edges refined
    for( e=1; e<=nel; e++ ){
      if( ! new_nodes(e,1) ) { numberingNewNodes( e, 1 ); stackcounter=1; }
      if( ! new_nodes(e,2) ) { numberingNewNodes( e, 2 ); stackcounter=1; }
      if( ! new_nodes(e,3) ) { numberingNewNodes( e, 3 ); stackcounter=1; }
      if( ! new_nodes(e,4) ) { numberingNewNodes( e, 4 ); stackcounter=1; }
      if( ! new_nodes(e,5) ) { numberingNewNodes( e, 5 ); stackcounter=1; }
      if( ! new_nodes(e,6) ) { numberingNewNodes( e, 6 ); stackcounter=1; }
    }
  else
    for( e=1; e<=nel; e++ )
      markAllEdgesRec( e );

  elmstack.redim(0);
  ednostack.redim(0);

  extraNumberingForT10();

  return true;
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: markElements( const VecSimple(int)& ele )
//-----------------------------------------------------------------------------
// If size of ele is nel, ele is nonzero for the elements to be refined
// otherwise ele contains numbers of the elements to be refined.
{
//DbP("");
  initNumbering();
  int e;
  if ( nsd==1 ) { 
// in 1D new_nodes contain the number of new elements caused by
// refinement of the element
    for( e=1; e<=nel; e++ )
      new_nodes(e,1) = ele(e);
    return true;
  }
//DbP("");

  stackcounter=1;
//DbP("");
if( ele.size() != nel )   {
//DbP("");
    errorFP("GridFEAdT:: markElements","wrong size of vector");
//DbP("");
}
  else {             // ele is nonzero for the elements to be marked 
//DbP("");
    for( e=1; e<=nel; e++ ) {
//DbP("");
      markAllEdgesRec( e, ele( e ) );
//DbP("");
    }
}
//DbP("");

  elmstack.redim(0);
  ednostack.redim(0);
//DbP("");

  extraNumberingForT10();
//DbP("");

  return true;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: extraNumberingForT10()
//-----------------------------------------------------------------------------
{
  if(nne != 10) 
    return; // if not ElmT10n3D

  errorFP("GridFEAdT:: extraNumberingForT10","not implemented yet");
}


//-----------------------------------------------------------------------------
void GridFEAdT:: markEdgeRec( int e, int edge_no )
//-----------------------------------------------------------------------------
// numbering of node at edge with edge_no in element e, note the recursion
{ //DbP("");
  if( ! new_nodes( e, edge_no ) ) { // if not marked before
//DbP("");
    int enei;
////    if ( nsd==2 ) {
////      numberingNewNodes( e, edge_no );
////      if( (enei = getElmNeighbor( e, edge_no )) ) // check if neighbor
////	markEdgeRec( enei, 1 ); // is longest edge marked recursive call  
////    }
    if ( nsd==2 ) {
      numberingNewNodes( e, edge_no );
      if( (enei = getElmNeighbor( e, edge_no )) ) { // check if neighbor
	markEdgeRec( enei, 1 ); // is longest edge marked recursive call  
	if( true ) {
	  // compute number of refined edges in the neighbor
	  int refedges = 1; // longest edge is always refined
	  if (new_nodes( enei, 2 ) ) refedges++;
	  if (new_nodes( enei, 3 ) ) refedges++;
	  if( refedges == 2 )
	    markEdgeRec( enei, getSecondLongestEdgeNo1( enei ) );
	}
      }
    }
    else { //DbP(""); // nsd == 3 
//  find the tetrahedras which share the same edge
      int start = stackcounter;
 //DbP(""); 
      numberingNewNodes( e, edge_no ); // stackcounter is increased
//DbP("");


      for( int i=start; i<stackcounter; i++ ) {
	enei = elmstack(i); // elm number of element neighbor to edge
//cout << i << " " << enei << " " << ednostack(i) << endl;
//cout << "2: " << edge_face2.size() << " 3: " <<edge_face3.size() << endl;
	switch ( ednostack(i) ) {
	case 1: // already numbered
	  break;
	case 2: 
	case 5: 
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face2(enei) ); // and longest of 2, 5 and 6
	  break;
	case 3:
	case 4:
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face3(enei) ); // and longest of 3, 4 and 6
	  break;
	case 6:
//DbP("");
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face2(enei) ); // and longest of 2, 5 and 6,
	  markEdgeRec( enei, edge_face3(enei) ); // and longest of 3, 4 and 6
	  break;
	default:
	  errorFP("GridFEAdT:: markEdgeRec",
		  "illegal edge number %d in element %d",
		  ednostack(i),elmstack(i));
	} 
      }
      stackcounter = start; // set stackcounter to the previous value
    }
  }
//DbP("");  
}

/*
//-----------------------------------------------------------------------------
void GridFEAdT:: markEdgeRec( int e, int edge_no )
//-----------------------------------------------------------------------------
// numbering of node at edge with edge_no in element e, note the recursion
{ 
  if( ! new_nodes( e, edge_no ) ) { // if not marked before
    int enei;
    if ( nsd==2 ) {
      numberingNewNodes( e, edge_no );
      if( (enei = getElmNeighbor( e, edge_no )) ) // check if neighbor
	markEdgeRec( enei, 1 ); // is longest edge marked recursive call  
    }
    else { // nsd == 3 
//  find the tetrahedras which share the same edge
      int start = stackcounter;
      numberingNewNodes( e, edge_no ); // stackcounter is increased
      for( int i=start; i<stackcounter; i++ ) {
	enei = elmstack(i); // elm number of element neighbor to edge
	switch ( ednostack(i) ) {
	case 1: // already numbered
	  break;
	case 2: 
	case 5: 
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face2(enei) ); // and longest of 2, 5 and 6
	  break;
	case 3:
	case 4:
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face3(enei) ); // and longest of 3, 4 and 6
	  break;
	case 6:
	  markEdgeRec( enei, 1 );                // number longest in element,
	  markEdgeRec( enei, edge_face2(enei) ); // and longest of 2, 5 and 6,
	  markEdgeRec( enei, edge_face3(enei) ); // and longest of 3, 4 and 6
	  break;
	default:
	  errorFP("GridFEAdT:: markEdgeRec",
		  "illegal edge number %d in element %d",
		  ednostack(i),elmstack(i));
	} 
      }
      stackcounter = start; // set stackcounter to the previous value
    }
  }
}
*/

//-----------------------------------------------------------------------------
void GridFEAdT:: markAllEdgesRec( int e )
//-----------------------------------------------------------------------------
// recursive numbering of edges of element e
{ 
  if ( nsd==2 ) {
    switch ( getRefinementMethod() ){
    case 1 : // mark only the longest edge of the element
      markEdgeRec( e, 1 );
      break;
    case 2 : // mark the two longest edges of the element
      markEdgeRec( e, 1 );
      markEdgeRec( e, getSecondLongestEdgeNo1( e ) );
      break;
    case 3 : // mark recursively all three edges of the element
      markEdgeRec( e, 1 );
      markEdgeRec( e, 2 );
      markEdgeRec( e, 3 );
      break;
    default:
      errorFP("GridFEAdT:: markAllEdgesRec","Refinement method %d"
	      "not implemented, must be 1-3.", getRefinementMethod() );
    }
  } else { // nsd==3

    int i;
    int no_mark_edges = getRefinementMethod();
    switch ( no_mark_edges ) {
    case 1 : // mark recursively the longest edge?
      markEdgeRec( e, 1 );  // is longest edge marked?
      break;
    case 2 : 
    case 3 : 
    case 4 : 
    case 5 :
      getEdgeLengthOrder(orderscratch,e);
      for(i=1;i<=no_mark_edges;i++) markEdgeRec( e, orderscratch(i) );
      break;
    case 6 : // mark recursively all six edges of the element
      for ( i=1; i<=6; i++ ) markEdgeRec( e, i );
      break;
    default:
      errorFP("GridFEAdT:: markAllEdgesRec","Refinement method %d"
	      "not implemented, must be 1-6.", getRefinementMethod() );
    }
  }
}


//-----------------------------------------------------------------------------
void GridFEAdT:: markAllEdgesRec( int e, int no_mark_edges )
//-----------------------------------------------------------------------------
// recursive marking of edges of element e
// (not used in 1D)
{//cout << "marking elm " << e << endl; cout.flush();
  if ( nsd==2 ) { // cout << "2D" <<  endl; cout.flush();
    switch ( no_mark_edges ){
    case 0 : break; // mark no edge
    case 1 : // mark only the longest edge of the element
      markEdgeRec( e, 1 );
      break;
    case 2 : // mark the two longest edges of the element
      markEdgeRec( e, 1 );
      markEdgeRec( e, getSecondLongestEdgeNo1( e ) );// assumes 1 is longest
      break;
    case 3 : // mark recursively all three edges of the element
      markEdgeRec( e, 1 );
      markEdgeRec( e, 2 );
      markEdgeRec( e, 3 );
      break;
    default:
      errorFP("GridFEAdT:: markAllEdgesRec","Mixed refinement method %d"
	      "not implemented, must be 0-3.", no_mark_edges);
    }
  } else { // nsd==3
//cout << "markine elm " << e << endl; cout.flush();
    int i;
    switch ( no_mark_edges ) {
    case 0 : break;
    case 1 : // mark recursively the longest edge?
      markEdgeRec( e, 1 );
      break;
    case 2 : 
    case 3 : 
    case 4 : 
    case 5 : 
      getEdgeLengthOrder(orderscratch,e);
      for(i=1;i<=no_mark_edges;i++)  markEdgeRec( e, orderscratch(i) );
      break;
    case 6 : // mark recursively all six edges of the element
//DbP("");
      for(i=1;i<=6;i++)  markEdgeRec( e, i );
//DbP("");
      break;
    default:
      errorFP("GridFEAdT:: markAllEdgesRec","Mixed refinement method %d"
	      "not implemented, must be 0-6.", no_mark_edges );
    }
  }
}

bool GridFEAdT:: lessThan( const int ed1, const int ed2, const int element)
{ // unique ordering of edges, which of edge ed1 and ed2 is longest?
// Routine to order edges that are of equal length
// return true if ed1 is longer than ed2
  int n1,n2,n3,n4;
  getNodesOnEdge ( element, ed1, n1, n2 );
  getNodesOnEdge ( element, ed2, n3, n4 );

if(useNodeForEqualTest()) {
// first check which edge contains the smallest node value
  int min12 = min(n1,n2), min34 = min(n3,n4);
  if ( min12 < min34 ) return true;
  if ( min12 > min34 ) return false;
// next if min12 and min34 equal check the sum
  if ( n1 + n2 < n3 + n4 ) return true;
  return false;
 }
  else {
// usefirst compute the coordinates of the midpoints
    for(int d=1; d<=nsd; d++) {
      real x1 = (coor(n1,d)+coor(n2,d));
      real x2 = (coor(n3,d)+coor(n4,d));
      if(x1<x2) return true;
      if(x1>x2) return false;
      // if equal use the next dimension
    }
    errorFP("GridFEAdT:: lessThan",
            "\nThe coordinates are identical,\n"
            "element %d, edge numbers %d %d, node numbers %d %d %d %d",
            element,ed1,ed2,n1,n2,n3,n4);
    return false;
  }

}

//-----------------------------------------------------------------------------
void GridFEAdT:: getEdgeLengthOrder( VecSimple(int)& order, int e )
//-----------------------------------------------------------------------------
// Get the edge number of the longest edge of element e in position 1
// the second longest in position 2 and so on
{
  int n = loc2glob( e, 1 );
  real x1 = coor(n,1), y1 = coor(n,2), z1 = coor(n,3);
  n = loc2glob( e, 2 );
  real x2 = coor(n,1), y2 = coor(n,2), z2 = coor(n,3);
  n = loc2glob( e, 3 );
  real x3 = coor(n,1), y3 = coor(n,2), z3 = coor(n,3);
  n = loc2glob( e, 4 );
  real x4 = coor(n,1), y4 = coor(n,2), z4 = coor(n,3);

  lengthsscratch(1) = sqr( x1 - x2 ) + sqr( y1 - y2 ) + sqr( z1 - z2 );
  lengthsscratch(2) = sqr( x2 - x3 ) + sqr( y2 - y3 ) + sqr( z2 - z3 );
  lengthsscratch(3) = sqr( x3 - x1 ) + sqr( y3 - y1 ) + sqr( z3 - z1 );
  lengthsscratch(4) = sqr( x1 - x4 ) + sqr( y1 - y4 ) + sqr( z1 - z4 );
  lengthsscratch(5) = sqr( x2 - x4 ) + sqr( y2 - y4 ) + sqr( z2 - z4 );
  lengthsscratch(6) = sqr( x3 - x4 ) + sqr( y3 - y4 ) + sqr( z3 - z4 ); 
//  order.redim(6); // assume that order and lengthsscratch has been redimed
  order(1)=1; order(2)=2; order(3)=3; order(4)=4; order(5)=5; order(6)=6;

// sort the lengths if equal length then "lessThan" is used
  int i,j,idx,idx2;
  real v;
  for (j=2;j<=6;j++) {
    v=lengthsscratch(idx=order(i=j));
    while( lengthsscratch(idx2=order(i-1)) < v ||
	   ( lengthsscratch(idx2) == v && lessThan(idx2,idx, e ) ) ) 
      { order(i)=idx2; i--; if(i==1) break; }
    order(i)=idx;
  }
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getLongestEdgeNo( int e )
//-----------------------------------------------------------------------------
// Get the local edge number of the longest edge in element e
{
  if ( nsd == 2 ) {
    int n = loc2glob( e, 1 );
    real x1 = coor(n,1), y1 = coor(n,2);
    n = loc2glob( e, 2 );
    real x2 = coor(n,1), y2 = coor(n,2);
    n = loc2glob( e, 3 );
    real x3 = coor(n,1), y3 = coor(n,2);
    real l1 = sqr( x1 - x2 ) + sqr( y1 - y2 );
    real l2 = sqr( x2 - x3 ) + sqr( y2 - y3 );
    real l3 = sqr( x3 - x1 ) + sqr( y3 - y1 );
    if( l1 >= l2 && l1 >= l3 ) return 1;
    return l2 >= l3 ? 2 : 3;
  }
  else if (nsd == 3 ) {
    getEdgeLengthOrder( orderscratch, e );
    return orderscratch(1);
  }
  errorFP("GridFEAdT:: getLongestEdgeNo",
	  "only implemented in 2D and 3D nsd=%d",nsd);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getSecondLongestEdgeNo1( int e )
//-----------------------------------------------------------------------------
// Get the local edge number of the second longest
// edge in element e, in 2D it is assumed that edge number 1 is longest
{
  if ( nsd==2 ) {
    int n = loc2glob( e, 2 );
    real x2 = coor(n,1), y2 = coor(n,2);
    n = loc2glob( e, 3 );
    real x3 = coor(n,1), y3 = coor(n,2);
    real l2 = sqr( x2 - x3 ) + sqr( y2 - y3 );
    real l3 = sqr( x3 - coor(n,1) ) + sqr( y3 - coor(n,2) );
    return l2 < l3 ? 3 : 2;
  }
  else if ( nsd==3 ) {
    getEdgeLengthOrder( orderscratch, e );
    return orderscratch(2);
  }
  errorFP("GridFEAdT:: getSecondLongestEdgeNo1",
	  "only implemented in 2D and 3D nsd=%d",nsd);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getShortestEdgeNo1( int e )
//-----------------------------------------------------------------------------
// Get the local edge number of the shortest edge in element e, in 2D it is 
// assumed edge 1 is longest
{
  if ( nsd==2 ) {
    int n = loc2glob( e, 2 );
    real x2 = coor(n,1), y2 = coor(n,2);
    n = loc2glob( e, 3 );
    real x3 = coor(n,1), y3 = coor(n,2);
    real l2 = sqr( x2 - x3 ) + sqr( y2 - y3 );
    real l3 = sqr( x3 - coor(n,1) ) + sqr( y3 - coor(n,2) );
    return l2 < l3 ? 2 : 3;
  }
  else if ( nsd==3 ) {
    getEdgeLengthOrder( orderscratch, e );
    return orderscratch(6);
  }
  errorFP("GridFEAdT:: getShortestEdgeNo1",
	  "only implemented in 2D and 3D nsd=%d",nsd);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getShortestEdgeNo( int e )
//-----------------------------------------------------------------------------
// Get the local edge number of the shortest edge in element e
{
  if ( nsd==2 ) {
    int n = loc2glob( e, 1 );
    real x1 = coor(n,1), y1 = coor(n,2);
    n = loc2glob( e, 2 );
    real x2 = coor(n,1), y2 = coor(n,2);
    n = loc2glob( e, 3 );
    real x3 = coor(n,1), y3 = coor(n,2);
    real l1 = sqr( x1 - x2 ) + sqr( y1 - y2 );
    real l2 = sqr( x2 - x3 ) + sqr( y2 - y3 );
    real l3 = sqr( x3 - x1 ) + sqr( y3 - y1 );
    if( l1 < l2 && l1 < l3 ) return 1;
    return l2 < l3 ? 2 : 3;
  }
  else if ( nsd==3 ) { 
    getEdgeLengthOrder( orderscratch, e );
    return orderscratch(6);
  }
  errorFP("GridFEAdT:: getShortestEdgeNo",
	  "only implemented in 2D and 3D nsd=%d",nsd);
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getShortestDiagonal( int e, MatSimple(real)& c )
//-----------------------------------------------------------------------------
// used for finding shortest diagonal in tetrahedron
// find the shortest internal diagonal of 5-10, 6-8, and 7-9
// it is not so important to avoid the arbitrariness of the 
// equal case, since the choice is an internal matter for an element.
// c contains the nodal coordinates (x,y,z)
{
  int n  = new_nodes( e, 1 ),  m  = new_nodes( e, 6 );
  real l1 = sqr(c(n,1)-c(m,1)) + sqr(c(n,2)-c(m,2)) + sqr(c(n,3)-c(m,3));
  n  = new_nodes( e, 2 );  m  = new_nodes( e, 4 );
  real l2 = sqr(c(n,1)-c(m,1)) + sqr(c(n,2)-c(m,2)) + sqr(c(n,3)-c(m,3));
  n  = new_nodes( e, 3 );  m  = new_nodes( e, 5 );
  real l3 = sqr(c(n,1)-c(m,1)) + sqr(c(n,2)-c(m,2)) + sqr(c(n,3)-c(m,3));

  if( l1 < l2 && l1 < l3 ) return 1;
  return l2 < l3 ? 2 : 3;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: getNodesOnEdge( int e, int edge_no, int& n1, int& n2 )
//-----------------------------------------------------------------------------
// get the global node numbers on local edge "edge_no" of a tetrahedron
// (only for ElmT4n3D) (no ordering of nodes on edge defined)
{
  switch ( edge_no ) {
  case 1: n1 = loc2glob(e,1); n2 = loc2glob(e,2); break;
  case 2: n1 = loc2glob(e,2); n2 = loc2glob(e,3); break;
  case 3: n1 = loc2glob(e,1); n2 = loc2glob(e,3); break;
  case 4: n1 = loc2glob(e,1); n2 = loc2glob(e,4); break;
  case 5: n1 = loc2glob(e,2); n2 = loc2glob(e,4); break;
  case 6: n1 = loc2glob(e,3); n2 = loc2glob(e,4); break;
  default: 
    errorFP("GridFEAdT:: getNodesOnEdge","illegal edge number %d",edge_no);
  }
}

//-----------------------------------------------------------------------------
int GridFEAdT::getElmAtFace(const int n1, const int n2, const int n3,
			     const int e)
//-----------------------------------------------------------------------------
{ 
// Given an element number e and the three global node numbers on the corners
// of an element face, compute the element number of the element 
// of the other side of the face.
// It is assumed that node-to-element (n2e) info for the grid is computed
// and that the elements are sorted in increasing order
// (which is the case for the present implementation of NeighborFE).
// 0 is returned if no other element is found, i.e., is the face
// is a boundary face.

  NeighborFE& n2e = additional_info.neighbor;
  int n1stop = n2e.nodeIrow(n1+1);
  int n2stop = n2e.nodeIrow(n2+1);
  int n3stop = n2e.nodeIrow(n3+1);
  int i = n2e.nodeIrow(n1);
  int j = n2e.nodeIrow(n2);
  int k = n2e.nodeIrow(n3);
  for( ; i<n1stop; i++ ) {
    int v = n2e.nodeJcol(i);
    if( v == e ) continue; // already know this common element
    while( j < n2stop && v > n2e.nodeJcol(j) ) j++; 
    if( j >= n2stop ) return 0; // outside range of elms to node 2
    if( v == n2e.nodeJcol(j) ) { // two elements equal
      while( k < n3stop && v > n2e.nodeJcol(k) ) k++;
      if( k >= n3stop ) return 0; // outside range of elms to node 3
      if( v == n2e.nodeJcol(k) ) return v;// if the last elm also equal, bingo!
    }
  }
  return 0;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getElmAtFace(const int n1, const int n2, const int e)
//-----------------------------------------------------------------------------
{ 
// Version for ElmT10n3D element
// Same as getElmAtFace(const int n1, const int n2, const int n3, const int e),
// but now it is sufficient with two midpoint nodes to identify face.

  NeighborFE& n2e = additional_info.neighbor;
  int n1stop = n2e.nodeIrow(n1+1);
  int n2stop = n2e.nodeIrow(n2+1);
  int i = n2e.nodeIrow(n1);
  int j = n2e.nodeIrow(n2);
  for( ; i<n1stop; i++ ) {
    int v = n2e.nodeJcol(i);
    if( v == e ) continue; // already know this common element
    while( j < n2stop && v > n2e.nodeJcol(j) ) j++; 
    if( j >= n2stop ) return 0; // outside range of elms to node 2
    if( v == n2e.nodeJcol(j) ) return v; // if the last elm also equal, bingo!
  }
  return 0;
}

//-----------------------------------------------------------------------------
void GridFEAdT::getElmsAtFaces(const int n1, const int n2, const int n3,
				const int n4, const int e, int& e3, int& e4)
//-----------------------------------------------------------------------------
{ 
// given an element number e and the four global node numbers on the corners
// of the element.
// Compute the element numbers of the elements of the other side of the two 
// faces which share the edge n1-n2.
// It is assumed that node-to-element (n2e) info for the grid is computed
// and that the elements their are sorted.
// 0 is returned if the neighbor element was not found, which means that
// a boundary face is found.

// e3 is the neighbor element to face n1 n2 n3 
// e4 is the neighbor element to face n1 n2 n4

  NeighborFE& n2e = additional_info.neighbor;
  int n1stop = n2e.nodeIrow(n1+1);
  int n2stop = n2e.nodeIrow(n2+1);
  int n3stop = n2e.nodeIrow(n3+1);
  int n4stop = n2e.nodeIrow(n4+1);
  int i = n2e.nodeIrow(n1); // start for first element n1 is part of
  int j = n2e.nodeIrow(n2);
  int k = n2e.nodeIrow(n3);
  int l = n2e.nodeIrow(n4);
  e3 = e4 = 0;
  for( ; i<n1stop; i++ ) {
    int v = n2e.nodeJcol(i);
    if( v == e ) continue; // already know this common element
    while( j < n2stop && v > n2e.nodeJcol(j) ) j++; 
    if( j >= n2stop ) return; // no more elms to node 2
    if( v == n2e.nodeJcol(j) ) { // both elements equal
      while( k < n3stop && v > n2e.nodeJcol(k) ) k++;
      if( k < n3stop && v == n2e.nodeJcol(k) ) { e3 = v; if(e4) return;}
      while( l < n4stop && v > n2e.nodeJcol(l) ) l++;
      if( l < n4stop && v == n2e.nodeJcol(l) ) { e4 = v; if(e3) return;}
    }
  }
}

//-----------------------------------------------------------------------------
void GridFEAdT::getElmsAtEdge( int e, int edge_no )
//-----------------------------------------------------------------------------
{ 
// efficient computation of all the elements which share the edge
// which has local edge number edge_no in element e.
// The local edge no of the elements are also computed.
// The result is stored last in two stacks: elmstack and ednostack,
// where stackcounter gives the position of the first position in the
// stack not used

  int n1,n2;
  getNodesOnEdge(e,edge_no,n1,n2);
  NeighborFE& n2e = additional_info.neighbor;// assumed n2e data computed
  int n1start = n2e.nodeIrow(n1);
  int n1stop = n2e.nodeIrow(n1+1);
  int n2start = n2e.nodeIrow(n2);
  int n2stop = n2e.nodeIrow(n2+1);
  int maxsize = max(n1stop-n1start,n2stop-n2start);
  int i;
  if ( elmstack.size() < stackcounter + maxsize ){ // dynamic size of stacks
// need to increase the size of the stacks
    int newsize = 2*elmstack.size() + maxsize; // new safe size for the stacks
// need scratch vector to move the data from the stack while the size of
// the stack is increased
    VecSimple(int) scratch(stackcounter-1); 
    for ( i=1; i<stackcounter; i++ ) scratch(i) = elmstack(i);
    elmstack.redim(newsize);
    for ( i=1; i<stackcounter; i++ ) elmstack(i) = scratch(i);
    for ( i=1; i<stackcounter; i++ ) scratch(i) = ednostack(i);
    ednostack.redim(newsize);
    for ( i=1; i<stackcounter; i++ ) ednostack(i) = scratch(i);
  }
  int j=n2start;
  for( i=n1start; i<n1stop; i++ ) {
    int v = n2e.nodeJcol(i);
    while( j < n2stop && v > n2e.nodeJcol(j) ) j++;
    if(j >= n2stop) return;
    if(v == n2e.nodeJcol(j)){
      ednostack(stackcounter) = getLocEdgeNo(n1,n2,v); 
      elmstack(stackcounter++) = v;
    }
  }
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getLocEdgeNo( int n1, int n2, int e )
//-----------------------------------------------------------------------------
// find the local edge number of element e which has nodes n1 and n2 
// used for ElmT4n3D
{ 
#ifdef SAFETY_CHECKS
// this version checks that n1 and n2 are nodes of the element
  int m1 = loc2glob(e,1);
  int m2 = loc2glob(e,2);
  if( ( n1 == m1 && n2 == m2 ) || ( n1 == m2 && n2 == m1 ) ) return 1;
  int m3 = loc2glob(e,3);
  if( n1 == m3 ) {
    if( n2 == m2 ) return 2;
    if( n2 == m1 ) return 3;    
  }
  if( n2 == m3 ) {
    if( n1 == m2 ) return 2;
    if( n1 == m1 ) return 3;    
  }
  int m4 = loc2glob(e,4);
  if( n1 == m4 ) {
    if( n2 == m1 ) return 4;
    if( n2 == m2 ) return 5;
    if( n2 == m3 ) return 6;
  }
  if( n2 == m4 ) {
    if( n1 == m1 ) return 4;
    if( n1 == m2 ) return 5;
    if( n1 == m3 ) return 6;
  }
  errorFP("GridFEAdT:: getLocEdgeNo",
	  "did not find edge in element %d with nodes %d and %d",e,n1,n2);
  return 0;
#else
  // in this version it is assumed that n1 and n2 are nodes of element e
  // if not so the result will be wrong
  if( n1 == loc2glob(e,1) ) {
    if( n2 == loc2glob(e,2) ) return 1;
    if( n2 == loc2glob(e,3) ) return 3;
    return 4;
  }
  if( n1 == loc2glob(e,2) ) {
    if( n2 == loc2glob(e,1) ) return 1;
    if( n2 == loc2glob(e,3) ) return 2;
    return 5;
  }
  if( n1 == loc2glob(e,3) ) {
    if( n2 == loc2glob(e,1) ) return 3;
    if( n2 == loc2glob(e,2) ) return 2;
    return 6;
  }
  if( n2 == loc2glob(e,1) ) return 4;
  if( n2 == loc2glob(e,2) ) return 5;
  return 6;
#endif
}

//-----------------------------------------------------------------------------
void GridFEAdT:: numberingNewNodes( int e, int edge_no )
//-----------------------------------------------------------------------------
// Go through the list of elements and their 
// marked edges and assign each of them a unique node number.
// A marked edge has a no value before it is assigned a (positive)
// node number.
{
  switch ( nne ) { // could use enumtype element
  case 3: { //      if( "ElmT3n2D" == elm_tp)
    int midptno = incCurrNodeNo();// a new node at the midpoint
    new_nodes( e, edge_no ) = midptno; // assign the primary position
    int elnei = getElmNeighbor( e, edge_no );
    if( elnei ){ // if nonzero means that the edge is internal
      if( getElmNeighbor( elnei, 1 ) == e ) // find edge no for neighbor elm
	new_nodes( elnei, 1 ) = midptno;    // assign the secondary position
      else if( getElmNeighbor( elnei, 2 ) == e ) 
	new_nodes( elnei, 2 ) = midptno;
      else 
	new_nodes( elnei, 3 ) = midptno;
    }
  } break;
  case 4: { //      if( "ElmT4n3D" == elm_tp)    
    int midptno = incCurrNodeNo();// a new node at the midpoint
    int i = stackcounter;
// note: stackcounter is a global variable which is increased in getElmsAtEdge
    getElmsAtEdge( e, edge_no ); // find tetrahedra which share the same edge
    for( ; i<stackcounter; i++)
      new_nodes(elmstack(i),ednostack(i)) = midptno;
  } break;
  case 6: { //      if( "ElmT6n2D" == elm_tp)
    int no1 = incCurrNodeNo(); // prior to mdptno
    int no2 = incCurrNodeNo(); // after in counterclockwise order
    new_nodes( e, edge_no     ) = no1; 
    new_nodes( e, edge_no + 3 ) = no2; 
    int elnei = getElmNeighbor( e, edge_no );
    if( elnei ){ // if nonzero means that the edge is internal
      if( getElmNeighbor( elnei, 1 ) == e ) { // find edge no for neighbor elm
	new_nodes( elnei, 1 ) = no2;          //  reverse order on other side
	new_nodes( elnei, 4 ) = no1;
      }
      else if( getElmNeighbor( elnei, 2 ) == e ) {
	new_nodes( elnei, 2 ) = no2;
	new_nodes( elnei, 5 ) = no1;
      }
      else {
	new_nodes( elnei, 3 ) = no2;
	new_nodes( elnei, 6 ) = no1;
      }
    }
  } break;
  case 10: { // if( "ElmT10n3D" == elm_tp)
    fatalerrorFP("GridFEAdT:: numberingNewNodes",
		 "ElmT10n3D not implemented (yet)");
// find the elements and local edgeno and put on stack(easier in this case)
    int midptno = loc2glob(e,4+edge_no); // numbering of edges and quad nodes
    int ed1 = incCurrNodeNo();
    int ed2 = incCurrNodeNo();
    int i = stackcounter;
// find better way for T10 using midpoint nodes
    getElmsAtEdge( e, edge_no ); // find tetrahedra which share the same edge
    for( ; i<stackcounter; i++) {
      int ne = elmstack(i);
      int nedno = ednostack(i);
      new_nodes(ne,nedno)    = midptno;
      new_nodes(ne,6+nedno)  = ed1; // check that the ordering is correct
      new_nodes(ne,12+nedno) = ed2; // could maybe change later
    }
  } break;
  default:
  fatalerrorFP("GridFEAdT:: numberingNewNodes",
	       "What element? wrong nne=%d",nne);
  }
}

//-----------------------------------------------------------------------------
void GridFEAdT:: getGlobNodesOnFace( int& n1, int& n2, int& n3, int f, int e )
//-----------------------------------------------------------------------------
// Get the global node numbers on face f of element e
{
  switch ( f ) {
  case 1: n1 = loc2glob(e,1); n2 = loc2glob(e,2); n3 = loc2glob(e,4);  break;
  case 2: n1 = loc2glob(e,2); n2 = loc2glob(e,3); n3 = loc2glob(e,4);  break;
  case 3: n1 = loc2glob(e,1); n2 = loc2glob(e,3); n3 = loc2glob(e,4);  break;
  case 4: n1 = loc2glob(e,1); n2 = loc2glob(e,2); n3 = loc2glob(e,3);  break;
  default:
    errorFP("GridFEAdT:: getNodesOnFace","illegal face number %d",f);
  }
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getFaceNo( int n1, int n2, int n3, int e, bool safe )
//-----------------------------------------------------------------------------
// Get the face number of element e which consists of the nodes n1, n2 and n3
// if safe is true the function will return 0 if no match is found.
{
  if(safe){
// a version which is safe in the case that n1,n2,n3 are not nodes of element e
    int m1 = loc2glob(e,1), m2 = loc2glob(e,2);
    int m3 = loc2glob(e,3), m4 = loc2glob(e,4);
    int f1=0, f2=0, f3=0, f4=0;

    if( n1 == m1 ) { f1++; f3++; f4++; }      // m1 is a part of face 1 3 and 4
    else if( n1 == m2 ) { f1++; f2++; f4++; } // m2 is a part of face 1 2 and 4
    else if( n1 == m3 ) { f2++; f3++; f4++; } // m3 is a part of face 2 3 and 4
    else if( n1 == m4 ) { f1++; f2++; f3++; } // m4 is a part of face 1 2 and 3

    if( n2 == m1 ) { f1++; f3++; f4++; }
    else if( n2 == m2 ) { f1++; f2++; f4++; }
    else if( n2 == m3 ) { f2++; f3++; f4++; }
    else if( n2 == m4 ) { f1++; f2++; f3++; }

    if( n3 == m1 ) { f1++; f3++; f4++; }
    else if( n3 == m2 ) { f1++; f2++; f4++; }
    else if( n3 == m3 ) { f2++; f3++; f4++; }
    else if( n3 == m4 ) { f1++; f2++; f3++; }

    if( f1 == 3 ) return 1; // if 3 hits for a face then correct face is found
    if( f2 == 3 ) return 2;
    if( f3 == 3 ) return 3;
    if( f4 == 3 ) return 4;

    return 0; // did not find any match
  } else {
// This version is not safe if n1,n2,n3 are not in element e
// but is is faster then the safe version ...
    int n = loc2glob(e,1);
    if( n != n1 && n != n2 && n != n3 ) return 2;
    n = loc2glob(e,2);
    if( n != n1 && n != n2 && n != n3 ) return 3;
    n = loc2glob(e,3);
    if( n != n1 && n != n2 && n != n3 ) return 1;
    return 4;
  }
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getParent( int e )
//-----------------------------------------------------------------------------
// Get the element number in prev_grid of the parent triangle of element e.
// Return 0 if there is no parent grid or if parent info is not computed
// use getParent_eff for efficiency if one knows that 
// isThisCoarsestGrid() || !isParentInfo() is false
{
  if( isThisCoarsestGrid() || !isParentInfo() ) return 0;
  return parent( e );
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getParent( int rel_level, int e )
//-----------------------------------------------------------------------------
// Get the parent element number of element e in grid rel_level levels up from
// from this grid.
// Return 0 if there is no parent grid or if parent info is not computed
// use getParent_eff for efficiency
{
  int parent_ele = e;
  GridFEAdT* g = this;
  for( int i=1; i<= rel_level; i++) {
    if( g->isThisCoarsestGrid() || ! g->isParentInfo() ) return 0;
    else { 
      parent_ele = g->getParent_eff( parent_ele );
      g = g->getParentGrid();
    }
  }
  return parent_ele;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getParent_eff( int rel_level, int e )
//-----------------------------------------------------------------------------
// More efficient version of getParent( int rel_level, int e )
{
  int parent_ele = e;
  GridFEAdT* g = this;
  for( int i=1; i<= rel_level; i++){
    parent_ele = g->getParent_eff( parent_ele );
    g = g->getParentGrid();
  }
  return parent_ele;
}

//-----------------------------------------------------------------------------
int GridFEAdT:: getChild(int e, int child_no)
//-----------------------------------------------------------------------------
// Get the element number in child grid of the child_no children
// of element e.
// Return 0 if there is no child grid or if children not computed or
// if the element does not have child_no children. 
// use getChild_eff for efficiency
{
  if( isThisFinestGrid() || ! isChildrenInfo() || 
      child_no > pos_child(e+1) - pos_child(e) || child_no < 1)
    return 0;
  return pos_child( e ) + child_no - 1;
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: checkParentInfo()
//-----------------------------------------------------------------------------
// Check for each element in this grid that its parent 
// has the element as one of the children.
{
  if( isThisCoarsestGrid() ) return true; // no parent, nothing to check
  GridFEAdT* pgrid = getParentGrid();
  if( pgrid->isThisFinestGrid() )
    fatalerrorFP("GridFEAdT::checkParentInfo",
		 "Inconsistency, the two grids do not agree.");
  for( int e=1; e <= nel; e++ )
    {
      int elm_no_par = getParent_eff(e);
      int no_of_ch = pgrid->getNoChildren( elm_no_par );
      bool found_child = false;
      for( int i=1; i<=no_of_ch; i++ )
	if( pgrid->getChild_eff( elm_no_par, i ) == e ){ 
	  found_child = true;
	  break;
	}
      if( ! found_child ) 
	return false; // Could not find child
    }
  return true; // everything seems ok
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: checkChildrenInfo()
//-----------------------------------------------------------------------------
// Check for all element in this grid that their children have the
// correct parent in this grid.
{
  if( isThisFinestGrid() ) return true; //has no children, nothing to check
  GridFEAdT* cgrid = getChildGrid();
  if( cgrid->isThisCoarsestGrid() )
    fatalerrorFP("GridFEAdT::checkChildrenInfo",
		 "Inconsistency, the two grids do not agree.");
  for( int e=1; e<=nel; e++ ) {
    int no_of_ch = getNoChildren(e);
    for( int i=1; i<=no_of_ch; i++ )
      if( cgrid->getParent_eff( getChild_eff( e, i )) != e ) 
	return false; 
  }
  return true; // everything seems ok
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: checkElementOrientation ()
//-----------------------------------------------------------------------------
// Check for all elements in this grid that they are counterclockwise oriented.
// for ElmT6n2D check also the "quadratic nodes" i.e. the midnodes.
{
  if ( nsd==1 ) {
    if( "ElmB2n1D" == elm_tp ){
      bool test = true;
      for( int e=1; e<=nel; e++ )
	if( coor(loc2glob( e, 1 ),1) > coor(loc2glob( e, 2 ),1) ) {
	  warningFP("GridFEAdT:: checkElementOrientation",
		    "Wrong orientation in element %d",e);
	  test = false;
	}
      return test;
    }
    else if( "ElmB3n1D" == elm_tp ){
      bool test = true;
      for( int e=1; e<=nel; e++ )
	if( coor(loc2glob( e, 1 ),1) > coor(loc2glob( e, 2 ),1) ||
	    coor(loc2glob( e, 2 ),1) > coor(loc2glob( e, 3 ),1) ) {
	  warningFP("GridFEAdT:: checkElementOrientation",
		    "Wrong orientation in element %d",e);
	  test = false;
	}
      return test;
    }
    else
      fatalerrorFP("GridFEAdT:: checkElementOrientation",
		   "Not implemented yet, element type = %s",elm_tp.chars() );
  }
  else if ( nsd==2 ) {
    if( "ElmT3n2D" != elm_tp && "ElmT6n2D" != elm_tp )
      fatalerrorFP("GridFEAdT:: checkElementOrientation",
		   "Not implemented yet, element type = %s",elm_tp.chars() );
    int n1,n2,n3;
    bool test = true;
    
    int times = ( "ElmT3n2D" == elm_tp ) ? 1 : 2;
    
    for ( int i=1; i<=times; i++ ) {
      for( int e=1; e<=nel; e++ ) {
	if(i==1){
	  n1 = loc2glob( e, 1 ); n2 = loc2glob( e, 2 ); n3 = loc2glob( e, 3 );
	} else {
	  n1 = loc2glob( e, 4 ); n2 = loc2glob( e, 5 ); n3 = loc2glob( e, 6 );
	}
	real x1 = coor(n1,1),  y1 = coor(n1,2);
	real x2 = coor(n2,1),  y2 = coor(n2,2);
	real x3 = coor(n3,1),  y3 = coor(n3,2);
	real det = (x1-x3)*(y2-y3) - (x2-x3)*(y1-y3);
	if( det < 0 ) {
	  warningFP("GridFEAdT:: checkElementOrientation",
		    "Wrong orientation in element %d",e);
	  test = false;
	}
      }
    }
    return test;
  }
  else if ( nsd==3 ) {
    if( "ElmT4n3D" != elm_tp && "ElmT10n3D" != elm_tp )
      fatalerrorFP("GridFEAdT:: checkElementOrientation",
		   "Not implemented yet, element type = %s",elm_tp.chars() );

    bool test = true;
    for( int e=1; e<=nel; e++ ) {
      FET4.refill(e);
      if( FET4.det() < 0 ) {
	errorFP("GridFEAdT:: checkElementOrientation",
		"\nWrong orientation in element %d"
		" n1=%d, n2=%d, n3=%d, n4=%d",e,
		FET4.n1(),FET4.n2(),FET4.n3(),FET4.n4());
//	putLoc2Glob(FET4.n2(),e,1); // does _not_ correct here
//	putLoc2Glob(FET4.n1(),e,2);
	test = false;
      }
    }
    return test;
  }
  return false; // should have returned before
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridL2Q()
//-----------------------------------------------------------------------------
// construct a grid with piecewise quadratic elements from a linear grid
{
  switch (nsd) {
  case 1: return changeGridB2toB3();
  case 2: return changeGridT3toT6();
  case 3: return changeGridT4toT10();
  default:
    fatalerrorFP("GridFEAdT::changeGridL2Q","nsd=%d not implemented yet",nsd);
  }
  return NULL;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridQ2L()
//-----------------------------------------------------------------------------
// construct a grid with piecewise linear elements from a quadratic grid
{ 
  switch (nsd) {
  case 1: return changeGridB3toB2();
  case 2: return changeGridT6toT3();
  case 3: return changeGridT10toT4();
  default: 
    fatalerrorFP("GridFEAdT::changeGridQ2L","nsd=%d not implemented yet",nsd);
  }
  return NULL;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: initL2QorQ2L( GridFEAdT* newgrid, const int new_nno )
//-----------------------------------------------------------------------------
// initialize changed grid. Linear to quadratic or quadratic to linear
{
  newgrid->setNonUniformMesh();
  newgrid->setRegBisectOrShort( getRegBisectOrShort() ); // only used in 2D
  newgrid->setNodeForEqualTest( useNodeForEqualTest() ); // only used in 3D

// copy the GridFEAdT specific information

  if( ! isThisFinestGrid() ){                       // there is finer grid...
    newgrid->setFinestGrid( false );
    newgrid->setChildGrid ( getChildGrid() ); // set pointer to child
    getChildGrid()->setParentGrid( newgrid ); // and back again
// Note that the old grid (this) will be incorrect since the grid it
// points to does not point back.
    if( isChildrenInfo() ){                   // children info exists
      newgrid->setChildrenInfo( true );
      newgrid->pos_child.redim(nel+1);
      newgrid->pos_child = pos_child;
    }
  }

  if( ! isThisCoarsestGrid() ){ // there is a coarser grid...
    newgrid->setCoarsestGrid( false );
    newgrid->setParentGrid ( getParentGrid() );// set pointer to parent
    getParentGrid()->setChildGrid( newgrid);   // and back again
// Note that the old grid (this) will be incorrect since the grid it
// points to does not point back.
    if( isParentInfo() ){                // parent info exists
      newgrid->setParentInfo( true );
      newgrid->parent.redim(nel);
      newgrid->parent = parent;
    }
  }
	
  removeHierInfo();  // maybe drastic?, but the info in old grid is now correct

  if ( isNeighborsComputed() ) {
    newgrid->setNeighborsComputed( true );
    newgrid->elm_neigh.redim(nel,nsd+1);
    newgrid->elm_neigh = elm_neigh;
  }

// will not copy new_nodes, (the matrix used to mark and number new nodes)

  newgrid->setRefinementMethod( getRefinementMethod() );

// copy the name of the boundary indicators
  int i,b;
  for ( i=1; i<=getNoBoInds(); i++) 
    newgrid->putBoIndName( getBoIndName(i), i );

// copy the old nodes which will be present in the new grid
// copy boundary indicators    

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

  if ( nsd==1 )
    for( i=1; i<=new_nno; i++ )
      newgrid->coor(i,1) = coor(i,1); 
  else if ( nsd==2 )
    for( i=1; i<=new_nno; i++ ){
      newgrid->coor(i,1) = coor(i,1); 
      newgrid->coor(i,2) = coor(i,2); 
    }
  else if ( nsd==3 )
    for( i=1; i<=new_nno; i++ ){
      newgrid->coor(i,1) = coor(i,1); 
      newgrid->coor(i,2) = coor(i,2); 
      newgrid->coor(i,3) = coor(i,3); 
    }
  else
    fatalerrorFP("GridFEAdT::initL2QorQ2L","Wrong nsd=%d",nsd);

  if ( getNoBoInds() )
    for( i=1; i<=new_nno; i++ )
      for( b=1; b<=getNoBoInds(); b++ )
	if( ind(i,b) == '1' ) 
	  nind(i,b) = '1';

  if( !onemat ) 
    newgrid->material_type = material_type;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridB2toB3()
//-----------------------------------------------------------------------------
// Create a grid with quadratic elements ElmB3n1d
// from linear elements ElmB2n1d 
{
  if( "ElmB3n1D" == elm_tp ) {
    warningFP("GridFEAdT:: changeGridB2toB3", "Already a ElmB3n1D grid");
    return this;
  }

  if( "ElmB2n1D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridB2toB3",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  if( !isNeighborsComputed() ) // Is the element-element info computed?
    initNeighbor();

  int b,e;

  int oldrefmethod = getRefinementMethod();
  setRefinementMethod(1); // set a new node on each edge
  markAllElements();
  setRefinementMethod(oldrefmethod);

  int new_nno = nno + nel;  // one new node for each changed element

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 3, getNoBoInds(), 3, onemat );
  newgrid->setElmType( "ElmB3n1D" );

  initL2QorQ2L(newgrid,nno);

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

// Go through all elements and compute position of the new node
// at the midpoint

  for( e=1; e<=nel; e++ ) {
    int n1 =  loc2glob( e, 1 );
    int n2 =  loc2glob( e, 2 );
    int m1 = incCurrNodeNo(); 
    newgrid->putLoc2Glob123 ( n1, m1, n2, e );
    newgrid->coor(m1,1) = (coor(n1,1)+coor(n2,1))/2;
    for( b=1; b<=getNoBoInds(); b++ )
      if( ind(n1,b) == '1' && ind(n2,b) == '1' ) // if both ends are set...
	nind(m1,b) = '1'; // then set midpoint
  }

#ifdef SAFETY_CHECKS
  if( !checkElementOrientation() )
    errorFP("GridFEAdT:: changeGridB2toB3", "Wrong element orientation.");
#endif

  return newgrid;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridB3toB2()
//-----------------------------------------------------------------------------
// Create a linear triangle elements grid ElmT3n2d from a grid 
// with quadratic elements ElmT6n2d.
{
  if( "ElmB2n1D" == elm_tp ) {
    warningFP("GridFEAdT:: changeGridT6toT3", "Already a ElmB2n1D grid");
    return this;
  }

  if( "ElmB3n1D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridB3toB2",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  int e;

// find largest node number for the nodes to get size for node vector
  int new_nno = 0;
  for( e=nel; e>0; e-- ) {
    if( loc2glob(e,1) > new_nno )  new_nno = loc2glob(e,1);
    if( loc2glob(e,3) > new_nno )  new_nno = loc2glob(e,3);
  }

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 2, getNoBoInds(), 2, onemat );
  newgrid->setElmType( "ElmB2n1D" );

  initL2QorQ2L(newgrid,new_nno);

// Go through all elements and copy elements

  for( e=1; e<=nel; e++ ) 
    newgrid->putLoc2Glob12 ( loc2glob( e, 1 ), loc2glob( e, 3 ), e );

  return newgrid;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridT3toT6()
//-----------------------------------------------------------------------------
// Create a grid with quadratic elements ElmT6n2d
// from alinear triangle elements ElmT3n2d  grid
{
  if( "ElmT6n2D" == elm_tp ) {
    warningFP("GridFEAdT:: changeGridT3toT6", "Already a ElmT6n2D grid");
    return this;
  }

  if( "ElmT3n2D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridT3toT6",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  if( !isNeighborsComputed() )// Is the element-element info computed?
    initNeighbor();

  int oldrefmethod = getRefinementMethod();
  setRefinementMethod(3); // set a new node on each edge
  markAllElements();
  setRefinementMethod(oldrefmethod);

  // the numbering of the new nodes has incremented the current_node_number

  int new_nno = getCurrNodeNo(); 

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 6, getNoBoInds(), 6, onemat );
  newgrid->setElmType( "ElmT6n2D" );

  initL2QorQ2L(newgrid,nno);

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

// Go through all elements and compute position of the new nodes
// at the midpoints of edges which have been assigned a node number.

  int b,e,n1,n2,n3,n4,n5,n6;

  for( e=1; e<=nel; e++ ) {

    n1 =  loc2glob( e, 1 );  n2 =  loc2glob( e, 2 );  n3 =  loc2glob( e, 3 );
    n4 = new_nodes( e, 1 );  n5 = new_nodes( e, 2 );  n6 = new_nodes( e, 3 );

    newgrid->putLoc2Glob123 ( n1, n2, n3, e );
    newgrid->putLoc2Glob456 ( n4, n5, n6, e );
    
    if( ! newgrid->coor(n4,1) ){ // to avoid computing things twice
      newgrid->coor(n4,1) = (coor(n1,1)+coor(n2,1))/2;
      newgrid->coor(n4,2) = (coor(n1,2)+coor(n2,2))/2;
      for( b=1; b<=getNoBoInds(); b++ )
	if( ind(n1,b)=='1' && ind(n2,b)=='1' ) //if both ends are set...
	  nind(n4,b) = '1'; // set midpoint
    }
    if( ! newgrid->coor(n5,1) ){ 
      newgrid->coor(n5,1) = (coor(n2,1)+coor(n3,1))/2;
      newgrid->coor(n5,2) = (coor(n2,2)+coor(n3,2))/2;
      for( b=1; b<=getNoBoInds(); b++ )
	if( ind(n2,b)=='1' && ind(n3,b)=='1' )
	  nind(n5,b) = '1';
    }
    if( ! newgrid->coor(n6,1) ){ 
      newgrid->coor(n6,1) = (coor(n3,1)+coor(n1,1))/2;
      newgrid->coor(n6,2) = (coor(n3,2)+coor(n1,2))/2;
      for( b=1; b<=getNoBoInds(); b++ )
	if( ind(n3,b)=='1' && ind(n1,b)=='1' )
	  nind(n6,b) = '1';
    }
  }

#ifdef SAFETY_CHECKS
  if( !checkElementOrientation() )
    errorFP("GridFEAdT:: changeGridT3toT6", "Wrong orientation.");
#endif

  return newgrid;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridT6toT3()
//-----------------------------------------------------------------------------
// Create a linear triangle elements grid ElmT3n2d from a grid 
// with quadratic elements ElmT6n2d.
{
  if( "ElmT3n2D" == elm_tp ) {
    warningFP("GridFEAdT:: changeGridT6toT3", "Already a ElmT3n2D grid");
    return this;
  }

  if( "ElmT6n2D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridT6toT3",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

// find largest node number for the corner nodes to get size for node vector
  int e;
  int new_nno = 0;
  for( e=nel; e>0; e-- ) {
    if( loc2glob(e,1) > new_nno )  new_nno = loc2glob(e,1);
    if( loc2glob(e,2) > new_nno )  new_nno = loc2glob(e,2);
    if( loc2glob(e,3) > new_nno )  new_nno = loc2glob(e,3);
  }

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 3, getNoBoInds(), 3, onemat );
  newgrid->setElmType( "ElmT3n2D" );

  initL2QorQ2L(newgrid,new_nno);

// Go through all elements and copy elements

  for( e=1; e<=nel; e++ ) 
    newgrid->putLoc2Glob123 ( loc2glob( e, 1 ),
			      loc2glob( e, 2 ),
			      loc2glob( e, 3 ), e );

  return newgrid;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridT4toT10()
//-----------------------------------------------------------------------------
// Create a grid with quadratic tetrahedral elements ElmT10n3D
// from a grid with linear tetrahedral elements ElmT4n3D.
{
  if( "ElmT10n3D" == elm_tp ) {
    warningFP("GridFEAdT:: changeGridT4toT10","Already a ElmT10n3D grid");
    return this;
  }

  if( "ElmT4n3D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridT4toT10",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

  if( !isNeighborsComputed() )// Is the element-element info computed?
    initNeighbor();

  int oldrefmethod = getRefinementMethod();
  setRefinementMethod(6);
  markAllElements();
  setRefinementMethod(oldrefmethod); // reset the old value

  // the numbering of the new nodes has incremented the current_node_number

  int new_nno = getCurrNodeNo(); 

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 10, getNoBoInds(), 10, onemat );
  newgrid->setElmType( "ElmT10n3D" );

  initL2QorQ2L(newgrid,nno);

  MatSimple(char)&  ind = boundaryData().indicatorAccess();
  MatSimple(char)&  nind = newgrid->boundaryData().indicatorAccess();

// Go through all elements and compute position of the new nodes
// at the midpoints of edges which have been assigned a node number.

  int b,i,e,n1,n2,n3,n4,n5,n6,n7,n8,n9,n10,mid,m1,m2;

  for( e=1; e<=nel; e++ ) {

    n1 =  loc2glob( e, 1 );  n2 =  loc2glob( e, 2 );
    n3 =  loc2glob( e, 3 );  n4 =  loc2glob( e, 4 );
    n5 = new_nodes( e, 1 );  n6 = new_nodes( e, 2 );
    n7 = new_nodes( e, 3 );  n8 = new_nodes( e, 4 );
    n9 = new_nodes( e, 5 );  n10= new_nodes( e, 6 );
    newgrid->putLoc2Glob1234 ( n1, n2, n3, n4, e );
    newgrid->putLoc2Glob5678910 ( n5, n6, n7, n8, n9, n10, e );

    for(i=1;i<=6;i++){
      switch (i) {
      case 1: mid=n5;  m1=n1;  m2=n2; break;
      case 2: mid=n6;  m1=n2;  m2=n3; break;
      case 3: mid=n7;  m1=n3;  m2=n1; break;
      case 4: mid=n8;  m1=n1;  m2=n4; break;
      case 5: mid=n9;  m1=n2;  m2=n4; break;
      case 6: mid=n10; m1=n3;  m2=n4; break;
      }
      if( mid && ! newgrid->coor(mid,1) ){ // to avoid recomputing things
	newgrid->coor(mid,1) = (coor(m1,1)+coor(m2,1))/2;
	newgrid->coor(mid,2) = (coor(m1,2)+coor(m2,2))/2;
	newgrid->coor(mid,3) = (coor(m1,3)+coor(m2,3))/2;
	for( b=1; b<=getNoBoInds(); b++ )
	  if( ind(m1,b)=='1' && ind(m2,b)=='1' ) // if both ends are set...
	    nind( mid, b )='1'; // then set midpoint
      }
    }
  }

#ifdef SAFETY_CHECKS
  if( !checkElementOrientation() )
    errorFP("GridFEAdT:: changeGridT4toT10", "Wrong orientation.");
#endif

  return newgrid;
}

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: changeGridT10toT4()
//-----------------------------------------------------------------------------
// Create a grid with linear tetrahedral grid ElmT4n3D
// from a grid with quadratic elements ElmT10n3D.
{
  if( "ElmT4n3D" == elm_tp )
    {
      warningFP("GridFEAdT:: changeGridT10toT4",
		"Already a ElmT4n3D grid");
      return this;
    }

  if( "ElmT10n3D" != elm_tp )
    fatalerrorFP("GridFEAdT:: changeGridT10toT4",
		 "Not implemented yet, element type = %s",elm_tp.chars() );

// find largest node number for the corner nodes to get size for node vector

  int e;
  int new_nno = 0;
  for( e=nel; e>=1; e-- ) {
    if( loc2glob(e,1) > new_nno )  new_nno = loc2glob(e,1);
    if( loc2glob(e,2) > new_nno )  new_nno = loc2glob(e,2);
    if( loc2glob(e,3) > new_nno )  new_nno = loc2glob(e,3);
    if( loc2glob(e,4) > new_nno )  new_nno = loc2glob(e,4);
  }

  GridFEAdT* newgrid = new GridFEAdT;
  newgrid->redim( nsd, new_nno, nel, 4, getNoBoInds(), 4, onemat );
  newgrid->setElmType( "ElmT4n3D" );

  initL2QorQ2L(newgrid,new_nno);

// Go through all elements and copy elements

  for( e=1; e<=nel; e++ )
    newgrid->putLoc2Glob1234 ( loc2glob( e, 1 ),
			       loc2glob( e, 2 ),
			       loc2glob( e, 3 ),
			       loc2glob( e, 4 ), e);

  return newgrid;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: makeSmoothing (int nrepetitions, int qualcrit)
//-----------------------------------------------------------------------------
{ // qualcrit is the quality criterion described in ElmTtools
  // Apply smoothing to the grid.
// short description of algorithm:
// for the nodes which no indicator is set,
// make   
  Ptv(real) x(nsd),xold(nsd);  // coordinates of a node
//  additional_info.initNeighbor(*this);
  additional_info.initNeighbor(*this,true,true,false);
  int nb,n,i,k,repetitions, no_of_nb_nodes, bestnode,d;
  real eta, besteta;
  NeighborFE& neighbor = additional_info.neighbor;
  VecSimplest(int) element_list;
  VecSimplest(int) node_list;
  VecSimple(int) changed(nno);
  changed.fill(1);
  real alpha = 0.1;
  int maxit = 1;
  int total_moves = 1;
  repetitions = 1;
  while( alpha>0.01 || repetitions <= nrepetitions){
//  for (repetitions=1; repetitions <= nrepetitions; repetitions++) {
    total_moves = 0;
    for (n=1; n<=nno; n++)
      {
	bool found_better = false;
	if ( (! boNode( n ) && (changed( n ))) ) 
// the boundary nodes are not to be moved
	  {
	    neighbor.node(n,element_list);   // neighbor elements
	    neighbor.couplings(n,node_list); // neighbor nodes
	    // save the old node coordinates.
	    xold = getCoor(n);
	    for(k=1;k<=maxit;k++) {
	      besteta = 0;
	      bestnode = 0;
	      no_of_nb_nodes = node_list.size();
	      for( i=1; i<=no_of_nb_nodes; i++ ){
		nb = node_list(i);
		if ( nb == n )
		  for(d=1;d<=nsd;d++)
		    x(d) = xold(d);
		else
		  for(d=1;d<=nsd;d++)
		    x(d) = (1-alpha)*xold(d) + alpha*getCoor(nb,d);
		putCoor(x,n);
		eta = getMinEta(element_list,qualcrit);
		if(eta > besteta + 1e-12 ) { // do avoid arbitrariness when =
		  besteta = eta;
		  bestnode = nb;
		}
	      }
	      if( bestnode != n )
		for(d=1;d<=nsd;d++)
		  xold(d) = (1-alpha)*xold(d) + alpha*getCoor(bestnode,d);
	      if ( bestnode == n ) break;
	      found_better = true;
	      total_moves++;
	    } 
	    putCoor( xold,n);
	    if ( found_better ) 
	      for( i=1; i<=no_of_nb_nodes; i++ )
		changed( node_list(i) ) = 1;
	    else
	      changed( n ) = 0;
	  }
	else // if indicator node, not to be changed
	  changed( n ) = 0;
      }
    cout<<"In smoothing iteration "<<repetitions<<
      " the total number of moved nodes was "<<total_moves<<endl;
    repetitions++;
    if(repetitions>nrepetitions) break;
    if ( total_moves == 0 ) {
      alpha/=2;// decrease the move
      changed.fill(1);
    }
  }
  setNonUniformMesh();
}

//-----------------------------------------------------------------------------
void GridFEAdT::smoothNode ( int node, int qualcrit)
//-----------------------------------------------------------------------------
{
  // Apply smoothing to a node in the grid.
  
  Ptv(real) x(nsd),xold(nsd);  // coordinates of a node
//  additional_info.initNeighbor(*this);
  additional_info.initNeighbor(*this,true,true,false);
  int nb,i,k, no_of_nb_nodes, bestnode;
  real eta, besteta;
  NeighborFE& neighbor = additional_info.neighbor;
  VecSimplest(int) element_list;
  VecSimplest(int) node_list;
  real alpha = 0.05;
  int maxit = 25;
  if (!boNode(node)) // the boundary nodes are not to be moved
    {
      neighbor.node(node,element_list);   // neighbor elements
      neighbor.couplings(node,node_list); // neighbor nodes
      
      // save the old node coordinates.
      xold = getCoor(node);
      for ( k=1; k<=maxit; k++ ) {
	besteta = 0;
	bestnode = 0;
	no_of_nb_nodes = node_list.size();
	for( i=1; i<=no_of_nb_nodes; i++ ){
	  nb = node_list(i);
	  x = (1-alpha)*xold + alpha*getCoor(nb);
	  putCoor(x,node);
	  eta = getMinEta(element_list,qualcrit);
	  if(eta > besteta + 1e-12 ) { // do avoid arbitrariness when =	
	    besteta = eta;
	    bestnode = nb;
	  }
	}
	xold = (1-alpha)*xold + alpha*getCoor(bestnode); // move the node
//	      cout<<"node "<<n<<"min eta = "<<besteta<<endl;  
	if ( bestnode == node ) break;
      } 
      putCoor( xold,node);
    }
  setNonUniformMesh();
}

//-----------------------------------------------------------------------------
void GridFEAdT::smoothNeighbors (int node, int qualcrit )
//-----------------------------------------------------------------------------
{
  // Apply smoothing to the the neighbors of a node in the grid.
//  additional_info.initNeighbor(*this);
  additional_info.initNeighbor(*this,true,true,false);
  VecSimplest(int) node_list;
  additional_info.neighbor.couplings(node,node_list); // neighbor nodes
  int length = node_list.size();
  for( int n =1;n<=length; n++)
    smoothNode(node_list(n),qualcrit);
}

//-----------------------------------------------------------------------------
real GridFEAdT:: getMinEta( VecSimplest(int)& element_list, int qualcrit )
//-----------------------------------------------------------------------------
{
  real mineta = 99999;
  int length = element_list.size();
  if(length==0) 
    warningFP("GridFEAdT:: getMinEta","length is 0");
  int e;
  real eta;
  if ( nsd==2 ) {
    for( e=1; e<=length; e++ ) {
      FET3.refill(element_list(e));
      if( FET3.det() <= 0.0 )
	eta = -1; // illegal grid, could happen in smoothing
      else
	eta = FET3.qualityMeasure(qualcrit);
      if( eta < mineta ) mineta = eta;
    }
  }
  else if( nsd==3 ) {
    for( e=1; e<=length; e++ ) {
      FET4.refill(element_list(e));
      if( FET4.det() <= 0.0 )
	eta = -1; // illegal grid, could happen in smoothing
      else
	eta = FET4.qualityMeasure(qualcrit);
      if( eta < mineta ) mineta = eta;
    }
  }
  else
    errorFP("GridFEAdT:: getMinEta","Wrong dimension %d",nsd);
  return mineta;
}

//-----------------------------------------------------------------------------
void GridFEAdT::makeStatistics (Os os)
//-----------------------------------------------------------------------------
{ // ksa: this one is not finished add more statistics
  real maxlength = 0;
  real minlength = 99999;
  real minangle = 99999;
  real maxangle = -99999;
  real maxeta = 0;
  real mineta = 99999;
  real maxlen,minlen,maxang,minang;
  
  int e;
  if ( nsd==2 ) {
    for ( e=1; e<=nel; e++ ) {
      FET3.refill(e);
      if( FET3.det() <= 0.0 ) {
	errorFP("GridFEAdT:: makeStatistics",
		"Wrong orientation in element %d, n1=%d,n2=%d,n3=%d,",e,
		FET3.n1(),FET3.n2(),FET3.n3());
      }
      maxang = FET3.maxAngle();
      minang = FET3.minAngle();
      real eta = FET3.qualityMeasure(1);
      maxlen = FET3.maxLength();
      minlen = FET3.minLength();
      if( eta > maxeta ) maxeta = eta;
      if( eta < mineta ) mineta = eta;
      if( maxlen > maxlength ) maxlength = maxlen;
      if( minlen < minlength ) minlength = minlen;
      if( maxang > maxangle ) maxangle = maxang;
      if( minang < minangle ) minangle = minang;
    }
  }
  else if ( nsd==3 ) {
    for ( e=1; e<=nel; e++ ) {
      FET4.refill(e);
      if( FET4.det() <= 0.0 ) {
	errorFP("GridFEAdT:: makeStatistics",
		"Wrong orientation in element %d, n1=%d,n2=%d,n3=%d,n4=%d",
		e,FET4.n1(),FET4.n2(),FET4.n3(),FET4.n4());
      }
      maxang = FET4.maxSolidAngle();
      minang = FET4.minSolidAngle();
      real eta = FET4.qualityMeasure(1);
      maxlen = FET4.maxLength();
      minlen = FET4.minLength();
      if( eta > maxeta ) maxeta = eta;
      if( eta < mineta ) mineta = eta;
      if( maxlen > maxlength ) maxlength = maxlen;
      if( minlen < minlength ) minlength = minlen;
      if( maxang > maxangle ) maxangle = maxang;
      if( minang < minangle ) minangle = minang;
    }
  }
  
  os<<endl;
  os<<"maxlength = "<<maxlength<<endl;
  os<<"minlength = "<<minlength<<endl;
  os<<"max angle = "<<maxangle<<endl;
  os<<"min angle = "<<minangle<<endl;
  os<<"max eta (crit 1)= "<<maxeta<<endl;
  os<<"min eta (crit 1)= "<<mineta<<endl;
  os<<endl;
}


//-----------------------------------------------------------------------------
int GridFEAdT:: findElm( Ptv(real)& x, int level0, int elm0, int level1 )
//-----------------------------------------------------------------------------
// given a point (x) which is in elm0 on level level0, find the element
// which the point is in in level1
{
  if(level1<level0) 
    warningFP("GridFEAdT:: findElm",
	      "getParent is more efficient if level1<level0");
  GridFEAdT* grid = getCoarsestGrid();
  int i;
  for(i=1;i<level0;i++) grid = grid->getChildGrid();
  int element = elm0;
  if( ! grid->isInsideElm( x, element ) )
    fatalerrorFP("GridFEAdT:: findElm","The point must be in element!");

//cout << pos_child;
  //pos_child.print(cout);
  for( i = level0; i<level1; i++) {
    //grid->printGrid(oform("level%d",i));
    if( grid->isChildrenInfo() ){
      int firstchildele = grid->getFirstChild(element);
      int lastchildele = grid->getFirstChild(element+1);
      //cout << "firstchildele " << firstchildele << endl;
      //cout << "lastchildele " << lastchildele << endl;
      grid = grid->getChildGrid();
// note that last element can also be tested
      for( element=firstchildele; element<lastchildele; element++ )
	if( grid->isInsideElm( x, element ) )
	  break;
      if( element == lastchildele ) 
	fatalerrorFP("GridFEAdT:: findElm","failed to find child!");
    }
    else
      fatalerrorFP("GridFEAdT:: findElm", 
		   "No children info exists, level=%d",i);
  }
  return element;
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: findElm( Ptv(real)& x, int level, int& element)
//-----------------------------------------------------------------------------
{
  GridFEAdT* grid = getCoarsestGrid();
  if( ! grid->simpleSearch(x,element) )
    return false;
  for( int level_counter = 1; level_counter<level; level_counter++) {
    if( grid->isChildrenInfo() ){
      int firstchildele = grid->getFirstChild(element);
      int lastchildele = grid->getFirstChild(element+1) - 1;
      grid =  grid->getChildGrid();
// note that last element is not tested
      for( element=firstchildele; element<lastchildele; element++ )
	if( grid->isInsideElm( x, element ) ) 
	  break;
    }
    else
      fatalerrorFP("GridFEAdT:: findElm", "No children info exists");
  }
  return true;
}


//-----------------------------------------------------------------------------
bool GridFEAdT:: simpleSearch( Ptv(real)& x, int& element)
//-----------------------------------------------------------------------------
{
  for( element=1; element<=nel; element++ ) 
    if( isInsideElm( x, element ) ) return true;
  return false;
}

//-----------------------------------------------------------------------------
bool GridFEAdT:: isInsideElm( Ptv(real)& gpt, int element )
//-----------------------------------------------------------------------------
{
  if ( nsd==1 ) {
    real x = gpt(1);
    int n1,n2;
    real x1,x2;
    real df; // safety factor to test if inside element
  
    if( isBoundingBoxesComputed() )
      { 
        box_counter++;
        if( x < bounding_box(element,1) || x > bounding_box(element,2) )
	  return false;
        df = 1e-10*exp(getGridLevelNo()*0.742);
      }
    else
      {
        box_counter++;
        df = 2e-10*exp(getGridLevelNo()*0.742);
      
        n1 = loc2glob( element, 1 );
        n2 = loc2glob( element, getNoNodesInElm(element) ); 
        // works for both linear and quadratic
      
        x1 = getCoor(n1,1);
        x2 = getCoor(n2,1);
      
        real xmin,xmax;
        if ( x1<x2 ){ xmin=x1; xmax=x2; }
        else   {      xmin=x2; xmax=x1; }
        real xdiff = (xmax-xmin)*df;
        if( x < xmin - xdiff || x > xmax + xdiff ) 
	  return false;
        df *= 0.5; //using a smaller value of safety factor for the next test
      }
    in_elm_counter++;
  }
  else if ( nsd==2 ) {
    real x = gpt(1), y = gpt(2);
    int n;
    real x1,x2,x3,y1,y2,y3;
    real df; // safety factor to test if inside element
  
    if( isBoundingBoxesComputed() )
      { 
        box_counter++;
        if( x < bounding_box(element,1) || x > bounding_box(element,2) ||
	    y < bounding_box(element,3) || y > bounding_box(element,4) )
	  return false;
	// prepare for element test
        df = 1e-10*exp(getGridLevelNo()*0.742);
        n = loc2glob( element, 1 );
        x1 = getCoor(n,1); y1 = getCoor(n,2);
        n = loc2glob( element, 2 );
        x2 = getCoor(n,1); y2 = getCoor(n,2);
        n = loc2glob( element, 3 );
        x3 = getCoor(n,1); y3 = getCoor(n,2);
      }
    else
      {
        box_counter++;
        df = 2e-10*exp(getGridLevelNo()*0.742);
        
        n = loc2glob( element, 1 );
        x1 = getCoor(n,1); y1 = getCoor(n,2);
        n = loc2glob( element, 2 );
        x2 = getCoor(n,1); y2 = getCoor(n,2);
        n = loc2glob( element, 3 );
        x3 = getCoor(n,1); y3 = getCoor(n,2);
        
        real xmin,xmax;
        if ( x1<x2 ){ xmin=x1; xmax=x2; }
        else   {      xmin=x2; xmax=x1; }
        if (x3<xmin) xmin=x3;
        else if (x3>xmax) xmax=x3;
        real xdiff = (xmax-xmin)*df;
        if( x < xmin - xdiff || x > xmax + xdiff ) 
  	  return false;
        
        real ymin,ymax;
        if ( y1<y2 ){ ymin=y1; ymax=y2; }
        else   {      ymin=y2; ymax=y1; }
        if (y3<ymin) ymin=y3;
        else if (y3>ymax) ymax=y3;
        real ydiff = (ymax-ymin)*df;
        if( y < ymin - ydiff || y > ymax + ydiff ) 
  	  return false;
        df *= 0.5; //using a smaller value of safety factor for the next test
    }
  
    in_elm_counter++;
  
    real x13 = x1 - x3,   y31 = y3 - y1,   x32 = x3 - x2,   y23 = y2 - y3;
    real det = x13 * y23 - x32 * y31;
    
    real meps = -df*det;
    real x03 = x - x3, y03 = y - y3;
    real detphi1 = y23 * x03 + x32 * y03; 
    if( detphi1 < meps )  
      return false;
    real detphi2 = y31 * x03 + x13 * y03;
    if( detphi2 < meps )  
      return false;
    real detphi3 = det - detphi1 - detphi2;
    if( detphi3 < meps )  
      return false;
  }
  else if ( nsd==3 ) {
    real x = gpt(1), y = gpt(2), z = gpt(3);
    real df; // safety factor to test if inside element
    
    if( isBoundingBoxesComputed() )
      {
        box_counter++;
        if( x < bounding_box(element,1) || x > bounding_box(element,2) ||
  	    y < bounding_box(element,3) || y > bounding_box(element,4) ||
  	    z < bounding_box(element,5) || z > bounding_box(element,6) )
  	return false;
        df = 1e-10*exp(getGridLevelNo()*0.742);
      }
    else
      {
        box_counter++;
	int n1,n2,n3,n4;
	real x1,x2,x3,x4, pt,min,max,diff;
        df = 2e-10*exp(getGridLevelNo()*0.742);
        n1 = loc2glob( element, 1 );
        n2 = loc2glob( element, 2 );
        n3 = loc2glob( element, 3 );
        n4 = loc2glob( element, 4 );
  
        for(int d=1;d<=nsd;d++){
  	  x1 = getCoor(n1,d);
  	  x2 = getCoor(n2,d); 
  	  x3 = getCoor(n3,d);
  	  x4 = getCoor(n4,d);
  	  if ( x1<x2 ){ min=x1; max=x2; }
  	  else   {      min=x2; max=x1; }
  	  if (x3<min) min=x3;
  	  else if (x3>max) max=x3;
  	  if (x4<min) min=x4;
  	  else if (x4>max) max=x4;
  	  diff = (max-min)*df;
  	  pt = gpt(d);
  	  if( pt < min - diff || pt > max + diff ) 
	    return false;
        }
        df*=0.5; // use a stricter rule for inclusion
      } 
    in_elm_counter++;

    FET4.refill(element);
    real u,r,s,t;
    FET4.barycentric(x,y,z,u,r,s,t);
    real mdf = -df;
    if( u < mdf || r < mdf || s < mdf || t < mdf )
      return false; // the point is outside the tetrahedron
  }
// if not returned yet, the point (x) is inside the element
  return true;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: initSearchStatAllGrids ()
//-----------------------------------------------------------------------------
{
  GridFEAdT* grid = getFinestGrid();
  while( ! grid->isThisCoarsestGrid() ){
    grid->initSearchStat();
    grid = grid->getParentGrid();
  }
  grid->initSearchStat();
}

//-----------------------------------------------------------------------------
void GridFEAdT:: printSearchStatAllGrids (Os os)
//-----------------------------------------------------------------------------
{
  GridFEAdT* grid = getFinestGrid();
  int no_of_levels = grid->getGridLevelNo();
  int total_box=0, total_elm_tests=0;
  while( no_of_levels-- ){
    int boxtests = grid->getNoBoxTests ();
    int eletests = grid->getNoElmTests ();
    total_box += boxtests;
    total_elm_tests += eletests;
    os <<" level = "<< no_of_levels+1
       <<" box tests = "<< boxtests 
       <<" ele tests = "<< eletests <<endl;
    grid = grid->getParentGrid();
  }
  os <<" total on all levels: "
     <<" total box tests = "<< total_box
     <<" total ele tests = "<< total_elm_tests <<endl;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: removeBoundingBoxes()
//-----------------------------------------------------------------------------
{
  bounding_box.redim(0,0);
  setBoundingBoxesComputed(false);
}


//-----------------------------------------------------------------------------
void GridFEAdT:: computeBoundingBoxes()
//-----------------------------------------------------------------------------
{
  bounding_box.redim(nel,2*nsd);
  setBoundingBoxesComputed(true);
// taking a safety factor twice the one used in the inside test.
  const real df = 2*1e-10*exp(getGridLevelNo()*0.742);
  if ( nsd==1 ) {
    real xmin,xmax;
    
    for( int e=1; e<=nel; e++ ) {
      int n1 = loc2glob( e, 1 );
      int n2 = loc2glob( e, getNoNodesInElm(e) );// works for ElmB2n and ElmB3n
      
      real x1 = getCoor(n1,1);
      real x2 = getCoor(n2,1);
      
      if ( x1<x2 ){ xmin=x1; xmax=x2; }
      else   {      xmin=x2; xmax=x1; }
      
      real xdiff = (xmax-xmin)*df;
      
      bounding_box(e,1)=xmin-xdiff;
      bounding_box(e,2)=xmax+xdiff;
    }
  }
  else if ( nsd==2 ) {
    
    real xmin,xmax,ymin,ymax;
    
    for( int e=1; e<=nel; e++ ) {
      int n = loc2glob( e, 1 );
      real x1 = getCoor(n,1),  y1 = getCoor(n,2);
      n = loc2glob( e, 2 );
      real x2 = getCoor(n,1),  y2 = getCoor(n,2);
      n = loc2glob( e, 3 );
      real x3 = getCoor(n,1),  y3 = getCoor(n,2);
      
      if ( x1<x2 ){ xmin=x1; xmax=x2; }
      else   {      xmin=x2; xmax=x1; }
      if ( x3<xmin )         xmin=x3;
      else if  ( x3>xmax )   xmax=x3;
      
      if ( y1<y2 ){ ymin=y1; ymax=y2; }
      else   {      ymin=y2; ymax=y1; }
      if ( y3<ymin )         ymin=y3;
      else if  ( y3>ymax )   ymax=y3;
      
      real xdiff = (xmax-xmin)*df;
      real ydiff = (ymax-ymin)*df;
      
      bounding_box(e,1)=xmin-xdiff;
      bounding_box(e,2)=xmax+xdiff;
      bounding_box(e,3)=ymin-ydiff;
      bounding_box(e,4)=ymax+ydiff;
    }
  }
  else {  // if(nsd==3) {
    
    real xmin,xmax,ymin,ymax,zmin,zmax;
    
    for( int e=1; e<=nel; e++ ) {
      
      int n = loc2glob( e, 1 );
      real x1 = coor(n,1), y1 = coor(n,2), z1 = coor(n,3);
      n = loc2glob( e, 2 );
      real x2 = coor(n,1), y2 = coor(n,2), z2 = coor(n,3);
      n = loc2glob( e, 3 );
      real x3 = coor(n,1), y3 = coor(n,2), z3 = coor(n,3);
      n = loc2glob( e, 4 );
      real x4 = coor(n,1), y4 = coor(n,2), z4 = coor(n,3);
      
      if ( x1<x2 ){ xmin=x1; xmax=x2; }
      else   {      xmin=x2; xmax=x1; }
      if ( x3<xmin )         xmin=x3;
      else if  ( x3>xmax )   xmax=x3;
      if ( x4<xmin )         xmin=x4;
      else if  ( x4>xmax )   xmax=x4;
      
      if ( y1<y2 ){ ymin=y1; ymax=y2; }
      else   {      ymin=y2; ymax=y1; }
      if ( y3<ymin )         ymin=y3;
      else if  ( y3>ymax )   ymax=y3;
      if ( y4<ymin )         ymin=y4;
      else if  ( y4>ymax )   ymax=y4;
      
      if ( z1<z2 ){ zmin=z1; zmax=z2; }
      else   {      zmin=z2; zmax=z1; }
      if ( z3<zmin )         zmin=z3;
      else if  ( z3>zmax )   zmax=z3;
      if ( z4<zmin )         zmin=z4;
      else if  ( z4>zmax )   zmax=z4;
      
      real xdiff = (xmax-xmin)*df;
      real ydiff = (ymax-ymin)*df;
      real zdiff = (zmax-zmin)*df;
      
      bounding_box(e,1)=xmin-xdiff;
      bounding_box(e,2)=xmax+xdiff;
      bounding_box(e,3)=ymin-ydiff;
      bounding_box(e,4)=ymax+ydiff;
      bounding_box(e,5)=zmin-zdiff;
      bounding_box(e,6)=zmax+zdiff;
    }
  }
}

//-----------------------------------------------------------------------------
void GridFEAdT:: presentGridHier (Os os)
//-----------------------------------------------------------------------------
{
  int total_number_of_grids = getFinestGridLevelNo();
  os << endl;
  os << " Hierarchical finite element mesh (GridFEAdT): "<< endl;
  os << endl;
  os << " This is mesh number " << getGridLevelNo() <<" out of "
     << total_number_of_grids << " meshes." << endl;
  os << endl;
  
  GridFEAdT* grid = getCoarsestGrid();
  
  for( int i=1; i<=total_number_of_grids; i++ ){
    os << " Level " << i << " is of type " << grid->elm_tp.chars() 
       << ", has " << grid->getNoNodes() << " nodes and "
       << grid->getNoElms() << " elements." << endl;
    grid = grid->getChildGrid();
  }
  os << endl;  
}

//-----------------------------------------------------------------------------
void GridFEAdT:: printGrid (Os os) 
//-----------------------------------------------------------------------------
// Print out all edges in the grid, assumes that the node numbers are
// ordered, works for linear and quadratic triangles in 
{
  
  if(nsd==1){
    int n;
    real minx = getCoor( 1,1 );
    real maxx = getCoor( 1,1 );
    for ( n=1; n<=nno; n++ ){
      real p = getCoor( n,1 );
      if(p<minx) minx=p;
      else if(p>maxx) maxx=p;
    }
    real diff = (maxx-minx)/200;
    for ( n=1; n<=nno; n++ ){
      real p = getCoor( n,1 );
// make a small vertical line at the node
      os <<p<<" 0 0"<<endl;  os <<p<<" "<<diff<<" 0"<<endl<<endl;
    }
    return;
  }
  
  if(nsd==3) os<<"$data=curve3d"<<endl;
  
  if ( elm_tp != "ElmT6n2D" ) {
//  additional_info.initNeighbor( *this );
    additional_info.initNeighbor(*this,true,true,false);
    NeighborFE& neighbor = additional_info.neighbor;
    
    Ptv(real) p;
    int nstart,nstop;
    for ( int n=1; n<=nno; n++ ) {
      nstart=neighbor.couplingsIrow(n);
      nstop=neighbor.couplingsIrow(n+1);
      p = getCoor( n );
      for(int k=nstart;k<nstop; k++){
	int nn = neighbor.couplingsJcol(k);
	if ( nn >= n ) break;
        os << "% lc=1\n";
	p.print( os ); os <<endl; getCoor(nn).print( os ); os <<endl;os <<endl;
      }
    }
  }
  else {
    if(nsd==3){
      Ptv(real) p1,p2,p3,p4;
      for( int e=1; e<=nel; e++ ){
	p1 = getCoor( loc2glob( e, 1 ) );
	p2 = getCoor( loc2glob( e, 2 ) );
	p3 = getCoor( loc2glob( e, 3 ) );
	p4 = getCoor( loc2glob( e, 4 ) );
	p1.print( os ); os <<endl; p2.print( os ); os <<endl;os <<endl;
	p2.print( os ); os <<endl; p3.print( os ); os <<endl;os <<endl;
	p3.print( os ); os <<endl; p1.print( os ); os <<endl;os <<endl;
	p1.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
	p2.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
	p3.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
      }
    }
    else {
      for( int e=1; e<=nel; e++ ){
	int n = 3;
	for( int i1=1; i1<=n; i1++ ){
	  int i2 = (i1==n) ? 1 : i1+1;
	  Ptv(real) p1 = getCoor( loc2glob( e, i1 ) );
	  Ptv(real) p2 = getCoor( loc2glob( e, i2 ) );
	  p1.print( os ); os << " 0" <<endl;
	  p2.print( os ); os << " 0" <<endl;
	  os << endl;  
	}
      }
    }
  }
  os<<"$end"<<endl;
}

//-----------------------------------------------------------------------------
void GridFEAdT:: printGrid(Os os, const int sbdtype) const
//-----------------------------------------------------------------------------
// Print out all edges in the grid, assumes that the node numbers are
// ordered, works for linear and quadratic triangles in 
{
  if( ! (nsd==2||nsd==3) ){
    warningFP("GridFEAdT:: printGrid","not implemented for nsd=%d",nsd);
    return;
  }
  if(nsd==3) os<<"$data=curve3d"<<endl;
  Ptv(real) p1,p2,p3,p4;
  for( int e=1; e<=nel; e++ ){
    if( getMaterialType (e) == sbdtype ) {
      if(nsd==3){
	p1 = getCoor( loc2glob( e, 1 ) );
	p2 = getCoor( loc2glob( e, 2 ) );
	p3 = getCoor( loc2glob( e, 3 ) );
	p4 = getCoor( loc2glob( e, 4 ) );
	p1.print( os ); os <<endl; p2.print( os ); os <<endl;os <<endl;
	p2.print( os ); os <<endl; p3.print( os ); os <<endl;os <<endl;
	p3.print( os ); os <<endl; p1.print( os ); os <<endl;os <<endl;
	p1.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
	p2.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
	p3.print( os ); os <<endl; p4.print( os ); os <<endl;os <<endl;
      } else if(nsd==2) {
	for( int i1=1; i1<=3; i1++ ){
	  int i2 = (i1==3) ? 1 : i1+1;
	  p1 = getCoor( loc2glob( e, i1 ) );
	  p2 = getCoor( loc2glob( e, i2 ) );
	  p1.print( os ); os << " 0" <<endl;
	  p2.print( os ); os << " 0" <<endl;
	  os << endl;  
	}
      }
    }
  }
  os<<"$end"<<endl;
}


//-----------------------------------------------------------------------------
void GridFEAdT:: printIndicators(Os os) 
//-----------------------------------------------------------------------------
// Print out the edges in the grid, which lies on the outer boundary
// or between different subdomaintypes
{
  if( ! (nsd==2||nsd==3) ){
    warningFP("GridFEAdT:: printIndicators","not implemented for nsd=%d",nsd);
    return;
  }

  if( !isNeighborsComputed() )// Is the element-element info computed?
    initNeighbor();
  
  Ptv(real) p1,p2,p3;
 
  if(nsd==2) {
    for(int e=1;e<=nel;e++){
      for( int i1=1; i1<=3; i1++ ){
	int sbde = getMaterialType (e);
	int nei = getElmNeighbor(e,i1);
	if( nei == 0 || getMaterialType (nei) != sbde ) {
	  if(nei) os<<"%lc="<<max(sbde,getMaterialType (nei))<<endl;
	  else os<<"%lc="<<sbde<<endl;
	  int i2 = (i1==3) ? 1 : i1+1;
	  p1 = getCoor( loc2glob( e, i1 ) );
	  p2 = getCoor( loc2glob( e, i2 ) );
	  p1.print( os ); os << " 0" <<endl;
	  p2.print( os ); os << " 0" <<endl;
	  os << endl;  
	}
      }
    } 
    os<<"$end"<<endl;
    return;
  }
  
  int i,e,n;
//  additional_info.initNeighbor(*this);
  additional_info.initNeighbor(*this,true,true,false);
  NeighborFE& neighbor = additional_info.neighbor;
  SparseDS ds;
  int coupsize = neighbor.couplingsSize ();
  ds.redim( nno, coupsize );
  int nno1 = nno+1;
  for ( n=1; n<=nno1; n++ )
    ds.irow(n) = neighbor.couplingsIrow(n);
  for ( n=1; n<=coupsize; n++ )
    ds.jcol(n) = neighbor.couplingsJcol(n);
  
  MatSparse(real) plotted(ds);
  
  plotted.fill(1);
  
  os<<"$data=curve3d"<<endl;
  int n1,n2,n3;
  for( e=1; e<=nel; e++ ){
    int sbde = getMaterialType (e);
    for(i=1;i<=4;i++){
      int nei = getElmNeighbor(e,i);
      if( nei == 0 || getMaterialType (nei) != sbde ) {
//      if( nei != 0 && getMaterialType (nei) != sbde ) {
	getGlobNodesOnFace( n1, n2, n3, i, e );
	if( plotted(n1,n2) ) {
	  if(nei) os<<"%lc="<<max(sbde,getMaterialType (nei))<<endl;
	  else os<<"%lc="<<sbde<<endl;
	  os<<getCoor(n1,1)<<" "<<getCoor(n1,2)<<" "<<getCoor(n1,3)<<endl
	    <<getCoor(n2,1)<<" "<<getCoor(n2,2)<<" "<<getCoor(n2,3)<<endl
	    <<endl;
	  plotted(n1,n2) = plotted(n2,n1) = 0;
	}
	if( plotted(n2,n3) ) {
	  if(nei) os<<"%lc="<<max(sbde,getMaterialType (nei))<<endl;
	  else os<<"%lc="<<sbde<<endl;
	  os<<getCoor(n2,1)<<" "<<getCoor(n2,2)<<" "<<getCoor(n2,3)<<endl
	    <<getCoor(n3,1)<<" "<<getCoor(n3,2)<<" "<<getCoor(n3,3)<<endl
	    <<endl;
	  plotted(n2,n3) = plotted(n3,n2) = 0;
	}
	if( plotted(n3,n1) ) {
	  if(nei) os<<"%lc="<<max(sbde,getMaterialType (nei))<<endl;
	  else os<<"%lc="<<sbde<<endl;
	  os<<getCoor(n1,1)<<" "<<getCoor(n1,2)<<" "<<getCoor(n1,3)<<endl
	    <<getCoor(n3,1)<<" "<<getCoor(n3,2)<<" "<<getCoor(n3,3)<<endl
	    <<endl;
	  plotted(n1,n3) = plotted(n3,n1) = 0;
	}
      }
    }
  }
  os<<"$end"<<endl;
  
}

//#include <VecSimple_VecSimple_int.h>

//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: createPartOfGrid(int subdomainno, int fromlevel)
//-----------------------------------------------------------------------------
// Create a copy of a part of the grid hierarchy
{
  presentGridHier(s_o); 

  int finest_level = getFinestGridLevelNo();
  if(fromlevel > finest_level )
    errorFP("GridFEAdT:: createPartOfGrid","0");
  GridFEAdT* g = getCoarsestGrid();
  int i,d,l,n,e;
  for(l=1;l < fromlevel; l++)  // find the base grid
    g = g->getChildGrid();

// copy the subdomain partition on the g grid to its finer grids.

// check that child information is computed
  VecSimple(int) checknodes;
// assumes that the numbering is consistent for hierarchy
  int noNodes = g->getFinestGrid()->getNoNodes();
  checknodes.redim(noNodes); // same checking for all grids
  checknodes.fill(0);
  int ncounter = 0;
  int subd;

  GridFEAdT* partGrid=NULL;
  GridFEAdT* coarsePartGrid;
  GridFEAdT* gg = g;
  GridFEAdT* ggchild;

  VecSimplest(VecSimple(int)) elmpart2glob(finest_level);
  VecSimplest(VecSimple(int)) elmglob2part(finest_level); 

  for( l=fromlevel; l <= finest_level; l++) {
    if( l < finest_level)
      ggchild = gg->getChildGrid();
    int noElms = gg->getNoElms();
    int ecounter = 0; 
    for(e=1;e<=noElms;e++) {
      subd = gg->getMaterialType(e);
      if( subd == subdomainno ) { // count number of elements and nodes
	ecounter++;
	for(i=1;i<=maxnne;i++)
	  if( ! checknodes( gg->loc2glob(e,i) ) )
	    checknodes( gg->loc2glob(e,i) ) = ++ncounter;
      }
      if( l < finest_level ) {
	int startchild = gg->getFirstChild(e);
	int stopchild = gg->getFirstChild(e+1);
	for(n=startchild;n<stopchild;n++)
//	  ggchild->material_type(n) = subd;
	 ggchild->setMaterialType(n,subd);
      }
    }
    if(l>fromlevel)
      coarsePartGrid = partGrid;
    partGrid = new GridFEAdT();
    partGrid->redim( nsd, ncounter, ecounter, nne, getNoBoInds(), maxnne, true);
    partGrid->setElmType ( gg->getElmType(1) );
    partGrid->setNonUniformMesh ();
    for(i=1;i<=getNoBoInds();i++)
      partGrid->putBoIndName(gg->getBoIndName (i), i );

    if(l>fromlevel) {
      partGrid->setParentGrid(coarsePartGrid);
      partGrid->setCoarsestGrid(false);
      coarsePartGrid->setChildGrid(partGrid);
      coarsePartGrid->setFinestGrid(false);
    }
    int new_node=0;
    for(n=1;n<=noNodes;n++)
      if(checknodes(n)) {
	++new_node;
	int nn = checknodes(n);
	for(d=1;d<=getNoBoInds();d++)
	  if( gg->boNode (n,d))
	    partGrid->setBoInd(nn, d);
	for(d=1;d<=nsd;d++)
	  partGrid->putCoor(gg->getCoor(n,d),nn,d);
      }
    if ( new_node != ncounter )
      fatalerrorFP("GridFEAdT:: createPartOfGrid","1");
    elmglob2part(l).redim(noElms);
    elmpart2glob(l).redim(ecounter);
    ecounter = 0; 
    for(e=1;e<=noElms;e++) {
      subd = gg->getMaterialType(e);
      if( subd == subdomainno ) {
	ecounter++;
	elmpart2glob(l)(ecounter) = e;
	elmglob2part(l)(e) = ecounter;

	for(i=1;i<=maxnne;i++)
	  partGrid->putLoc2Glob (checknodes(gg->loc2glob(e,i)), ecounter, i);
      }
    }
    if(l>fromlevel) {
      partGrid->parent.redim(ecounter);
      for(e=1;e<=ecounter;e++) {
	int part_elm = e;
	int glob_elm = elmpart2glob(l)(part_elm);
	int glob_elm_par = gg->getParent_eff(glob_elm);
	int part_elm_par = elmglob2part(l-1)(glob_elm_par);
	partGrid->parent(part_elm) =  part_elm_par;
//	partGrid->parent(e) = 
//	  elmglob2part(l-1)(gg->getParent(elmpart2glob(l)(e)));
      }
      partGrid->setParentInfo(true);
      GridFEAdT* pglob = gg->getParentGrid();
      int noelminpar = coarsePartGrid->getNoElms();
      coarsePartGrid->pos_child.redim(noelminpar+1);
      for(e=1;e<=noelminpar;e++) {
	int part_elm = e;
	int glob_elm = elmpart2glob(l-1)(part_elm);
	int glob_elm_par = pglob->getFirstChild(glob_elm);
	int part_elm_par = elmglob2part(l)(glob_elm_par);
	coarsePartGrid->pos_child( e ) = part_elm_par;
      }
      coarsePartGrid->pos_child( noelminpar+1 ) = partGrid->getNoElms()+1;
      coarsePartGrid->setChildrenInfo( true );
      elmglob2part(l-1).redim(0);
    }
	
    if( ! gg->isThisFinestGrid() ) {
      gg = ggchild;
      ggchild = gg->getChildGrid();
    }

  }
  partGrid->presentGridHier(s_o);
  GridFEAdT* gh = partGrid->getCoarsestGrid();
  while( ! gh->isThisFinestGrid() ) {
    if( gh->checkParentInfo())
      s_o<<"checkParentInfo is ok\n";
    else
      s_o<<"checkParentInfo is not ok\n";
    if( gh->checkChildrenInfo())
      s_o<<"checkChildrenInfo is ok\n";
    else
      s_o<<"checkChildrenInfo is not ok\n";
    gh = gh->getChildGrid();
  }
    if( gh->checkParentInfo())
      s_o<<"checkParentInfo is ok\n";
    else
      s_o<<"checkParentInfo is not ok\n";
    if( gh->checkChildrenInfo())
      s_o<<"checkChildrenInfo is ok\n";
    else
      s_o<<"checkChildrenInfo is not ok\n";

  return partGrid;
}


//-----------------------------------------------------------------------------
GridFEAdT* GridFEAdT:: createPartOfGrid(VecSimple(int)& partition, 
					  int start_level,
					  int stop_level)
//-----------------------------------------------------------------------------
// Create a copy of a part of the grid hierarchy
// partition contains the element numbers of the elements in the
// start_level grid which should be part of the partition
// copy the partition part of the children grid up to stop_level
{

  int finest_level = getFinestGridLevelNo();
  if(start_level < 1 ){
    warningFP("GridFEAdT:: createPartOfGrid",
	      "\nstart level number %d to small, will use start level=1",
	      start_level);
    start_level = 1;
  }
  if(start_level > stop_level ){
    warningFP("GridFEAdT:: createPartOfGrid",
	      "\nstart level number %d is larger than stop level %d,"
	      "will use stop level = start level",start_level,stop_level);
    stop_level = start_level;
  }
  if(stop_level > finest_level ){
    warningFP("GridFEAdT:: createPartOfGrid",
	      "\nstop level number %d is larger than finest level %d,"
	      "will use stop level = finest level",stop_level,finest_level);
    stop_level = finest_level;
  }

  GridFEAdT* g = getCoarsestGrid();
  int i,d,l,n,e;
  for( l=1; l < start_level; l++ )  // find the start level grid
    g = g->getChildGrid();

// test if the element numbers in partition are legal
  int noPaElm = partition.size();
  int maxElm = 0;
  int minElm = g->getNoElms();
  for(e=1;e<=noPaElm;e++) {
    int pe = partition(e);
    if ( pe > maxElm )
      maxElm = pe;
    else if ( pe < minElm )
      minElm = pe;
  }
  if( maxElm > g->getNoElms() ){
    fatalerrorFP("GridFEAdT:: createPartOfGrid",
		 "\nToo large element number are used in partition"
		 "max number = %d, max in grid = %d",maxElm,g->getNoElms());
    return g;
  }
  if( minElm < 1 ){
    fatalerrorFP("GridFEAdT:: createPartOfGrid",
		 "\nToo small element number are used in partition"
		 "min number = %d, min in grid = %d",minElm,1);
    return g;
  }

  GridFEAdT* gstop = getCoarsestGrid();
  for(l=1;l < stop_level; l++) 
    gstop = gstop->getChildGrid();

  int noNodes = gstop->getNoNodes();
  VecSimple(int) checknodes(noNodes);// same checking for all grids
  checknodes.redim(noNodes);
// assumes that the numbering is consistent for hierarchy
  checknodes.fill(0);
  int ncounter = 0;

  GridFEAdT* partGrid=NULL;
  GridFEAdT* coarsePartGrid=NULL;

// will not use the first levels if startlevel > 1;
  VecSimplest(VecSimple(int)) partitionAllLevels(stop_level); 
  partitionAllLevels(start_level) = partition; // initialize

  int pe;
  int childcounter = 0;
  int stopchild;
  for( l = start_level; l <= stop_level; l++ ) {
    VecSimple(int)& pl = partitionAllLevels(l);
    int noElms = pl.size();
// need to go through and count number of nodes and children (if not finest)
    if( l < stop_level ) {
      childcounter = 0;
      for(e=1;e<=noElms;e++) {
	pe = pl(e);
	for(i=1;i<=maxnne;i++)
	  if( ! checknodes( g->loc2glob(pe,i) ) )// check if used before
	    checknodes( g->loc2glob(pe,i) ) = ++ncounter;
	childcounter += g->getNoChildren(pe);
      }
    } else { //  l == stop_level
      for(e=1;e<=noElms;e++) {
	pe = pl(e);
	for(i=1;i<=maxnne;i++)
	  if( ! checknodes( g->loc2glob(pe,i) ) )
	    checknodes( g->loc2glob(pe,i) ) = ++ncounter;
      }
    }
// now we know the size of the partitioned grid
    coarsePartGrid = partGrid; // not used for the start level
    partGrid = new GridFEAdT();
    int gnbind = g->getNoBoInds();
    partGrid->redim( g->nsd, ncounter, noElms, g->nne, gnbind, g->maxnne, 
		     g->onemat);
    partGrid->setElmType ( g->elm_tp );
    partGrid->setNonUniformMesh ();
    partGrid->setRefinementMethod( g->getRefinementMethod() );
    partGrid->setRegBisectOrShort( g->getRegBisectOrShort() ); // only in 2D
    partGrid->setNodeForEqualTest( g->useNodeForEqualTest() ); // only in 3D


    for( i=1; i<=gnbind; i++ )
      partGrid->putBoIndName(g->getBoIndName (i), i );

// set all the used nodes in the partitioned grid
    MatSimple(char)&  ind = g->boundaryData().indicatorAccess();
    MatSimple(char)&  nind = partGrid->boundaryData().indicatorAccess();

    int noNodesIng = g->getNoNodes();
    for( n=1; n<=noNodesIng; n++ )// go through all nodes in the grid
      if(checknodes(n)) {
	int nn = checknodes(n);
	for(d=1;d<=gnbind;d++) if( ind(n,d) == '1' ) nind(nn,d) = '1';
	for(d=1;d<=nsd;d++) partGrid->putCoor(g->getCoor(n,d),nn,d);
      }
// set subdomaintype
    if(! g->onemat ){
      partGrid->material_type.redim(noElms); // not sure if this is necessary
      for(e=1;e<=noElms;e++)
	partGrid->material_type(e) = g->material_type(pl(e));
    }

    if( l < stop_level ) {
      partGrid->pos_child.redim(noElms+1);
      partGrid->pos_child(1) = 1;
      partGrid->setChildrenInfo( true ); 
      VecSimple(int)& pl1 = partitionAllLevels(l+1);
      pl1.redim(childcounter);
      childcounter = 1;
      for(e=1;e<=noElms;e++) {
	pe = pl(e);
	for(i=1;i<=maxnne;i++)
	  partGrid->putLoc2Glob (checknodes(g->loc2glob(pe,i)), e, i);
	stopchild = g->getFirstChild(pe+1);
	for(n=g->getFirstChild(pe);n<stopchild;n++)
	  pl1(childcounter++) = n;
	partGrid->pos_child(e+1) = childcounter;
      } 
    } else { // l == stop_level
      for(e=1;e<=noElms;e++) {
	pe = pl(e);
	for(i=1;i<=maxnne;i++)
	  partGrid->putLoc2Glob (checknodes(g->loc2glob(pe,i)), e, i);
      } 
    }

    pl.redim(0); // is not needed anymore
    if( l > start_level ) {
      partGrid->setParentGrid(coarsePartGrid);
      partGrid->setCoarsestGrid(false);
      coarsePartGrid->setChildGrid(partGrid);
      coarsePartGrid->setFinestGrid(false);
      partGrid->setParentInfo(true);
      partGrid->parent.redim(noElms);
      int noElmsCoarse = coarsePartGrid->getNoElms();
      for(e=1;e<=noElmsCoarse;e++) {
    	stopchild = coarsePartGrid->getFirstChild(e+1);
	for(n=coarsePartGrid->getFirstChild(e);n<stopchild;n++)
	  partGrid->parent(n) = e;
      }
    }
	
    if( ! g->isThisFinestGrid() )
      g = g->getChildGrid();
  }

// s_o<<"in Hier: address: "<<this<<'\n'; s_o->flush();


  return partGrid;
}

